/*
 * Copyright 2005, U.S. Robotics Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND U.S. Robotics GRANTS NO WARRANTIES OF
 * ANY KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. U.S.
 * Robotics SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */
/*`
 * Broadcom Home Gateway Reference Design
 * Web Page Configuration Support Routines
 *
 * Copyright 2005, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 * 
 */

#ifdef WEBS
#include <webs.h>
#include <uemf.h>
#include <ej.h>
#else /* !WEBS */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>
#include <httpd.h>
#endif /* WEBS */


/* Required for hash table headers*/
#if defined(linux)
/* Use SVID search */
#define __USE_GNU
#include <search.h>
#endif

#include <typedefs.h>
#include <proto/ethernet.h>
#include <bcmparams.h>
#include <bcmnvram.h>
#include <bcmutils.h>
#include <shutils.h>
#include <netconf.h>
#include <nvparse.h>
#include <wlutils.h>
#include <bcmcvar.h>
#include <ezc.h>
#include <bcmconfig.h>
#include <trxhdr.h>
/* USR */
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <usrhdr.h>
#include "projects.h"
#include "wan_detection.h"
#include "check_fw.h"
#include <syslog.h>
#include <opencrypto.h>
#include <time.h>
#include <epivers.h>


int internal_init(void);

static char * encrypt_var(char *varname, char *ctext, int ctext_len, char *ptext, int *ptext_len,char *key, int keylen);
static char * decrypt_var(char *varname, char *ptext, int ptext_len, char *ctext, int *ctext_len,char *key, int keylen);
static char * rfctime(const time_t *timep);
static char * reltime(unsigned int seconds);
static bool find_ethaddr_in_list (void *ethaddr, struct maclist *list);

#define wan_prefix(unit, prefix)	snprintf(prefix, sizeof(prefix), "wan%d_", unit)
/* USR */
#define SIGNAL_MIN -200
//-95
#define SIGNAL_MAX -10


/*
 * Country names and abbreviations from ISO 3166
 */
typedef struct {
	char *name;     /* Long name */
	char *abbrev;   /* Abbreviation */
} country_name_t;
country_name_t country_names[];     /* At end of this file */

struct variable variables[];
extern struct nvram_tuple router_defaults[];

enum {
	NOTHING,
	REBOOT,
	RESTART,
};
//next 2 variables are used for backup/restore function
static int restore_ret;
static char posterr_msg[255];

#define ERR_MSG_SIZE sizeof(posterr_msg)
#if defined(linux)

#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sys/klog.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <net/if.h>

typedef u_int64_t u64;
typedef u_int32_t u32;
typedef u_int16_t u16;
typedef u_int8_t u8;
#include <linux/ethtool.h>
#include <linux/sockios.h>
#include <net/if_arp.h>

#define sys_restart() kill(1, SIGHUP)
#define sys_reboot() kill(1, SIGTERM)
#define sys_stats(url) eval("stats", (url))

#ifndef WEBS

#define MIN_BUF_SIZE	4096

//USR global variables used in site_survey
wlc_ssid_t new_ssid_g;
char eaddr_g[40] = "";   //used to store MAC address of AP during repeater site-survey
int error = 0;   /* used to keep track of errors */
int flag = 0;	//used to detect reporting errors

#ifdef CLIENT
static const char * const apply_header =
"<head>"
"<title>U.S. Robotics 802.11g Gaming Adapter and Ethernet Bridge: INFO</title>"
"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">"
"<style type=\"text/css\">"
"DIV#idTextHeader { float: right; padding: 0.2em 0.4em; margin-right: -0.4em; font-family: Verdana, Arial, sans-serif; font-size: x-large; font-weight: bold; color: white; background-color: red; letter-spacing: -0.05em; }"
"DIV#idTextHeader SUP { font-size: x-small; }"
"BODY { background-color: white; margin: 0 0 0.5em; font-family: Verdana, Arial, sans-serif; font-size: small; }"
"DIV#idContent { margin: 0 0.7em }"
"DIV#idHeader { padding: 0.7em 0.5em; margin-bottom: 0.5em; font-family: Verdana, Arial, sans-serif; font-weight: bold; font-size: x-large; color: white; background-color: red; letter-spacing: 1pt; }"
"H3 { font-weight: bold; letter-spacing: 1pt; padding: 0.4em; font-size: large; width: 60%%; border: 1px solid black; background-color: #cccc99; }"
"P.clsSave { border-top: 1px solid #cccc99; padding-top: 0.5em; }"
"INPUT.clsBtnSave { font-weight: bold; color: black; background: #cccc99; border: 2px outset silver; }"
".clsNotice { border: 3px double #EB1E24; padding: 1em; margin: 0.5em 0 0.5em 0; width: 60%%; background-color: #F8F8FF; }"
"P.clsCopyright { font-family: Verdana, Arial, sans-serif; font-size: xx-small; white-space: nowrap; color: black; border-top: 2px solid black; padding-top: 0.5em; padding-left: 0.5em }"
"</style>"
"</head>"
"<body>"
"<div id=\"idTextHeader\">U.S.Robotics<sup>&reg;</sup></div>"
"<div id=\"idHeader\">Information</div>"
"<div id=\"idContent\">"
"<form method=\"get\" action=\"apply.cgi\">"
;
#else
static const char * const apply_header =
"<head>"
#ifdef ROUTER
"<title>U.S. Robotics Wireless MAXg Router</title>"
#endif
#ifdef AP
"<title>U.S. Robotics Wireless MAXg Access Point</title>"
#endif
#ifdef APB
"<title>U.S. Robotics MAXg Professional Access Point</title>"
#endif
#ifdef REPEATER
"<title>U.S. Robotics Wireless MAXg Range Extender</title>"
#endif
#ifdef GAMING
"<title>U.S. Robotics Wireless MAXg Gaming Adapter</title>"
#endif
#ifdef BRIDGE
"<title>U.S. Robotics Wireless MAXg Bridge</title>"
#endif
"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">"
"<style type=\"text/css\">"
"DIV#idTextHeader { float: right; padding: 0.2em 0.4em; margin-right: -0.4em; font-family: Verdana, Arial, sans-serif; font-size: x-large; font-weight: bold; color: white; background-color: red; letter-spacing: -0.05em; }"
"DIV#idTextHeader SUP { font-size: x-small; }"
"BODY { background-color: white; margin: 0 0 0.5em; font-family: Verdana, Arial, sans-serif; font-size: small; }"
"DIV#idContent { margin: 0 0.7em }"
"DIV#idHeader { padding: 0.7em 0.5em; margin-bottom: 0.5em; font-family: Verdana, Arial, sans-serif; font-weight: bold; font-size: x-large; color: white; background-color: red; letter-spacing: 1pt; }"
"H3 { font-weight: bold; letter-spacing: 1pt; padding: 0.4em; font-size: large; width: 60%%; border: 1px solid #202080; background-color: #F8F8FF; }"
"P.clsSave { border-top: 1px solid #202080; padding-top: 0.5em; }"
"INPUT.clsBtnSave { font-weight: bold; color: white; background: #202080; border: 2px outset silver; }"
".clsNotice { border: 3px double red; padding: 1em; margin: 0.5em 0 0.5em 0; width: 60%%; background-color: #F8F8FF; }"
"P.clsCopyright { font-family: Verdana, Arial, sans-serif; font-size: xx-small; white-space: nowrap; color: #202080; border-top: 2px solid #202080; padding-top: 0.5em; padding-left: 0.5em }"
"</style>"
"</head>"
"<body>"
"<div id=\"idTextHeader\">U.S.Robotics<sup>&reg;</sup></div>"
"<div id=\"idHeader\">Information</div>"
"<div id=\"idContent\">"
"<form method=\"get\" action=\"apply.cgi\">"
;
#endif

#ifdef AP
static const char * const apply_footer_check_fw_new =
"<div class=\"clsNotice\"><p>A new version of firmware for this access point is available.</p>"
"<p>Firmware version: <strong>%s</strong></p>"
"<p>Click <strong>Download</strong> to save the new firmware image on your computer.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Download\" OnClick=\"document.location.href='" DEFAULT_5451_URL "';\">"
"</form>"
"<form method=\"post\" action=\"apply.cgi\""
"<p>Press <strong>Continue</strong> to proceed to the <strong>Device</strong> page where you can continue the upgrade process by following steps 3 and 4 in the Upgrade Access Point section.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='device.asp';\"></div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#endif

#ifdef APB
static const char * const apply_footer_check_fw_new =
"<div class=\"clsNotice\"><p>A new version of firmware for this access point is available.</p>"
"<p>Firmware version: <strong>%s</strong></p>"
"<p>Click <strong>Download</strong> to save the new firmware image on your computer.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Download\" OnClick=\"document.location.href='" DEFAULT_5453_URL "';\">"
"</form>"
"<form method=\"post\" action=\"apply.cgi\""
"<p>Press <strong>Continue</strong> to proceed to the <strong>Device</strong> page where you can continue the upgrade process by following steps 3 and 4 in the Upgrade Access Point section.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='device.asp';\"></div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#endif

#ifdef REPEATER
static const char * const apply_footer_check_fw_new =
"<div class=\"clsNotice\"><p>A new version of firmware for this wireless range extender is available.</p>"
"<p>Firmware version: <strong>%s</strong></p>"
"<p>Click <strong>Download</strong> to save the new firmware image on your computer.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Download\" OnClick=\"document.location.href='" DEFAULT_5441_URL "';\">"
"</form>"
"<form method=\"post\" action=\"apply.cgi\""
"<p>Press <strong>Continue</strong> to proceed to the <strong>Device</strong> page where you can continue the upgrade process by following steps 3 and 4 in the Upgrade Range Extender section.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='device.asp';\"></div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#endif

#ifdef GAMING
static const char * const apply_footer_check_fw_new =
"<div class=\"clsNotice\"><p>A new version of firmware for this gaming adapter is available.</p>"
"<p>Firmware version: <strong>%s</strong></p>"
"<p>Click <strong>Download</strong> to save the new firmware image on your computer.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Download\" OnClick=\"document.location.href='" DEFAULT_5431_URL "';\">"
"</form>"
"<form method=\"post\" action=\"apply.cgi\""
"<p>Press <strong>Continue</strong> to proceed to the <strong>Device</strong> page where you can continue the upgrade process by following steps 3 and 4 in the Upgrade Gaming Adapter section.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='device.asp';\"></div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#endif

#ifdef BRIDGE
static const char * const apply_footer_check_fw_new =
"<div class=\"clsNotice\"><p>A new version of firmware for this bridge is available.</p>"
"<p>Firmware version: <strong>%s</strong></p>"
"<p>Click <strong>Download</strong> to save the new firmware image on your computer.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Download\" OnClick=\"document.location.href='" DEFAULT_5432_URL "';\">"
"</form>"
"<form method=\"post\" action=\"apply.cgi\""
"<p>Press <strong>Continue</strong> to proceed to the <strong>Device</strong> page where you can continue the upgrade process by following steps 3 and 4 in the Upgrade Bridge section.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='device.asp';\"></div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#endif

#ifdef ROUTER
static const char * const apply_header_detect =
"<head>"
"<title>U.S. Robotics 5461 Router</title>"
"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\">"
"<style type=\"text/css\">"
"DIV#idTextHeader { float: right; padding: 0.2em 0.4em; margin-right: -0.4em; font-family: Verdana, Arial, sans-serif; font-size: x-large; font-weight: bold; color: white; background-color: red; letter-spacing: -0.05em; }"
"DIV#idTextHeader SUP { font-size: x-small; }"
"BODY { background-color: white; margin: 0 0 0.5em; font-family: Verdana, Arial, sans-serif; font-size: small; }"
"DIV#idContent { margin: 0 0.7em }"
"DIV#idHeader { padding: 0.7em 0.5em; margin-bottom: 0.5em; font-family: Verdana, Arial, sans-serif; font-weight: bold; font-size: x-large; color: white; background-color: red; letter-spacing: 1pt; }"
"H3 { font-weight: bold; letter-spacing: 1pt; padding: 0.4em; font-size: large; width: 60%%; border: 1px solid #202080; background-color: #F8F8FF; }"
"P.clsSave { border-top: 1px solid #202080; padding-top: 0.5em; }"
"INPUT.clsBtnSave { font-weight: bold; color: white; background: #202080; border: 2px outset silver; }"
".clsNotice { border: 3px double red; padding: 1em; margin: 0.5em 0 0.5em 0; width: 60%%; background-color: #F8F8FF; }"
"P.clsCopyright { font-family: Verdana, Arial, sans-serif; font-size: xx-small; white-space: nowrap; color: #202080; border-top: 2px solid #202080; padding-top: 0.5em; padding-left: 0.5em }"
"</style>"
"</head>"
"<body>"
"<div id=\"idTextHeader\">U.S.Robotics<sup>&reg;</sup></div>"
"<div id=\"idHeader\">Establish Internet Connection</div>"
"<div id=\"idContent\">"
"<form method=\"get\" action=\"apply.cgi\">"
;

static const char * const apply_footer_check_fw_new =
"<div class=\"clsNotice\"><p>A new version of firmware for this router is available.</p>"
"<p>Firmware version: <strong>%s</strong></p>"
"<p>Click <strong>Download</strong> to save the new firmware image on your computer.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Download\" OnClick=\"document.location.href='" DEFAULT_5461_URL "';\">"
"</form>"
"<form method=\"post\" action=\"apply.cgi\""
"<p>Press <strong>Continue</strong> to proceed to the <strong>Device</strong> page where you can continue the upgrade process by following steps 3 and 4 in the Upgrade Router section.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='device.asp';\"></div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;

static const char * const apply_footer_detect =
"<script>document.location = \'setup_sec.asp\';</script>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;

static const char * const apply_footer_nodetect =
"<p>To re-enter your user name and password and try again, press <strong>Retry</strong>.&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
"<input type=\"button\" class=\"clsBtnSave\" name=\"btnSubmit\" value=\"Retry -->     \" onclick=\"document.location.href='%s';\"></p>"
"<p>To continue with your current user name and password, press <strong>Continue</strong>.&nbsp;"
"<input type=\"button\" class=\"clsBtnSave\" name=\"btnSubmit\" value=\"Continue -->\" onclick=\"document.location.href = 'setup_sec.asp';\"></p>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;

static const char * const apply_footer_restart =
"<div id=\"idContinue\" style=\"display: none;\">"
"<p class=\"clsSave\">Please press the <b>Continue</b> button to proceed to the <b>Internet</b> page.</p>"
"<p><input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" onclick=\"document.location.href='wan.asp';\"></p>"
"</div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"<script type=\"text/javascript\" defer>"
"onTimer(10);"
"function setVisibility(id, state){ var elt = document.getElementById(id); if ((elt == null) || !elt.style || (elt.style.display == null)) return;"
"if (state) elt.style.display = \"\"; else elt.style.display = \"none\";}"
"function onTimer(iCount){ var s; if (iCount == 1) s = iCount + \" second \"; else s = iCount + \" seconds \"; document.getElementById(\"idSeconds\").innerHTML = s;"
"if (iCount > 0) setTimeout(\"onTimer(\" + (iCount - 1) + \")\", 1000); else { setVisibility(\"idWait\", false); setVisibility(\"idContinue\", true);}}"
"</script>"
"</body>"
;
#endif

static const char * const apply_footer =
"<p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='status.asp';\"></p>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;

static const char * const apply_footer_error =
"<p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='%s';\"></p>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;

#ifdef ROUTER
static const char * const apply_footer_lan =
"</div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"<script type=\"text/javascript\" defer>"
"onTimer(15);"
"function setVisibility(id, state){ var elt = document.getElementById(id); if ((elt == null) || !elt.style || (elt.style.display == null)) return;"
"if (state) elt.style.display = \"\"; else elt.style.display = \"none\";}"
"function onTimer(iCount){ var s; if (iCount == 1) s = iCount + \" second \"; else s = iCount + \" seconds \"; document.getElementById(\"idSeconds\").innerHTML = s;"
"if (iCount > 0) setTimeout(\"onTimer(\" + (iCount - 1) + \")\", 1000); else { setVisibility(\"idWait\", false); setVisibility(\"idContinue\", true);}}"
"</script>"
"</body>"
;
#else
static const char * const apply_footer_lan =
"<p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='\" http://<% nvram_get(\"lan_ipaddr\"); %>/lan.asp \"';\"></p>"
//"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick='document.location.href=http://<% nvram_get(\"lan_ipaddr\"); %>/lan.asp;'></p>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#endif

#ifdef CLIENT
static const char * const apply_footer_finish =
"<p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='%s';\"></p>"
"</form>"
"<form method=\"post\" action=\"apply.cgi\">"
"<input type=\"submit\" class=\"clsBtnSave\" name=\"action\" value=\"Finish\" OnClick=\"document.location.href='status.asp';\">"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;

static const char * const apply_footer_save =
"<div id=\"idContinue\" style=\"display: none;\">"
"<p><b>You may now connect your U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge to your gaming console or other Ethernet-ready device.<br>"
"Power cycle your U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge.<br>"
"Be sure to reconnect your wireless router or access point to your computer, if necessary.</b></p>"
"<p class=\"clsSave\">Please press the <b>Continue</b> button to proceed to the <b>Status</b> page.</p>"
"<p><input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" onclick=\"document.location.href='status.asp';\"></p>"
"</div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"<script type=\"text/javascript\" defer>"
"onTimer(20);"
"function setVisibility(id, state){ var elt = document.getElementById(id); if ((elt == null) || !elt.style || (elt.style.display == null)) return;"
"if (state) elt.style.display = \"\"; else elt.style.display = \"none\";}"
"function onTimer(iCount){ var s; if (iCount == 1) s = iCount + \" second \"; else s = iCount + \" seconds \"; document.getElementById(\"idSeconds\").innerHTML = s;"
"if (iCount > 0) setTimeout(\"onTimer(\" + (iCount - 1) + \")\", 1000); else { setVisibility(\"idWait\", false); setVisibility(\"idContinue\", true);}}"
"</script>"
"</body>"
;

static const char * const apply_footer_check_fw_new =
"<div class=\"clsNotice\"><p>A new version of firmware for this wireless client is available.</p>"
"<p>Firmware version: <strong>%s</strong></p>"
"<p>Click <strong>Download</strong> to save the new firmware image on your computer.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Download\" OnClick=\"document.location.href='" DEFAULT_5431_URL "';\">"
"</form>"
"<form method=\"post\" action=\"apply.cgi\""
"<p>Press <strong>Continue</strong> to proceed to the <strong>Device</strong> page where you can continue the upgrade process by following steps 3 and 4 in the Upgrade wireless client section.</p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='device.asp';\"></div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#else
static const char * const apply_footer_finish =
"<p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='status.asp';\"></p>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;
#endif

static const char * const apply_footer_restore =
"<div id=\"idContinue\" style=\"display: none;\">"
"<p class=\"clsSave\">Please press the <b>Continue</b> button to proceed to the <b>Setup</b> page.</p>"
"<p><input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" onclick=\"document.location.href='index.asp';\"></p>"
"</div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"<script type=\"text/javascript\" defer>"
#ifdef CLIENT
"onTimer(20);"
#else
"onTimer(15);"
#endif
"function setVisibility(id, state){ var elt = document.getElementById(id); if ((elt == null) || !elt.style || (elt.style.display == null)) return;"
"if (state) elt.style.display = \"\"; else elt.style.display = \"none\";}"
"function onTimer(iCount){ var s; if (iCount == 1) s = iCount + \" second \"; else s = iCount + \" seconds \"; document.getElementById(\"idSeconds\").innerHTML = s;"
"if (iCount > 0) setTimeout(\"onTimer(\" + (iCount - 1) + \")\", 1000); else { setVisibility(\"idWait\", false); setVisibility(\"idContinue\", true);}}"
"</script>"
"</body>"
;

static const char * const apply_footer_check_fw_nonew =
"<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the <strong>Device</strong> page.</p>"
"<p>"
"<input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" OnClick=\"document.location.href='%s';\"></p>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"</body>"
;

static const char * const apply_footer_reboot =
"<div id=\"idContinue\" style=\"display: none;\">"
"<p class=\"clsSave\">Please press the <b>Continue</b> button to proceed to the <b>Status</b> page.</p>"
"<p><input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" onclick=\"document.location.href='status.asp';\"></p>"
"</div>"
"</form>"
"</div>"
"<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>"
"<script type=\"text/javascript\" defer>"
#ifdef CLIENT
"onTimer(20);"
#else
"onTimer(15);"
#endif
"function setVisibility(id, state){ var elt = document.getElementById(id); if ((elt == null) || !elt.style || (elt.style.display == null)) return;"
"if (state) elt.style.display = \"\"; else elt.style.display = \"none\";}"
"function onTimer(iCount){ var s; if (iCount == 1) s = iCount + \" second \"; else s = iCount + \" seconds \"; document.getElementById(\"idSeconds\").innerHTML = s;"
"if (iCount > 0) setTimeout(\"onTimer(\" + (iCount - 1) + \")\", 1000); else { setVisibility(\"idWait\", false); setVisibility(\"idContinue\", true);}}"
"</script>"
"</body>"
;

/* USR - validation */
void validation_page(webs_t wp)
{
	flag = 1;
	websHeader(wp);
	websWrite(wp, (char_t *) apply_header);
	websWrite(wp, "<!--\r\n");
	websWrite(wp, "1:: <p>(Invalid |Error, |This setting is invalid\\.)\r\n");
	websWrite(wp, "::: ::START HERE-->");
}
/* Upgrade from remote server or socket stream */
static int
sys_upgrade(char *url, FILE *stream, int *total)
{
	char upload_fifo[] = "/tmp/uploadXXXXXX";
	FILE *fifo = NULL;
	char *write_argv[] = { "write", upload_fifo, "linux", NULL };
	pid_t pid;
	char *buf = NULL;
	int count, ret = 0;
	long flags = -1;
	int size = BUFSIZ;

	assert(stream);
	assert(total);
	
	if (url)
		return eval("write", url, "linux");

	/* Feed write from a temporary FIFO */
	if (!mktemp(upload_fifo) ||
	    mkfifo(upload_fifo, S_IRWXU) < 0||
	    (ret = _eval(write_argv, NULL, 0, &pid)) ||
	    !(fifo = fopen(upload_fifo, "w"))) {
		if (!ret)
			ret = errno;
		goto err;
	}

	/* Set nonblock on the socket so we can timeout */
	if ((flags = fcntl(fileno(stream), F_GETFL)) < 0 ||
	    fcntl(fileno(stream), F_SETFL, flags | O_NONBLOCK) < 0) {
		ret = errno;
		goto err;
	}

	/*
	* The buffer must be at least as big as what the stream file is
	* using so that it can read all the data that has been buffered
	* in the stream file. Otherwise it would be out of sync with fn
	* select specially at the end of the data stream in which case
	* the select tells there is no more data available but there in
	* fact is data buffered in the stream file's buffer. Since no
	* one has changed the default stream file's buffer size, let's
	* use the constant BUFSIZ until someone changes it.
	*/
	if (size < MIN_BUF_SIZE)
		size = MIN_BUF_SIZE;
	if ((buf = malloc(size)) == NULL) {
		ret = ENOMEM;
		goto err;
	}

	/* Pipe the rest to the FIFO */
	cprintf("sys_upgrade: Upgrading\n");
	
	while (total && *total) {
		if (waitfor(fileno(stream), 5) <= 0)
			break;
		count = safe_fread(buf, 1, size, stream);
		if (!count && (ferror(stream) || feof(stream)))
			break;
		*total -= count;
		safe_fwrite(buf, 1, count, fifo);
		cprintf(".");
	}
	fclose(fifo);
	fifo = NULL;

	/* Wait for write to terminate */
	waitpid(pid, &ret, 0);
	cprintf("done\n");

	/* Reset nonblock on the socket */
	if (fcntl(fileno(stream), F_SETFL, flags) < 0) {
		ret = errno;
		goto err;
	}

 err:
 	if (buf)
		free(buf);
	if (fifo)
		fclose(fifo);
	unlink(upload_fifo);
	return ret;
}

#endif /* WEBS */

/*USR - retrieve MAC used in site_survey for REPEATER*/
static int
ej_retrieve_mac(int eid, webs_t wp, int argc, char_t **argv)
{
	printf("retrieve_mac: %s \n", eaddr_g);
	return websWrite(wp, "%s", eaddr_g);
}

/* Dump firewall log */
static int
ej_dumplog(int eid, webs_t wp, int argc, char_t **argv)
{
	char buf[4096], *line, *next, *s;
	int len, ret = 0;

	time_t tm;
	char *verdict, *src, *dst, *proto, *spt, *dpt;

	if (klogctl(3, buf, 4096) < 0) {
		websError(wp, 400, "Insufficient memory\n");
		return -1;
	}
	
	for (next = buf; (line = strsep(&next, "\n"));) {
		if (!strncmp(line, "<4>DROP", 7))
			verdict = "denied";
		else if (!strncmp(line, "<4>ACCEPT", 9))
			verdict = "accepted";
		else
			continue;

		/* Parse into tokens */
		s = line;
		len = strlen(s);
		while (strsep(&s, " "));

		/* Initialize token values */
		time(&tm);
		src = dst = proto = spt = dpt = "n/a";

		/* Set token values */
		for (s = line; s < &line[len] && *s; s += strlen(s) + 1) {
			if (!strncmp(s, "TIME=", 5))
				tm = strtoul(&s[5], NULL, 10);
			else if (!strncmp(s, "SRC=", 4))
				src = &s[4];
			else if (!strncmp(s, "DST=", 4))
				dst = &s[4];
			else if (!strncmp(s, "PROTO=", 6))
				proto = &s[6];
			else if (!strncmp(s, "SPT=", 4))
				spt = &s[4];
			else if (!strncmp(s, "DPT=", 4))
				dpt = &s[4];
		}
		ret += websWrite(wp, "<tr><td>");
		ret += websWrite(wp, "%s %s connection %s to %s:%s from %s:%s\n",
				 rfctime(&tm), proto, verdict, dst, dpt, src, spt);
		ret += websWrite(wp, "</td></tr>");
		//send the log to syslog
		//openlog("USR5461", LOG_PID, LOG_DAEMON);
		//syslog(LOG_SYSLOG | LOG_INFO, "%s %s connection %s to %s:%s from %s:%s\n",
		//		 rfctime(&tm), proto, verdict, dst, dpt, src, spt);
		//closelog();
	}

	return ret;
}

static int
ej_syslog(int eid, webs_t wp, int argc, char_t **argv)
{
	FILE *fp;
	char buf[256] = "/sbin/logread > ";
	char tmp[] = "/tmp/log.XXXXXX";
	int ret;

	if (!nvram_match("log_ram_enable", "1")) {
		websError(wp, 400, "\"Syslog in RAM\" is not enabled.\n");
		return (-1);
	}
	
	mktemp(tmp);
	strcat(buf, tmp);
	system(buf);
	
	fp = fopen(tmp, "r");

	unlink(tmp);

	if (fp == NULL) {
		websError(wp, 400, "logread error\n");
		return (-1);
	}

	websWrite(wp, "<pre>");	

	ret = 0;
	while(fgets(buf, sizeof(buf), fp))
		ret += websWrite(wp, buf);

	ret += websWrite(wp, "</pre>");

	fclose(fp);
		
	return (ret);
}

struct lease_t {
	unsigned char chaddr[16];
	u_int32_t yiaddr;
	u_int32_t expires;
	char hostname[64];
};

/* Dump leases in <tr><td>hostname</td><td>MAC</td><td>IP</td><td>expires</td></tr> format */
static int
ej_lan_leases(int eid, webs_t wp, int argc, char_t **argv)
{
	FILE *fp = NULL;
	struct lease_t lease;
	int i;
	struct in_addr addr;
	unsigned long expires = 0;
	char sigusr1[] = "-XX";
	int ret = 0;

	/* Write out leases file */
	sprintf(sigusr1, "-%d", SIGUSR1);
	eval("killall", sigusr1, "udhcpd");

	if (!(fp = fopen("/tmp/udhcpd.leases", "r"))) {
		printf("udhcpd.leases 0\n");
		return 0;
	}
//wl clients	
	int unit;
	char tmp[100], prefix[] = "wlXXXXXXXXXX_";
	char *name;
	struct maclist *auth;
	int max_sta_count, maclist_size;
	char wl_mac[100];
	char tmp1[3];
	bool flag_lan;
	
	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	/* buffers and length */
	max_sta_count = 256;
	maclist_size = sizeof(auth->count) + max_sta_count * sizeof(struct ether_addr);

	auth = malloc(maclist_size);
	
	/* query wl for authenticated sta list */
	strcpy((char*)auth, "authe_sta_list");
		
	if (auth && !wl_ioctl(name, WLC_GET_VAR, auth, maclist_size)) {
	/* build authenticated/associated/authorized sta list */
		for (i = 0; i < auth->count; i ++) {
			//printf("i= %d\n", i);
			char ea[ETHER_ADDR_STR_LEN];
			snprintf(&wl_mac[i], sizeof(ea), "%s", ether_etoa((void *)&auth->ea[i], ea));
			//printf("wl_clients: %s\n", &wl_mac[i]);
		}
	}
	else
		printf("wireless authentication error!\n");
	//wl_clients end
	
	while (fread(&lease, sizeof(lease), 1, fp)) {
		/* Do not display reserved leases */
		if (ETHER_ISNULLADDR(lease.chaddr)) {
			printf("null lease!\n");
			continue;
		}
		strncpy(tmp,"", sizeof(tmp));
		flag_lan = FALSE;
		
		for (i = 0; i < 6; i++) {
			//printf("chaddr: %02x\n", lease.chaddr[i]);
			snprintf(tmp1, sizeof(tmp1), "%02X", lease.chaddr[i]);
			strcat(tmp, tmp1);
			//printf("tmp1: %s\n", tmp1);
			//printf("tmp: %s\n", tmp);
			if (i != 5) strcat(tmp,":");
		}
		for (i = 0; i < auth->count; i ++) {
			//printf("i-1= %d\n", i);
			//printf("wl_mac: %s\n", &wl_mac[i]);
			if (!strcmp(tmp, &wl_mac[i])) {
				flag_lan = TRUE;
				ret += websWrite(wp, "<tr><td>wireless</td>");
				continue;
			}
		}
		if (!flag_lan)
			ret += websWrite(wp, "<tr><td>wired</td>");	
		ret += websWrite(wp, "<td>%s</td><td>", lease.hostname);
		for (i = 0; i < 6; i++) {
			ret += websWrite(wp, "%02X", lease.chaddr[i]);
			if (i != 5) ret += websWrite(wp, ":");
		}
		addr.s_addr = lease.yiaddr;
		ret += websWrite(wp, "</td><td>%s</td><td>", inet_ntoa(addr));
		expires = ntohl(lease.expires);
		if (!expires)
			ret += websWrite(wp, "Expired");
		else
			ret += websWrite(wp, "%s", reltime(expires));
		ret += websWrite(wp, "</td></tr>");
	}

	fclose(fp);
	/* error/exit */
	if (auth) free(auth);
	
	return ret;
}

/* Dump leases in <tr><td>hostname</td><td>MAC</td><td>IP</td><td>expires</td></tr> format */
static int
ej_lan_clients(int eid, webs_t wp, int argc, char_t **argv)
{
	FILE *fp = NULL;
	struct lease_t lease;
	int i;
	char sigusr1[] = "-XX";
	int ret = 0;

	/* Write out leases file */
	sprintf(sigusr1, "-%d", SIGUSR1);
	eval("killall", sigusr1, "udhcpd");

	if (!(fp = fopen("/tmp/udhcpd.leases", "r")))
		return 0;

	while (fread(&lease, sizeof(lease), 1, fp)) {
		/* Do not display reserved leases */
		if (ETHER_ISNULLADDR(lease.chaddr))
			continue;
		ret += websWrite(wp, "<option value=\"");              
		for (i = 0; i < 6; i++) {
			ret += websWrite(wp, "%02X", lease.chaddr[i]);
			if (i != 5) ret += websWrite(wp, ":");
		}
		ret += websWrite(wp, "\"> %s (", lease.hostname);
		for (i = 0; i < 6; i++) {
			ret += websWrite(wp, "%02X", lease.chaddr[i]);
			if (i != 5) ret += websWrite(wp, ":");
		}
		ret += websWrite(wp, ")</option>");
	}

	fclose(fp);

	return ret;
}

/* return printer status */
static int
ej_get_printer(int eid, webs_t wp, int argc, char_t **argv)
{
	int ret = 0;
	
	if ( (access("/dev/usb/lp0", 0)) == -1) {
		//printf("can't open lp0!\n");
		ret += websWrite(wp, "0"); 
		return ret;
	}
	
	//printf("printer is ready!\n");
	ret += websWrite(wp, "1");
	return ret;
}

/* Renew lease */
static int
sys_renew(void)
{
	int unit;
	char tmp[NVRAM_BUFSIZE];
	char *str = NULL;
	int pid;

	if ((unit = atoi(nvram_safe_get("wan_unit"))) < 0)
		unit = 0;

	snprintf(tmp, sizeof(tmp), "/var/run/udhcpc%d.pid", unit);
	if ((str = file2str(tmp))) {
		pid = atoi(str);
		free(str);
		return kill(pid, SIGUSR1);
	}
	
	return -1;
}

/* Release lease */
static int
sys_release(void)
{
	int unit;
	char tmp[NVRAM_BUFSIZE];
	char *str= NULL;
	int pid;

	if ((unit = atoi(nvram_safe_get("wan_unit"))) < 0)
		unit = 0;
	
	snprintf(tmp, sizeof(tmp), "/var/run/udhcpc%d.pid", unit);
	if ((str = file2str(tmp))) {
		pid = atoi(str);
		free(str);
		return kill(pid, SIGUSR2);
	}
	
	return -1;
}

#ifdef __CONFIG_NAT__
#ifdef ROUTER
#define sin_addr(s) (((struct sockaddr_in *)(s))->sin_addr)

/* Return WAN link state 
static int
ej_wan_link(int eid, webs_t wp, int argc, char_t **argv)
{
	char *wan_ifname;
	int s;
	struct ifreq ifr;
	struct ethtool_cmd ecmd;
	FILE *fp= NULL;
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";

	if ((unit = atoi(nvram_safe_get("wan_unit"))) < 0)
		unit = 0;
	wan_prefix(unit, prefix);

	// non-exist and disabled 
	if (nvram_match(strcat_r(prefix, "proto", tmp), "") ||
	    nvram_match(strcat_r(prefix, "proto", tmp), "disabled")) {
		return websWrite(wp, "N/A");
	}
	// PPPoE connection status 
	else if (nvram_match(strcat_r(prefix, "proto", tmp), "pppoe")) {
		wan_ifname = nvram_safe_get(strcat_r(prefix, "pppoe_ifname", tmp));
		if ((fp = fopen(strcat_r("/tmp/ppp/link.", wan_ifname, tmp), "r"))) {
			fclose(fp);
			return websWrite(wp, "Connected");
		} else
			return websWrite(wp, "Disconnected");
	}
	// Get real interface name 
	else
		wan_ifname = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	// Open socket to kernel 
	if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		return websWrite(wp, "N/A");

	// Check for hardware link 
	//strncpy(ifr.ifr_name, wan_ifname, IFNAMSIZ);
	strncpy(ifr.ifr_name, "eth0", IFNAMSIZ);
	ifr.ifr_data = (void *) &ecmd;
	ecmd.cmd = ETHTOOL_GSET;
	if (ioctl(s, SIOCETHTOOL, &ifr) < 0) {
		close(s);
		return websWrite(wp, "Unknown");
	}
	if (!ecmd.speed) {
		close(s);
		return websWrite(wp, "Disconnected");
	}

	// Check for valid IP address 
	//strncpy(ifr.ifr_name, wan_ifname, IFNAMSIZ);
	strncpy(ifr.ifr_name, "eth0", IFNAMSIZ);
	if (ioctl(s, SIOCGIFADDR, &ifr) < 0) {
		close(s);
		return websWrite(wp, "Connecting");
	}

	// Otherwise we are probably configured 
	close(s);
	return websWrite(wp, "Connected");
}*/

/* Display IP Address lease */
static int
ej_wan_lease(int eid, webs_t wp, int argc, char_t **argv)
{
	unsigned long expires = 0;
	int ret = 0;
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";

	if ((unit = atoi(nvram_safe_get("wan_unit"))) < 0)
		unit = 0;
	wan_prefix(unit, prefix);
	
	if (nvram_match(strcat_r(prefix, "proto", tmp), "dhcp")) {
		char *str;
		time_t now;

		snprintf(tmp, sizeof(tmp), "/tmp/udhcpc%d.expires", unit); 
		if ((str = file2str(tmp))) {
			expires = atoi(str);
			free(str);
		}
		time(&now);
		if (expires <= now)
			ret += websWrite(wp, "Expired");
		else
			ret += websWrite(wp, "%s", reltime(expires - now));
	} else
		ret += websWrite(wp, "N/A");

	return ret;
}
#endif   /* ROUTER */
#endif	/* __CONFIG_NAT__ */

/* Report sys up time */
static int
ej_sysuptime(int eid, webs_t wp, int argc, char_t **argv)
{
	char *str = file2str("/proc/uptime");
	if (str) {
		unsigned int up = atoi(str);
		free(str);
		return websWrite(wp, reltime(up));
	}
	return websWrite(wp, "N/A");
}

/*#ifdef __CONFIG_NAT__
// Return a list of wan interfaces (eth0/eth1/eth2/eth3) 
static int
ej_wan_iflist(int eid, webs_t wp, int argc, char_t **argv)
{
	char name[IFNAMSIZ], *next;
	int ret = 0;
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";
	char ea[64];
	int s;
	struct ifreq ifr;

	// current unit # 
	if ((unit = atoi(nvram_safe_get("wan_unit"))) < 0)
		unit = 0;
	wan_prefix(unit, prefix);

	if ((s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
		return errno;

	// build wan interface name list 
	foreach(name, nvram_safe_get("wan_ifnames"), next) {
		strncpy(ifr.ifr_name, name, IFNAMSIZ);
		if (ioctl(s, SIOCGIFHWADDR, &ifr))
			continue;
		ret += websWrite(wp, "<option value=\"%s\" %s>%s (%s)</option>", name,
				 nvram_match(strcat_r(prefix, "ifname", tmp), name) ? "selected" : "",
				 name, ether_etoa(ifr.ifr_hwaddr.sa_data, ea));
	}

	close(s);

	return ret;
} 
#endif	// __CONFIG_NAT__ */


#endif /* vxworks */

static char *
rfctime(const time_t *timep)
{
	static char s[201];
	struct tm tm;

#if defined(linux)
	setenv("TZ", nvram_safe_get("time_zone"), 1);
#endif
	memcpy(&tm, localtime(timep), sizeof(struct tm));
	strftime(s, 200, "%a, %d %b %Y %H:%M:%S %z", &tm);
	return s;
}

static char *
reltime(unsigned int seconds)
{
	static char s[] = "XXXXX days, XX hours, XX minutes, XX seconds";
	char *c = s;

	if (seconds > 60*60*24) {
		c += sprintf(c, "%d days, ", seconds / (60*60*24));
		seconds %= 60*60*24;
	}
	if (seconds > 60*60) {
		c += sprintf(c, "%d hours, ", seconds / (60*60));
		seconds %= 60*60;
	}
	if (seconds > 60) {
		c += sprintf(c, "%d minutes, ", seconds / 60);
		seconds %= 60;
	}
	c += sprintf(c, "%d seconds", seconds);

	return s;
}

/* Report time in RFC-822 format */
static int
ej_localtime(int eid, webs_t wp, int argc, char_t **argv)
{
	time_t tm;

	time(&tm);
	return websWrite(wp, rfctime(&tm));
}

/*
static int
ej_wl_mode_list(int eid, webs_t wp, int argc, char_t **argv) 
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *name, *next;
	int ap = 0, sta = 0, wet = 0, wds = 0;
	char cap[WLC_IOCTL_SMLEN];
	char caps[WLC_IOCTL_SMLEN];

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	if (wl_get_val(name, "cap", (void *)caps, WLC_IOCTL_SMLEN))
		return -1;
	
	foreach(cap, caps, next) {
		if (!strcmp(cap, "ap"))
			ap = wds = 1;
		else if (!strcmp(cap, "sta"))
			sta = 1;
		else if (!strcmp(cap, "wet"))
			wet = 1;
        }

	if (ap)
		websWrite(wp, "<option value=\"ap\" %s>Access Point</option>\n", 
				nvram_match("wl_mode", "ap" ) ? "selected" : "");
	if (wds)
		websWrite(wp, "<option value=\"wds\" %s>Wireless Bridge</option>\n", 
				nvram_match("wl_mode", "wds" ) ? "selected" : "");
	if (wet)
		websWrite(wp, "<option value=\"wet\" %s>Wireless Ethernet</option>\n", 
			nvram_match("wl_mode", "wet" ) ? "selected" : "");

	return 0;

}

static int
ej_wl_inlist(int eid, webs_t wp, int argc, char_t **argv) 
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *name=NULL, *next=NULL;
	char cap[WLC_IOCTL_SMLEN];
	char caps[WLC_IOCTL_SMLEN];
	char *var=NULL, *item=NULL;

	if (ejArgs(argc, argv, "%s %s", &var, &item) < 2) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	if (wl_iovar_get(name, var, (void *)caps, WLC_IOCTL_SMLEN))
		return -1;

	foreach(cap, caps, next) {
		if (!strcmp(cap, item))
			return websWrite(wp, "1");
	}

	return websWrite(wp, "0");
}
*/
#ifdef REPEATER
static int
ej_wl_wds_status(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *macs=NULL, *next=NULL, *name=NULL;
	char mac[100];
	int i=0, len=0;
	sta_info_t *sta=NULL;
	char buf[300];
	int ret = 0;
	
	if (ejArgs(argc, argv, "%d", &i) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));
	macs = nvram_safe_get(strcat_r(prefix, "wds", tmp));
	
	foreach(mac, macs, next) {
		//if (i-- == 0) {
			len = sprintf(buf, "sta_info");
			ether_atoe(mac, (unsigned char *)&buf[len + 1]);
			//if (atoi(nvram_safe_get(strcat_r(prefix, "wds_timeout", tmp))) &&
			if(!wl_ioctl(name, WLC_GET_VAR, buf, sizeof(buf))) {
				sta = (sta_info_t *)buf;
				ret += websWrite(wp, "<tr><td class=\"clsStatic\">%s</td> %s</tr>", mac, (sta->flags & WL_WDS_LINKUP) ? "<td><img src=\"check.gif\" border=\"0\" /></td>" : "");
			}
		//}
	}	

	return ret;
}
#endif
/*
static int
ej_ses_button_display(int eid, webs_t wp, int argc, char_t **argv)
{
	return 1;
}


static int
ej_wl_radio_roam_option(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *name=NULL;
	int radio_status = 0;		

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0) {
		websError(wp, 400, "unit number variable doesn't exist\n");
		return -1;
	}
	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));
	wl_ioctl(name, WLC_GET_RADIO, &radio_status, sizeof (radio_status));
	if (!radio_status) // Radio on
		websWrite(wp, "<input type=\"submit\" name=\"action\" value=\"RadioOff\" >");
	else // Radio Off 
		websWrite(wp, "<input type=\"submit\" name=\"action\" value=\"RadioOn\" >");

	return 1;


}
static int 
wl_radio_onoff(webs_t wp, int disable)
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *name=NULL;
	char *interface_status=NULL;		
	uint radiomaskval = 0;


	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0) {
		websError(wp, 400, "unit number variable doesn't exist\n");
		return -1;
	}

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	interface_status = nvram_safe_get(strcat_r(prefix, "radio", tmp));

	if (interface_status != NULL) {
		if (!strcmp(interface_status, "1")) {
			snprintf(prefix, sizeof(prefix), "wl%d_", unit);
			name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));
			radiomaskval = WL_RADIO_SW_DISABLE << 16 | disable;
			wl_ioctl(name, WLC_SET_RADIO, &radiomaskval, sizeof (disable));
		}
		else {
			websWrite(wp, "Interface is not Enabled...");
			return 0;
		}		
	}
	else {
		websWrite(wp, "Interface  status UnKnown...");
		return 0;
	}


	return 0;

}
*/
/*
 * Example: 
 * lan_ipaddr=192.168.1.1
 * <% nvram_get("lan_ipaddr"); %> produces "192.168.1.1"
 * <% nvram_get("undefined"); %> produces ""
 */
static int
ej_nvram_get(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL, *c=NULL;
	int ret = 0;
	char temp[NVRAM_BUFSIZE] = "";
	char temp1[4];
	
	if (ejArgs(argc, argv, "%s", &name) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	assert(name);
	/*
	for (c = nvram_safe_get(name); *c; c++) {
		if (isprint((int) *c) &&
		    *c != '"' && *c != '&' && *c != '<' && *c != '>')
			ret += websWrite(wp, "%c", *c);
		else
			ret += websWrite(wp, "&#%d", *c);
	}*/
	for (c = nvram_safe_get(name); *c; c++) {
		if (*c != '"' && *c != '&' && *c != '<' && *c != '>') {
			sprintf(temp1, "%c", *c);
			strcat(temp, temp1);
		}
		else if (*c == '"')
			strcat(temp,"&quot;");
		else if (*c == '&')
			strcat(temp,"&amp;");
		else if (*c == '<')
			strcat(temp,"&lt;");
		else if (*c == '>')
			strcat(temp,"&gt;");
	}		
	//printf("c is: %s\n", c);
	//printf("mystring is: %s\n", temp);
	ret = websWrite(wp, "%s", temp);
	return ret; 
}

/*
 * Example: 
 * wan_proto=dhcp
 * <% nvram_match("wan_proto", "dhcp", "selected"); %> produces "selected"
 * <% nvram_match("wan_proto", "static", "selected"); %> does not produce
 */
static int
ej_nvram_match(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL, *match=NULL, *output=NULL;

	if (ejArgs(argc, argv, "%s %s %s", &name, &match, &output) < 3) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	assert(name);
	assert(match);
	assert(output);
	
	if (nvram_match(name, match))
		return websWrite(wp, output);

	return 0;
}	
/*
static int
ej_wme_match_op(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL, *match=NULL, *output=NULL;
	char word[256], *next=NULL;

	if (ejArgs(argc, argv, "%s %s %s", &name, &match, &output) < 3) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	assert(name);
	assert(match);
	assert(output);
	
	foreach(word, nvram_safe_get(name), next) {
		if (!strcmp(word, match))
			return websWrite(wp, output);
	}

	return 0;
}
*/
#ifdef CLIENT
static int
ej_nvram_adhoc(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name, *match, *output;
	int infra;

	if (ejArgs(argc, argv, "%s %s %s", &name, &match, &output) < 3) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	if ((infra = atoi(nvram_safe_get("wl_infra"))) == 0) {
		if (nvram_invmatch(name, match))
			return websWrite(wp, output);
	}
	else {
		if (nvram_match(name, match))
			return websWrite(wp, output);
	}
	return 0;
}

static int
wl_print_channel_list(webs_t wp, char *name, char *phytype, char *abbrev)
{
	int j, status = 0;
	wl_channels_in_country_t *cic = (wl_channels_in_country_t *)malloc(WLC_IOCTL_MAXLEN);
		
	assert(name);
	assert(phytype);
	assert(abbrev);
	
	if (!cic) {
		status = -1;
		goto exit;
	}

	cic->buflen = WLC_IOCTL_MAXLEN;
	strcpy(cic->country_abbrev, abbrev);
	if (!strcmp(phytype, "a"))
		cic->band = WLC_BAND_A;
	else if ((!strcmp(phytype, "b")) || (!strcmp(phytype, "g")))
		cic->band = WLC_BAND_B;
	else {
		status = -1;
		goto exit;
	}

	if (wl_ioctl(name, WLC_GET_CHANNELS_IN_COUNTRY, cic, cic->buflen) == 0) {
		if (cic->count == 0) {
			status = 0;
			goto exit;
		}
		websWrite(wp, "\t\tif (country == \"%s\")\n\t\t\tchannels = new Array(0",
			abbrev);
		for(j = 0; j < cic->count; j++)
			websWrite(wp, ", %d", cic->channel[j]);
		websWrite(wp,");\n");
	}

exit:
	if (cic)
		free((void *)cic);
	return status;
}

static int
ej_wl_country_list(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, i = 0, status = 0;
	char *name=NULL;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *phytype = NULL;
	wl_country_list_t *cl = (wl_country_list_t *)malloc(WLC_IOCTL_MAXLEN);
	country_name_t *cntry=NULL;
	char *abbrev=NULL;

	if (!cl) {
		status = -1;
		goto exit;
	}

	if (ejArgs(argc, argv, "%s", &phytype) < 1) {
		websError(wp, 400, "Insufficient args\n");
		status = -1;
		goto exit;
	}
	
	assert(phytype);
	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0) {
		status = -1;
		goto exit;
	}

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	cl->buflen = WLC_IOCTL_MAXLEN;
	cl->band_set = TRUE;

	if (!strcmp(phytype, "a"))
		cl->band = WLC_BAND_A;
	else if ((!strcmp(phytype, "b")) || (!strcmp(phytype, "g")))
		cl->band = WLC_BAND_B;
	else {
		status = -1;
		goto exit;
	}

	if (wl_ioctl(name, WLC_GET_COUNTRY_LIST, cl, cl->buflen) == 0) {
		websWrite(wp, "\t\tvar countries = new Array(");
		for(i = 0; i < cl->count; i++) {
			abbrev = &cl->country_abbrev[i*WLC_CNTRY_BUF_SZ];
			websWrite(wp, "\"%s\"", abbrev);
			if (i != (cl->count - 1))
				websWrite(wp, ", ");
		}
		websWrite(wp, ");\n");
		for(i = 0; i < cl->count; i++) {
			abbrev = &cl->country_abbrev[i*WLC_CNTRY_BUF_SZ];
			for(cntry = country_names;
				cntry->name && strcmp(abbrev, cntry->abbrev);
				cntry++); 
			websWrite(wp, "\t\tdocument.forms[0].wl_country_code[%d] = new Option(\"%s\", \"%s\");\n",
				i, cntry->name ? cntry->name : abbrev, abbrev);
		}
	}

exit:
	if (cl)
		free((void *)cl);
	return status;
}

static int
ej_wl_channel_list(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, i, status = 0;
	char *name=NULL;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *phytype = NULL;
	wl_country_list_t *cl = (wl_country_list_t *)malloc(WLC_IOCTL_MAXLEN);
	char *abbrev=NULL;

	if (!cl) {
		status = -1;
		goto exit;
	}

	if (ejArgs(argc, argv, "%s", &phytype) < 1) {
		websError(wp, 400, "Insufficient args\n");
		status = -1;
		goto exit;
	}

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0) {
		status = -1;
		goto exit;
	}

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	cl->buflen = WLC_IOCTL_MAXLEN;
	cl->band_set = TRUE;

	if (!strcmp(phytype, "a"))
		cl->band = WLC_BAND_A;
	else if ((!strcmp(phytype, "b")) || (!strcmp(phytype, "g")))
		cl->band = WLC_BAND_B;
	else {
		status = -1;
		goto exit;
	}

	if (wl_ioctl(name, WLC_GET_COUNTRY_LIST, cl, cl->buflen) == 0) {
		for(i = 0; i < cl->count; i++) {
			abbrev = &cl->country_abbrev[i*WLC_CNTRY_BUF_SZ];
			wl_print_channel_list(wp, name, phytype, abbrev);
		}
	}


exit:
	if (cl)
		free((void *)cl);
	return status;
}
#endif

static int
ej_wl_clients(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char tmp[100], prefix[] = "wlXXXXXXXXXX_";
	char *name=NULL, *macs=NULL, *next=NULL;
	struct maclist *auth=NULL;
	int max_sta_count, maclist_size;
	int i = 0;
	
	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	/* buffers and length */
	max_sta_count = 256;
	maclist_size = sizeof(auth->count) + max_sta_count * sizeof(struct ether_addr);

	auth = malloc(maclist_size);
	
	if (!auth)
		goto exit;

	/* query wl for authenticated sta list */
	strcpy((char*)auth, "authe_sta_list");
	if (wl_ioctl(name, WLC_GET_VAR, auth, maclist_size)) {
		//printf("EXIT AUTH!\n");
		goto exit;
	}
	/* //query wl for associated sta list 
	assoc->count = max_sta_count;

	if (wl_ioctl(name, WLC_GET_ASSOCLIST, assoc, maclist_size))
		goto exit;

	// query wl for authorized sta list 
	strcpy((char*)authorized, "autho_sta_list");
	if (wl_ioctl(name, WLC_GET_VAR, authorized, maclist_size))
		goto exit;  */
	/* add WDS list first */
	macs = nvram_safe_get("wl_wds");
	foreach(tmp, macs, next) {
		websWrite(wp, "\"%s\",", tmp);
	}
	/* build authenticated/associated/authorized sta list */
	int count = auth->count;
	//printf("count= %d\n", count);
	for (i = 0; i < count; i ++) {
		//printf("i= %d\n", i);
		char ea[ETHER_ADDR_STR_LEN];
		if (i != (count - 1))
			websWrite(wp, "\"%s\",", ether_etoa((void *)&auth->ea[i], ea));
		else
			websWrite(wp, "\"%s\"", ether_etoa((void *)&auth->ea[i], ea));              
	}
	if (!count) {
		websWrite(wp, "\"\"");              
	}
		

	/* error/exit */
exit:
	if (auth) free(auth);
	return 0;
}

static bool find_ethaddr_in_list (void *ethaddr, struct maclist *list)
{
	int i = 0;
	
	assert(ethaddr);
	assert(list);
	
	for (i = 0; i < list->count; i ++) {
		if (!bcmp(ethaddr, (void *)&list->ea[i], ETHER_ADDR_LEN)) {
			return TRUE;
		}
	}

	return FALSE;
}

static int
ej_wl_auth_list(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *name=NULL;
	struct maclist *auth=NULL, *assoc=NULL, *authorized=NULL, *wme=NULL;
	int max_sta_count, maclist_size;
	int i=0;
	
	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	/* buffers and length */
	max_sta_count = MAX_STA_COUNT;
	maclist_size = sizeof(auth->count) + max_sta_count * sizeof(struct ether_addr);

	auth = malloc(maclist_size);
	assoc = malloc(maclist_size);
	authorized = malloc(maclist_size);
	wme = malloc(maclist_size);

	if (!auth || !assoc || !authorized || !wme)
		goto exit;

	/* query wl for authenticated sta list */
	strcpy((char*)auth, "authe_sta_list");
	if (wl_ioctl(name, WLC_GET_VAR, auth, maclist_size))
		goto exit;

	/* query wl for associated sta list */
	assoc->count = max_sta_count;
	if (wl_ioctl(name, WLC_GET_ASSOCLIST, assoc, maclist_size))
		goto exit;

	/* query wl for authorized sta list */
	strcpy((char*)authorized, "autho_sta_list");
	if (wl_ioctl(name, WLC_GET_VAR, authorized, maclist_size))
		goto exit;
#ifdef REPEATER
	;
#else
	/* query wl for WME sta list */
	strcpy((char*)wme, "wme_sta_list");
	if (wl_ioctl(name, WLC_GET_VAR, wme, maclist_size))
		goto exit;
#endif
	/* build authenticated/associated/authorized sta list */
	for (i = 0; i < auth->count; i ++) {
		char ea[ETHER_ADDR_STR_LEN];
		char *value;
		websWrite(wp, "<tr><td>%s</td>", ether_etoa((void *)&auth->ea[i], ea));

		value = (find_ethaddr_in_list ((void *)&auth->ea[i], assoc))? "Yes" : "No";
		websWrite(wp, "<td>%s</td>", value);

		value = (find_ethaddr_in_list ((void *)&auth->ea[i], authorized))? "Yes" : "No";
		websWrite(wp, "<td>%s</td>", value);
#ifdef REPEATER
		;
#else
		value = (find_ethaddr_in_list ((void *)&auth->ea[i], wme))? "Yes" : "No";
		websWrite(wp, "<td>%s</td>", value);
#endif
		websWrite(wp, "</tr>");
	}

	/* error/exit */
exit:
	if (auth) free(auth);
	if (assoc) free(assoc);
	if (authorized) free(authorized);
	if (wme) free(wme);
	
	return 0;
}

/*
 * Example: 
 * wan_proto=dhcp
 * <% nvram_invmatch("wan_proto", "dhcp", "disabled"); %> does not produce
 * <% nvram_invmatch("wan_proto", "static", "disabled"); %> produces "disabled"
 */
static int
ej_nvram_invmatch(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL, *invmatch=NULL, *output=NULL;

	if (ejArgs(argc, argv, "%s %s %s", &name, &invmatch, &output) < 3) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	assert(name);
	assert(invmatch);
	assert(output);
	
	if (nvram_invmatch(name, invmatch))
		return websWrite(wp, output);

	return 0;
}	

/*
 * Example: 
 * filter_maclist=00:12:34:56:78:00 00:87:65:43:21:00
 * <% nvram_list("filter_maclist", 1); %> produces "00:87:65:43:21:00"
 * <% nvram_list("filter_maclist", 100); %> produces ""
 */
/* KEO - extend nvram_list to meet our needs
 * Example: 
 * wan_pptp_auth=0 1 1 1
 * <% nvram_list("wan_pptp_auth", 0, "0", "checked"; %> produces "checked"
 * <% nvram_list("wan_pptp_auth", 0, "1", "checked"; %> produces ""
 * <% nvram_list("wan_pptp_auth", 1, "1", "1", "0"; %> produces "1"
 * <% nvram_list("wan_pptp_auth", 1, "0", "1", "0"; %> produces "0"
 */
static int
ej_nvram_list(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL, *test=NULL, *posRes=NULL, *negRes=NULL;
	int which = 0, err = 0;
	char word[256], *next=NULL;
    
	switch(argc)
	{
	    case 5:
		if (ejArgs(argc, argv, "%s %d %s %s %s", 
			&name, &which, &test, &posRes, &negRes ) < 5) 
		    err = 1;
		break;
	    case 4:
		if (ejArgs(argc, argv, "%s %d %s %s", 
			&name, &which, &test, &posRes) < 4) 
		    err = 1;
		break;
	    case 2:
		if (ejArgs(argc, argv, "%s %d",	&name, &which) < 2) 
		    err = 1;
		break;
	}

	if (err) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	foreach(word, nvram_safe_get(name), next) {
	    if (which-- == 0)
	    {
	        if (argc ==  2)
		    return websWrite(wp, word);
		else if (!strcmp(word,test))
		    return websWrite(wp, posRes);
		else if (argc == 5)
		    return websWrite(wp, negRes);
	    }
	}
	return 0;
}

static int
ej_nvram_inlist(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL, *item=NULL, *output=NULL;
	char word[256], *next=NULL;

	if (ejArgs(argc, argv, "%s %s %s", &name, &item, &output) < 3) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	foreach(word, nvram_safe_get(name), next) {
		if (!strcmp(word, item))
			return websWrite(wp, output);
	}

	return 0;
}

static int
ej_nvram_invinlist(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL, *item=NULL, *output=NULL;
	char word[256], *next=NULL;

	if (ejArgs(argc, argv, "%s %s %s", &name, &item, &output) < 3) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	foreach(word, nvram_safe_get(name), next) {
		if (!strcmp(word, item))
			return 0;
	}

	return websWrite(wp, output);
}


/*USR - display MAC lists */
static int
ej_lan_list(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL;
	int ret = 0;
	char word[256], *next=NULL;

	if (ejArgs(argc, argv, "%s", &name) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	foreach(word, nvram_safe_get(name), next) {
		ret += websWrite(wp, "<tr>");
		ret += websWrite(wp, "<td><class=\"clsEntry\">LAN</td>");
		ret += websWrite(wp, "<td></td>");
		ret += websWrite(wp, "<td>%s</td>", word);
		ret += websWrite(wp, "</tr>");
	}
	
	return 0;
}

static int
ej_wireless_list(int eid, webs_t wp, int argc, char_t **argv)
{
	char *name=NULL;
	int ret = 0;
	char word[256], *next=NULL;

	if (ejArgs(argc, argv, "%s", &name) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	foreach(word, nvram_safe_get(name), next) {
		ret += websWrite(wp, "<tr>");
		ret += websWrite(wp, "<td><class=\"clsEntry\">Wireless</td>");
		ret += websWrite(wp, "<td></td>");
		ret += websWrite(wp, "<td>%s</td>", word);
		ret += websWrite(wp, "</tr>");
	}
	
	return 0;
}

#ifdef __CONFIG_NAT__
#ifdef ROUTER
/*
 * Example: 
 * <% filter_client(1, 10); %> produces a table of the first 10 client filter entries
 
static int
ej_filter_client(int eid, webs_t wp, int argc, char_t **argv)
{
	int i, n, j, ret = 0;
	netconf_filter_t start, end;
	bool valid;
	char port[] = "XXXXX";
	char *days[] = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };
	char *hours[] = {
		"12:00 AM", "1:00 AM", "2:00 AM", "3:00 AM", "4:00 AM", "5:00 AM",
		"6:00 AM", "7:00 AM", "8:00 AM", "9:00 AM", "10:00 AM", "11:00 AM",
		"12:00 PM", "1:00 PM", "2:00 PM", "3:00 PM", "4:00 PM", "5:00 PM",
		"6:00 PM", "7:00 PM", "8:00 PM", "9:00 PM", "10:00 PM", "11:00 PM"
	};

	if (ejArgs(argc, argv, "%d %d", &i, &n) < 2) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	for (; i <= n; i++) {
		valid = get_filter_client(i, &start, &end);

		ret += websWrite(wp, "<tr>");
		//ret += websWrite(wp, "<td></td>");

		// Print address range 
		ret += websWrite(wp, "<td><input name=\"filter_client_from_start%d\" value=\"%s\" size=\"13\" maxlength=\"15\"></td>",
				 i, valid ? inet_ntoa(start.match.src.ipaddr) : "");
		ret += websWrite(wp, "<td>-</td>");
		ret += websWrite(wp, "<td><input name=\"filter_client_from_end%d\" value=\"%s\" size=\"13\" maxlength=\"15\"></td>",
				 i, valid ? inet_ntoa(end.match.src.ipaddr) : "");
		//ret += websWrite(wp, "<td></td>");

		// Print protocol 
		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"filter_client_proto%d\">", i);
		ret += websWrite(wp, "<option value=\"tcp\" %s>TCP</option>",
				 valid && start.match.ipproto == IPPROTO_TCP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"udp\" %s>UDP</option>",
				 valid && start.match.ipproto == IPPROTO_UDP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"tcp/udp\" %s>TCP/UDP</option>",
				 valid && start.match.ipproto == IPPROTO_TCP ? "selected" : "");
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		//ret += websWrite(wp, "<td></td>");

		// Print port range 
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(start.match.dst.ports[0]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"filter_client_to_start%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		ret += websWrite(wp, "<td>-</td>");
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(start.match.dst.ports[1]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"filter_client_to_end%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		//ret += websWrite(wp, "<td></td>");

		// Print day range 
		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"filter_client_from_day%d\">", i);
		for (j = 0; j < ARRAYSIZE(days); j++)
			ret += websWrite(wp, "<option value=\"%d\" %s>%s</option>",
					 j, valid && start.match.days[0] == j ? "selected" : "", days[j]);
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		ret += websWrite(wp, "<td>-</td>");
		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"filter_client_to_day%d\">", i);
		for (j = 0; j < ARRAYSIZE(days); j++)
			ret += websWrite(wp, "<option value=\"%d\" %s>%s</option>",
					 j, valid && start.match.days[1] == j ? "selected" : "", days[j]);
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		//ret += websWrite(wp, "<td></td>");

		//Print time range 
		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"filter_client_from_sec%d\">", i);
		for (j = 0; j < ARRAYSIZE(hours); j++)
			ret += websWrite(wp, "<option value=\"%d\" %s>%s</option>",
					 j * 3600, valid && start.match.secs[0] == (j * 3600) ? "selected" : "", hours[j]);
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		ret += websWrite(wp, "<td>-</td>");

		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"filter_client_to_sec%d\">", i);
		for (j = 0; j < ARRAYSIZE(hours); j++)
			ret += websWrite(wp, "<option value=\"%d\" %s>%s</option>",
					 j * 3600, valid && start.match.secs[1] == (j * 3600) ? "selected" : "", hours[j]);
		// Special case for 11:59:59 PM 
		ret += websWrite(wp, "<option value=\"%d\" %s>12:00 AM</option>",
				 24 * 3600 - 1, valid && start.match.secs[1] == (24 * 3600 - 1) ? "selected" : "");
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		//ret += websWrite(wp, "<td></td>");

		// Print enable 
		ret += websWrite(wp, "<td><input type=\"checkbox\" name=\"filter_client_enable%d\" %s></td>",
				 i, valid && !(start.match.flags & NETCONF_DISABLED) ? "checked" : "");

		ret += websWrite(wp, "</tr>");
	}

	return ret;
}
*/
static int
ej_usr_filter_client(int eid, webs_t wp, int argc, char_t **argv)
{
	int i, ret = 0;
	netconf_filter_t start, end;
	bool valid;
	char port0[] = "XXXXX";
	char port1[] = "XXXXX";
	
	
	if (ejArgs(argc, argv, "%d", &i) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}
		
	valid = get_filter_client(i, &start, &end);
	
	/* Print port range */
	if (valid) {
		snprintf(port0, sizeof(port0), "%d", ntohs(start.match.dst.ports[0]));
		snprintf(port1, sizeof(port1), "%d", ntohs(start.match.dst.ports[1]));
	}
	else {
		*port0 = '\0';
		*port1 = '\0';
	}
		
	/* Print address range */
	ret += websWrite(wp, "%s", valid ? inet_ntoa(start.match.src.ipaddr) : "");
	ret += websWrite(wp, "-");
	ret += websWrite(wp, "%s", valid ? inet_ntoa(end.match.src.ipaddr) : "");
	ret += websWrite(wp, ":");
	/* Print port range */
	ret += websWrite(wp, "%s-%s", port0, port1);
	ret += websWrite(wp, ",");
	/* Print protocol*/
	if (valid && (start.match.ipproto == IPPROTO_TCP)) {
		ret += websWrite(wp, "%s", "tcp");
	}
	else if (valid && (start.match.ipproto == IPPROTO_UDP)) {
		ret += websWrite(wp, "%s", "udp");
	}
	else {
	//	printf("");
		ret += websWrite(wp, "%s", "");
	}
	//printf(",");
	ret += websWrite(wp, ",");
	/* Print day range */
	ret += websWrite(wp, "%d-%d", start.match.days[0], start.match.days[1]);
	ret += websWrite(wp, ",");
	/* Print time range */
	ret += websWrite(wp, "%d-%d", start.match.secs[0], start.match.secs[1]);
	ret += websWrite(wp, ",");
	/* Print enable */
	ret += websWrite(wp, "%s", valid && !(start.match.flags & NETCONF_DISABLED) ? "on" : "off");
	
	return ret;
}

/*
 * Example: 
 * <% forward_port(1, 10); %> produces a table of the first 10 port forward entries
 
static int
ej_forward_port(int eid, webs_t wp, int argc, char_t **argv)
{
	int i, n, ret = 0;
	netconf_nat_t nat;
	bool valid;
	char port[] = "XXXXX";

	if (ejArgs(argc, argv, "%d %d", &i, &n) < 2) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	for (; i <= n; i++) {
		valid = get_forward_port(i, &nat);

		ret += websWrite(wp, "<tr>");
		//ret += websWrite(wp, "<td></td>");

		// Print protocol 
		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"forward_port_proto%d\">", i);
		ret += websWrite(wp, "<option value=\"tcp\" %s>TCP</option>",
				 valid && nat.match.ipproto == IPPROTO_TCP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"udp\" %s>UDP</option>",
				 valid && nat.match.ipproto == IPPROTO_UDP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"tcp/udp\" %s>TCP/UDP</option>",
				 valid && nat.match.ipproto == IPPROTO_TCP ? "selected" : "");
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		//ret += websWrite(wp, "<td></td>");

		// Print WAN destination port range 
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(nat.match.dst.ports[0]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"forward_port_from_start%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		ret += websWrite(wp, "<td>-</td>");
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(nat.match.dst.ports[1]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"forward_port_from_end%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		//ret += websWrite(wp, "<td>></td>");

		// Print address range
		ret += websWrite(wp, "<td><input name=\"forward_port_to_ip%d\" value=\"%s\" size=\"15\" maxlength=\"15\"></td>",
				 i, valid ? inet_ntoa(nat.ipaddr) : "");
		ret += websWrite(wp, "<td>:</td>");

		// Print LAN destination port range 
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(nat.ports[0]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"forward_port_to_start%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		ret += websWrite(wp, "<td>-</td>");
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(nat.ports[1]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"forward_port_to_end%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		//ret += websWrite(wp, "<td></td>");

		// Print enable 
		ret += websWrite(wp, "<td><input type=\"checkbox\" name=\"forward_port_enable%d\" %s></td>",
				 i, valid && !(nat.match.flags & NETCONF_DISABLED) ? "checked" : "");

		ret += websWrite(wp, "</tr>");
	}

	return ret;
}
*/
static int
ej_usr_forward_port(int eid, webs_t wp, int argc, char_t **argv)
{
	int i, ret = 0;
	netconf_nat_t nat;
	bool valid;
	char port0[] = "XXXXX";
	char port1[] = "XXXXX";
	char port2[] = "XXXXX";
	char port3[] = "XXXXX";

	if (ejArgs(argc, argv, "%d", &i) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}
	
	valid = get_forward_port(i, &nat);
	
	/* Save destination port range */
	if (valid) {
		snprintf(port0, sizeof(port0), "%d", ntohs(nat.match.dst.ports[0]));
		snprintf(port1, sizeof(port1), "%d", ntohs(nat.match.dst.ports[1]));
		snprintf(port2, sizeof(port2), "%d", ntohs(nat.ports[0]));
		snprintf(port3, sizeof(port3), "%d", ntohs(nat.ports[1]));
	}
	else {
		*port0 = '\0';
		*port1 = '\0';
		*port2 = '\0';
		*port3 = '\0';
	}
	
	/* Print WAN destination port range */
	ret += websWrite(wp, "%s-%s", port0, port1);
	ret += websWrite(wp, ">");
	/* Print ip address */
	ret += websWrite(wp, "%s", valid ? inet_ntoa(nat.ipaddr) : "");
	ret += websWrite(wp, ":");
	/* Print LAN destination port range */
	ret += websWrite(wp, "%s-%s", port2, port3);
	ret += websWrite(wp, ",");
	/* Print protocol*/
	if (valid && (nat.match.ipproto == IPPROTO_TCP)) 
		ret += websWrite(wp, "%s", "tcp");
	else if (valid && (nat.match.ipproto == IPPROTO_UDP))
		ret += websWrite(wp, "%s", "udp");
	else
		ret += websWrite(wp, "%s", "");
	ret += websWrite(wp, ",");
	/* Print enable */
	ret += websWrite(wp, "%s", valid && !(nat.match.flags & NETCONF_DISABLED) ? "on" : "off");
	
	return ret;
}
/*
static int
ej_autofw_port(int eid, webs_t wp, int argc, char_t **argv)
{
	int i, n, ret = 0;
	netconf_app_t app;
	bool valid;
	char port[] = "XXXXX";

	if (ejArgs(argc, argv, "%d %d", &i, &n) < 2) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	for (; i <= n; i++) {
		valid = get_autofw_port(i, &app);

		// Parse out_proto:out_port,in_proto:in_start-in_end>to_start-to_end,enable,desc 
		ret += websWrite(wp, "<tr>");
		//ret += websWrite(wp, "<td></td>");

		// Print outbound protocol 
		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"autofw_port_out_proto%d\">", i);
		ret += websWrite(wp, "<option value=\"tcp\" %s>TCP</option>",
				 valid && app.match.ipproto == IPPROTO_TCP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"udp\" %s>UDP</option>",
				 valid && app.match.ipproto == IPPROTO_UDP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"tcp/udp\" %s>TCP/UDP</option>",
				 valid && app.match.ipproto == IPPROTO_TCP ? "selected" : "");
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		//ret += websWrite(wp, "<td></td>");

		// Print outbound port 
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(app.match.dst.ports[0]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"autofw_port_out_start%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		ret += websWrite(wp, "<td>-</td>");
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(app.match.dst.ports[1]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"autofw_port_out_end%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		//ret += websWrite(wp, "<td></td>");

		// Print related protocol 
		ret += websWrite(wp, "<td>");
		ret += websWrite(wp, "<select name=\"autofw_port_in_proto%d\">", i);
		ret += websWrite(wp, "<option value=\"tcp\" %s>TCP</option>",
				 valid && app.proto == IPPROTO_TCP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"udp\" %s>UDP</option>",
				 valid && app.proto == IPPROTO_UDP ? "selected" : "");
		ret += websWrite(wp, "<option value=\"tcp/udp\" %s>TCP/UDP</option>",
				 valid && app.proto == IPPROTO_TCP ? "selected" : "");		 
		ret += websWrite(wp, "</select>");
		ret += websWrite(wp, "</td>");
		//ret += websWrite(wp, "<td></td>");

		// Print related destination port range 
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(app.dport[0]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"autofw_port_in_start%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		ret += websWrite(wp, "<td>-</td>");
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(app.dport[1]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"autofw_port_in_end%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		//ret += websWrite(wp, "<td></td>");

		// Print mapped destination port range  
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(app.to[0]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"autofw_port_to_start%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		ret += websWrite(wp, "<td>-</td>");
		if (valid)
			snprintf(port, sizeof(port), "%d", ntohs(app.to[1]));
		else
			*port = '\0';
		ret += websWrite(wp, "<td><input name=\"autofw_port_to_end%d\" value=\"%s\" size=\"5\" maxlength=\"5\"></td>",
				 i, port);
		//ret += websWrite(wp, "<td></td>"); 

		// Print enable 
		ret += websWrite(wp, "<td><input type=\"checkbox\" name=\"autofw_port_enable%d\" %s></td>",
				 i, valid && !(app.match.flags & NETCONF_DISABLED) ? "checked" : "");

		ret += websWrite(wp, "</tr>");
	}

	return ret;
}
*/
static int
ej_usr_autofw_port(int eid, webs_t wp, int argc, char_t **argv)
{
	int i, ret = 0;
	netconf_app_t app;
	bool valid;
	char port0[] = "XXXXX";
	char port1[] = "XXXXX";
	char port2[] = "XXXXX";
	char port3[] = "XXXXX";
	char port4[] = "XXXXX";
	char port5[] = "XXXXX";

	if (ejArgs(argc, argv, "%d", &i) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	valid = get_autofw_port(i, &app);
		
	/* Save destination port range */
	if (valid) {
		snprintf(port0, sizeof(port0), "%d", ntohs(app.match.dst.ports[0]));
		snprintf(port1, sizeof(port1), "%d", ntohs(app.match.dst.ports[1]));
		snprintf(port2, sizeof(port2), "%d", ntohs(app.dport[0]));
		snprintf(port3, sizeof(port3), "%d", ntohs(app.dport[1]));
		snprintf(port4, sizeof(port4), "%d", ntohs(app.to[0]));
		snprintf(port5, sizeof(port5), "%d", ntohs(app.to[1]));
	}
	else {
		*port0 = '\0';
		*port1 = '\0';
		*port2 = '\0';
		*port3 = '\0';
		*port4 = '\0';
		*port5 = '\0';
	}
	
	/* Print outbound protocol */
	if (valid && (app.match.ipproto == IPPROTO_TCP))
		ret += websWrite(wp, "%s", "tcp" );
	else if (valid && (app.match.ipproto == IPPROTO_UDP))
		ret += websWrite(wp, "%s", "udp" );
	else
		ret += websWrite(wp, "%s", "" );
	ret += websWrite(wp, ":"); 
	/* Print outbound port */
	ret += websWrite(wp, "%s-%s", port0, port1);
	ret += websWrite(wp, ",");
	/* Print related protocol */
	if (valid && (app.proto == IPPROTO_TCP))
		ret += websWrite(wp, "%s", "tcp");
	else if (valid && (app.proto == IPPROTO_UDP))
		ret += websWrite(wp, "%s", "udp");
	else
		ret += websWrite(wp, "%s", "");
	ret += websWrite(wp, ":"); 
	/* Print related destination port */
	ret += websWrite(wp, "%s-%s", port2, port3);
	ret += websWrite(wp, ">");
	/* Print mapped destination port range */ 
	ret += websWrite(wp, "%s-%s", port4, port5);
	ret += websWrite(wp, ",");
	/* Print enable */
	ret += websWrite(wp, "%s", valid && !(app.match.flags & NETCONF_DISABLED) ? "on" : "off");
	
	return ret;
}	
#endif   /* ROUTER */
#endif  /* __CONFIG_NAT__ */
/*
 * Example: 
 * lan_route=192.168.2.0:255.255.255.0:192.168.2.1:1
 * <% lan_route("ipaddr", 0); %> produces "192.168.2.0"
 */
static int
ej_lan_route(int eid, webs_t wp, int argc, char_t **argv)
{
	char *arg;
	int which;
	char word[256], *next;
	char *ipaddr, *netmask, *gateway, *metric;

	if (ejArgs(argc, argv, "%s %d", &arg, &which) < 2) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	foreach(word, nvram_safe_get("lan_route"), next) {
		if (which-- == 0) {
			netmask = word;
			ipaddr = strsep(&netmask, ":");
			if (!ipaddr || !netmask)
				continue;
			gateway = netmask;
			netmask = strsep(&gateway, ":");
			if (!netmask || !gateway)
				continue;
			metric = gateway;
			gateway = strsep(&metric, ":");
			if (!gateway || !metric)
				continue;
			if (!strcmp(arg, "ipaddr"))
				return websWrite(wp, ipaddr);
			else if (!strcmp(arg, "netmask"))
				return websWrite(wp, netmask);
			else if (!strcmp(arg, "gateway"))
				return websWrite(wp, gateway);
			else if (!strcmp(arg, "metric"))
				return websWrite(wp, metric);
		}
	}

	return 0;
}

#ifdef __CONFIG_NAT__
#ifdef ROUTER
/*
 * Example: 
 * wan_route=192.168.10.0:255.255.255.0:192.168.10.1:1
 * <% wan_route("ipaddr", 0); %> produces "192.168.10.0"
 */
static int
ej_wan_route(int eid, webs_t wp, int argc, char_t **argv)
{
	char *arg;
	int which;
	char word[256], *next;
	char *ipaddr, *netmask, *gateway, *metric;
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";


	if (ejArgs(argc, argv, "%s %d", &arg, &which) < 2) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	if ((unit = atoi(nvram_safe_get("wan_unit"))) < 0)
		unit = 0;
	wan_prefix(unit, prefix);
	
	foreach(word, nvram_safe_get(strcat_r(prefix, "route", tmp)), next) {
		if (which-- == 0) {
			netmask = word;
			ipaddr = strsep(&netmask, ":");
			if (!ipaddr || !netmask)
				continue;
			gateway = netmask;
			netmask = strsep(&gateway, ":");
			if (!netmask || !gateway)
				continue;
			metric = gateway;
			gateway = strsep(&metric, ":");
			if (!gateway || !metric)
				continue;
			if (!strcmp(arg, "ipaddr"))
				return websWrite(wp, ipaddr);
			else if (!strcmp(arg, "netmask"))
				return websWrite(wp, netmask);
			else if (!strcmp(arg, "gateway"))
				return websWrite(wp, gateway);
			else if (!strcmp(arg, "metric"))
				return websWrite(wp, metric);
		}
	}

	return 0;
}
#endif  /* ROUTER */
#endif	/* __CONFIG_NAT__ */

/* Return a list of the currently present wireless interfaces 
static int
ej_wl_list(int eid, webs_t wp, int argc, char_t **argv)
{
	char name[IFNAMSIZ], *next;
	int unit, ret = 0;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *hwaddr, *ssid;
	char ifnames[256];

	snprintf(ifnames, sizeof(ifnames), "%s %s", 
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		// Probe for wl interfaces 
		if (wl_probe(name) ||
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)))
			continue;

		// Get configured SSID 
		snprintf(prefix, sizeof(prefix), "wl%d_", unit);
		hwaddr = nvram_get(strcat_r(prefix, "hwaddr", tmp));
		ssid = nvram_get(strcat_r(prefix, "ssid", tmp));
		if (!hwaddr || !*hwaddr || !ssid || !*ssid)
			continue;

		ret += websWrite(wp, "<option value=\"%d\" %s>%s (%s)</option>", unit,
				 unit == atoi(nvram_safe_get("wl_unit")) ? "selected" : "",
				 ssid, hwaddr);
	}

	if (!ret)
		ret += websWrite(wp, "<option value=\"-1\" selected>None</option>");

	return ret;
}
*/
/* Return current wireless channel */
static int
ej_wl_cur_channel(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char *name=NULL;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	channel_info_t ci;

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	wl_ioctl(name, WLC_GET_CHANNEL, &ci, sizeof(ci));
	return websWrite(wp, "%d", ci.target_channel);
}
#ifdef CLIENT
// Return a list of the supported bands on the currently selected wireless interface 
static int
ej_wl_phytypes(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, ret = 0;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *phytype=NULL;
	char *phylist=NULL;
	int i = 0;

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return websWrite(wp, "None");

	// Get available phy types of the currently selected wireless interface 
	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	phylist = nvram_safe_get(strcat_r(prefix, "phytypes", tmp));

	// Get configured phy type 
	phytype = nvram_safe_get("wl_phytype");

	for (i = 0; i < strlen(phylist); i++) {
		ret += websWrite(wp, "<option value=\"%c\" %s>802.11%c (%s GHz)</option>",
				 phylist[i], phylist[i] == *phytype ? "selected" : "", phylist[i], 
				 phylist[i] == 'a' ? "5" : "2.4");
	}

	return ret;
}

// Return a radio ID given a phy type
static int
ej_wl_radioid(int eid, webs_t wp, int argc, char_t **argv)
{
	char *phytype=NULL, var[NVRAM_BUFSIZE], *next;
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	int which;

	if (ejArgs(argc, argv, "%s", &phytype) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}
	
	assert(phytype);
	
	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return websWrite(wp, "None");
	snprintf(prefix, sizeof(prefix), "wl%d_", unit);

	which = strcspn(nvram_safe_get(strcat_r(prefix, "phytypes", tmp)), phytype);
	foreach(var, nvram_safe_get(strcat_r(prefix, "radioids", tmp)), next) {
		if (which == 0)
			return websWrite(wp, var);
		which--;
	}

	return websWrite(wp, "None");
}

// Return current core revision 
static int
ej_wl_corerev(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return websWrite(wp, "None");
	snprintf(prefix, sizeof(prefix), "wl%d_", unit);

	return websWrite(wp, nvram_safe_get(strcat_r(prefix, "corerev", tmp)));
}

/* USR  Return current country */
static int
ej_wl_cur_country(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char *name=NULL;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char buf[WLC_CNTRY_BUF_SZ];

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	wl_ioctl(name, WLC_GET_COUNTRY, buf, sizeof(buf));
	if (!strcmp(buf,"US"))
		strncpy(buf,"FCC", 4);
	if (!strcmp(buf,"FR")) 
		strncpy(buf,"ETSI",5);
	return websWrite(wp, "%s", buf);  
}
#endif
/*USR Return current Wireless Mode */
static int
ej_wl_cur_mode(int eid, webs_t wp, int argc, char_t **argv)
{
	char *value;
	int tmp1, tmp2;
		
	value = nvram_safe_get("wl_mode");
	tmp1 = atoi(nvram_safe_get("router_disable"));
	tmp2 = atoi(nvram_safe_get("wl_infra"));
        if ((!strcmp(value,"wet")) && (tmp1 == 1) && (tmp2 == 1))
	#ifdef CLIENT	
		value="Client";
	#else
		value="Infrastructure";
	#endif
	else if ((!strcmp(value,"wet")) && (tmp1 == 1) && (tmp2 == 0))
	#ifdef CLIENT	
		value="Ad Hoc";
	#else
		value="Ad hoc";
	#endif
	else if (!strcmp(value,"wds"))
		value="WDS Bridge";
	else if (!strcmp(value,"wet"))
		value="Configuration Mode";
	else if (!strcmp(value,"ap"))
	#ifdef REPEATER
		value="Range Extender";
	#else
		value="Access Point";
	#endif
	
	return websWrite(wp, "%s", value);
}

#ifdef CLIENT
/*USR Return SSID of associated network */
static int
ej_wl_cur_SSID(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];

	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 && unit == unit2) {
			wlc_ssid_t ssid;
			wl_ioctl(name, WLC_GET_SSID, &ssid, sizeof(ssid));
			return websWrite(wp, "%s", ssid.SSID);
		}
	}
	return websWrite(wp, "");
}

/*USR Return BSSID of associated network */
static int
ej_wl_cur_BSSID(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];

	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
			unsigned char buf[ETHER_ADDR_LEN];
		        char eaddr[32];
			if (wl_ioctl(name, WLC_GET_BSSID, buf, sizeof(buf)) == 0)
			{	if (strcmp(ether_etoa(buf,eaddr),"00:00:00:00:00:00"))
					return websWrite(wp, "%s", ether_etoa(buf,eaddr));
				else
					return websWrite(wp, "Device is not associated");
			}
		}
	}
	return websWrite(wp, "Device is not associated");
}

/* USR get signal strength */
static int
ej_wl_cur_signal(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];
	int rssi = 0;
	int percent = 0;
		
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
			if (wl_ioctl(name, WLC_GET_RSSI, &rssi, sizeof(rssi)) == 0)
			{	//printf("rssi: %d, \n", rssi);
				if (rssi < SIGNAL_MIN)
					rssi = SIGNAL_MIN;
				if (rssi > SIGNAL_MAX)
					rssi = SIGNAL_MAX;
				percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
				//printf("RSSI: %d, \n", rssi);
				return websWrite(wp, "%d%%", percent);
			}
		}
	}
	return websWrite(wp, "0%%");
}

/* USR get signal strength */
static int
ej_wl_cur_rate(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];
	int rate;
	char buf[WLC_IOCTL_MAXLEN];
	
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
			if (wl_ioctl(name, WLC_GET_RATE, &rate, sizeof(rate)) == 0)
			{
				//printf("rate in 500kbps: %d, \n", rate);
				if (rate >= 2) {
					rate = rate * 500/1000;
					//get afterburner value
					memset(buf, 0, WLC_IOCTL_MAXLEN);
					strcpy(buf, "afterburner");
					if (wl_ioctl(name, WLC_GET_VAR, buf, WLC_IOCTL_MAXLEN) < 0)
						printf("afterburner error!\n");
					else {	
						printf("afterbuner rate: %d\n", rate);
						int * tm = (int *)buf;
						printf("afterburner: %d\n", *tm);
						if ((rate == 54) && (*tm == 1)) 
							return websWrite(wp, "125 Mbps");
					}
					return websWrite(wp, "%d Mbps", rate);
				}
				else {
					rate = rate * 500;
					//printf("rate in kbps: %d, \n", rate);
					return websWrite(wp, "%d Kbps", rate);
				}
			}
		}
	}
	return websWrite(wp, "0 Mbps");
}

/* USR implementation of wl scan */
static int
ej_wl_scan(int eid, webs_t wp, int argc, char_t **argv)
{
	unsigned int i;
	char eaddr[40];
	uchar ssid[40];
	char *wep, *adhoc;
	int ret = 0;
	char schecked[10];
	char ifnames[256];
	char name[IFNAMSIZ];
	char *next;
	int unit, unit2;
	struct scan_results {
		uint32 buflen;
		uint32 version;
		uint32 count;
		unsigned buf[4096]; /*arbitrary- enough to hold 30+ bss info.IOCTL buffer's size is limited to  WLC_IOCTL_MAXLEN(8192)*/
	}scan_results;
	wl_bss_info_t *pbss_info;
	int channel = 0;
	int rssi = 0;
	int percent = 0;
	
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
                return -1;

        snprintf(ifnames, sizeof(ifnames), "%s %s",
                nvram_safe_get("lan_ifnames"),
                nvram_safe_get("wan_ifnames"));
        foreach(name, ifnames, next) {
                if (wl_probe(name) == 0 &&
                    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 && unit == unit2)
			break;
	}

	if(*name == 0) {//shoudn't happen
		return  websWrite(wp,"Found No Wireless Interface.");
	}
	usleep(1000000);
	//printf("Client defined! \n");
	if(!wl_ioctl(name, WLC_SCAN, &channel, sizeof(channel))){
		usleep(3000000);
		memset((char *)&scan_results, 0, sizeof(scan_results));
		scan_results.buflen = sizeof(scan_results);
		scan_results.version = WL_BSS_INFO_VERSION;
		if((! wl_ioctl(name, WLC_SCAN_RESULTS, &scan_results, sizeof(scan_results))) ||
			(scan_results.count != 0)) {
			//printf("scan_results.count: %d, \n", scan_results.count);
			pbss_info = (wl_bss_info_t *)scan_results.buf;
			for (i = 0; i < scan_results.count; i++) {
				if(i == 0)
	   				strcpy(schecked,"checked"); //set the first SSSID as default
				else
		   			schecked[0] = '\0';
				memset(ssid,0,sizeof(ssid));
			 	strncpy(ssid,(char *)pbss_info->SSID,pbss_info->SSID_len);
				ether_etoa(pbss_info->BSSID.octet,eaddr);
				rssi = pbss_info->RSSI;
				if (rssi < SIGNAL_MIN)
					rssi = SIGNAL_MIN;
				if (rssi > SIGNAL_MAX)
					rssi = SIGNAL_MAX;
				percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
				//printf("RSSI: %d, \n", rssi);
				channel = pbss_info->channel;
				if(pbss_info->capability & DOT11_CAP_PRIVACY)
					wep = "Enabled";
				else
					wep = "Disabled";
				
				if(pbss_info->capability & DOT11_CAP_IBSS)
					adhoc = "Yes";
				else
					adhoc = "No";
						
				ret += websWrite(wp, "<tr><td><input type=\"radio\" name=\"ap_sel\" value=\"%s %s %s %d %d\" %s></td>", name,ssid,eaddr,percent,channel,schecked);
				ret += websWrite(wp, "<td>");
				ret += websWrite(wp, ssid);
		       		ret += websWrite(wp, "</td><td>");
				ret += websWrite(wp, eaddr);
				ret += websWrite(wp, "</td><td>");
				ret += websWrite(wp, wep);
				ret += websWrite(wp, "</td><td>");
				ret += websWrite(wp, adhoc);
				ret += websWrite(wp, "</td><td>");
				ret += websWrite(wp, "  %d", channel);
				ret += websWrite(wp, "</td><td>");
				ret += websWrite(wp, "  %d%%", percent);
				ret += websWrite(wp, "</td></tr>");
				pbss_info = (wl_bss_info_t *)(((char *) pbss_info) + pbss_info->length);
  			}
		}
		else{
			ret += websWrite(wp,"<tr><td colspan=\"3\"> Found no AP! Click Cancel to exit the scan process.</td></tr>");
		}
	}
	else{
		ret += websWrite(wp,"<tr><td colspan=\"3\"> Scan process failed.</td></tr>");
	}
	return ret;
}
#endif

#ifdef REPEATER
static int
ej_wl_scan(int eid, webs_t wp, int argc, char_t **argv)
{
	unsigned int i;
	char eaddr[40];
	uchar ssid[40];
	char *wep;
	int ret = 0;
	char ifnames[256];
	char name[IFNAMSIZ];
	char *next;
	int unit, unit2;
	struct scan_results {
		uint32 buflen;
		uint32 version;
		uint32 count;
		unsigned buf[4096]; /*arbitrary- enough to hold 30+ bss info.IOCTL buffer's size is limited to  WLC_IOCTL_MAXLEN(8192)*/
	}scan_results;
	wl_bss_info_t *pbss_info;
	int channel = 0;
	int rssi = 0;
	int percent = 0;
	wlc_ssid_t ssid_temp;
	brcm_ie_t *ptr_ie;
	uint32 ie_len;
	uint8 len, id, flags;
	uchar oui[10];
	char *wds;
	static char temp[] = {0x0, 0x90, 0x4C};
		
	#define BRF_LZYWDS 0x04
	
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
                return -1;

        snprintf(ifnames, sizeof(ifnames), "%s %s",
                nvram_safe_get("lan_ifnames"),
                nvram_safe_get("wan_ifnames"));
        foreach(name, ifnames, next) {
                if (wl_probe(name) == 0 &&
                    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 && unit == unit2)
			break;
	}

	if(*name == 0) {//shoudn't happen
		cprintf("Found No Wireless Interface.");
		return 0;
	}
	
	/* to be able to start the scanning process we have to kill SSID*/
	/* Nuke SSID  */
	//printf("Repeater defined! \n");
	ssid_temp.SSID_len = 0;
	ssid_temp.SSID[0] = '\0';
	//cprintf("name_scan: %s \n", name);
	if (wl_ioctl(name, WLC_SET_SSID, &ssid_temp, sizeof(ssid_temp))) {
		cprintf("SET SSID failed!\n");
		cprintf("ssid: %s\n", ssid_temp.SSID);
	}
	wl_ioctl(name, WLC_DOWN, NULL, 0);
	wl_ioctl(name, WLC_UP, NULL, 0);
	usleep(2000000);
	//cprintf("start scan!\n");
	if (wl_ioctl(name, WLC_SCAN, &channel, sizeof(channel)))
	{
		cprintf("Scan failed!\n");
		//restore SSID
		ssid_temp.SSID_len = strlen(nvram_safe_get("wl0_ssid"));
		strncpy(ssid_temp.SSID, nvram_safe_get("wl0_ssid"), ssid_temp.SSID_len);
		wl_ioctl(name, WLC_SET_SSID, &ssid_temp, sizeof(ssid_temp));
		return ret;	
	}
	usleep(1000000);
	//cprintf("wait 1 sec\n");
	memset((char *)&scan_results, 0, sizeof(scan_results));
	scan_results.buflen = sizeof(scan_results);
	scan_results.version = WL_BSS_INFO_VERSION;
	if((! wl_ioctl(name, WLC_SCAN_RESULTS, &scan_results, sizeof(scan_results))) ||
		(scan_results.count != 0)) {
		printf("scan_results.count: %d, \n", scan_results.count);
		pbss_info = (wl_bss_info_t *)scan_results.buf;
		for (i = 0; i < scan_results.count; i++) {
			memset(ssid,0,sizeof(ssid));
		 	strncpy(ssid,(char *)pbss_info->SSID,pbss_info->SSID_len);
			ether_etoa(pbss_info->BSSID.octet,eaddr);
			rssi = pbss_info->RSSI;
			if (rssi < SIGNAL_MIN)
				rssi = SIGNAL_MIN;
			if (rssi > SIGNAL_MAX)
				rssi = SIGNAL_MAX;
			percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
			//cprintf("RSSI: %d, \n", rssi);
			channel = pbss_info->channel;
			//cprintf("channel: %d, \n", channel);
			if(pbss_info->capability & DOT11_CAP_PRIVACY)
				wep = "true";
			else
				wep = "false";
		        //cprintf("wep: %s, \n", wep);
			ie_len = pbss_info->ie_length;
			//cprintf("ie_len: %d \n", ie_len);
			ptr_ie = (brcm_ie_t *)(((char *) pbss_info) + sizeof(wl_bss_info_t));
			wds = "false";
			while (ie_len > 0) {
				id = ptr_ie->id;
				//cprintf("id: %d ", id);
				len = ptr_ie->len; 
				//cprintf("length: %d ", len);
				if (id == 221) {
					memset(oui,0,10);
					strncpy(oui,(char *)ptr_ie->oui,3);
					if (!strncmp(oui,temp,3)) {
						flags = ptr_ie->flags;
						//cprintf("flags: %d\n", flags);
						if (flags & BRF_LZYWDS)
						{
							//cprintf("LAZYWDS = 1\n");
							wds = "true";
						}
						else
						{	//cprintf("LAZYWDS = 0\n");
							;//wds = "false";
							//break;
						}
					}
				}
				ie_len = ie_len - len - 2;
				ptr_ie = (brcm_ie_t *)(((char *)ptr_ie) + len + 2);
			}//while
			if (i < scan_results.count - 1) 					
				ret += websWrite(wp, "new Site(\"%s\", %s, \"%s\", %d, %d, %s),", ssid, wds, eaddr, channel, percent, wep);
			else
				ret += websWrite(wp, "new Site(\"%s\", %s, \"%s\", %d, %d, %s)", ssid, wds, eaddr, channel, percent, wep);
			pbss_info = (wl_bss_info_t *)(((char *) pbss_info) + pbss_info->length);
  		}
	}
	else{
		cprintf("Found no AP!");
	}
	
	//restore SSID
	ssid_temp.SSID_len = strlen(nvram_safe_get("wl0_ssid"));
	strncpy(ssid_temp.SSID, nvram_safe_get("wl0_ssid"), ssid_temp.SSID_len);
	wl_ioctl(name, WLC_SET_SSID, &ssid_temp, sizeof(ssid_temp));
	
	return ret;
}
#endif

#ifdef GAMING
/*USR Return BSSID of associated network */
static int
ej_wl_cur_BSSID(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];

	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
		    	unsigned char buf[ETHER_ADDR_LEN];
		        char eaddr[32];
			if (wl_ioctl(name, WLC_GET_BSSID, buf, sizeof(buf)) == 0)
			{				
				if (strcmp(ether_etoa(buf,eaddr),"00:00:00:00:00:00"))
					return websWrite(wp, "%s", ether_etoa(buf,eaddr));
				else
					return websWrite(wp, "No connection");
			}
		}
	}
	return websWrite(wp, "No connection");
}

/* USR get signal strength */
static int
ej_wl_cur_signal(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];
	int rssi = 0;
	int percent = 0;
		
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
		    	if (wl_ioctl(name, WLC_GET_RSSI, &rssi, sizeof(rssi)) == 0) {
				//printf("rssi: %d, \n", rssi);
				if (rssi < SIGNAL_MIN)
					rssi = SIGNAL_MIN;
				if (rssi > SIGNAL_MAX)
					rssi = SIGNAL_MAX;
				percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
				//printf("RSSI: %d, \n", rssi);
				return websWrite(wp, "%d%%", percent);
			}
		}
	}
	return websWrite(wp, "0%%");
}

/* USR get signal strength */
static int
ej_wl_cur_rate(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];
	int rate;
	char buf[WLC_IOCTL_MAXLEN];
	
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
			if (wl_ioctl(name, WLC_GET_RATE, &rate, sizeof(rate)) == 0)
			{
				//printf("rate in 500kbps: %d, \n", rate);
				if (rate >= 2) {
					rate = rate * 500/1000;
					//get afterburner value
					memset(buf, 0, WLC_IOCTL_MAXLEN);
					strcpy(buf, "afterburner");
					if (wl_ioctl(name, WLC_GET_VAR, buf, WLC_IOCTL_MAXLEN) < 0)
						printf("afterburner error!\n");
					else {	
						printf("afterbuner rate: %d\n", rate);
						int * tm = (int *)buf;
						printf("afterburner: %d\n", *tm);
						if ((rate == 54) && (*tm == 1)) 
							return websWrite(wp, "125 Mbps");
					}
					return websWrite(wp, "%d Mbps", rate);
				}
				else {
					rate = rate * 500;
					//printf("rate in kbps: %d, \n", rate);
					return websWrite(wp, "%d Kbps", rate);
				}
			}
		}
	}
	return websWrite(wp, "0 Mbps");
}

/* USR implementation of wl scan for WET*/
static int
ej_wl_scan(int eid, webs_t wp, int argc, char_t **argv)
{
	unsigned int i;
	char eaddr[40];
	uchar ssid[40];
	char *wep, *adhoc;
	int ret = 0;
	char ifnames[256];
	char name[IFNAMSIZ];
	char *next;
	int unit, unit2;
	struct scan_results {
		uint32 buflen;
		uint32 version;
		uint32 count;
		unsigned buf[4096]; /*arbitrary- enough to hold 30+ bss info.IOCTL buffer's size is limited to  WLC_IOCTL_MAXLEN(8192)*/
	}scan_results;
	wl_bss_info_t *pbss_info;
	int channel = 0;
	int rssi = 0;
	int percent = 0;
	
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
                return -1;

        snprintf(ifnames, sizeof(ifnames), "%s %s",
                nvram_safe_get("lan_ifnames"),
                nvram_safe_get("wan_ifnames"));
        foreach(name, ifnames, next) {
                if (wl_probe(name) == 0 &&
                    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 && unit == unit2)
			break;
	}

	if(*name == 0) {//shoudn't happen
		cprintf("Found No Wireless Interface.");
		return 0;
	}
	
	//printf("Client defined! \n");
	usleep(2000000);
	if(!wl_ioctl(name, WLC_SCAN, &channel, sizeof(channel))){
		usleep(1000000);
		memset((char *)&scan_results, 0, sizeof(scan_results));
		scan_results.buflen = sizeof(scan_results);
		scan_results.version = WL_BSS_INFO_VERSION;
		if((! wl_ioctl(name, WLC_SCAN_RESULTS, &scan_results, sizeof(scan_results))) ||
			(scan_results.count != 0)) {
			//printf("scan_results.count: %d, \n", scan_results.count);
			pbss_info = (wl_bss_info_t *)scan_results.buf;
			for (i = 0; i < scan_results.count; i++) {
				memset(ssid,0,sizeof(ssid));
			 	strncpy(ssid,(char *)pbss_info->SSID,pbss_info->SSID_len);
				ether_etoa(pbss_info->BSSID.octet,eaddr);
				rssi = pbss_info->RSSI;
				if (rssi < SIGNAL_MIN)
					rssi = SIGNAL_MIN;
				if (rssi > SIGNAL_MAX)
					rssi = SIGNAL_MAX;
				percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
				//printf("RSSI: %d, \n", rssi);
				channel = pbss_info->channel;
				//printf("RSSI: %d, \n", rssi);
				if(pbss_info->capability & DOT11_CAP_PRIVACY)
					wep = "true";
				else
					wep = "false";
			       	
				if(pbss_info->capability & DOT11_CAP_IBSS)
					adhoc = "true";
				else
					adhoc = "false";
				//cprintf("adhoc is %s\n", adhoc);	
				
				if (i < scan_results.count - 1) 					
					ret += websWrite(wp, "new Site(\"%s\", \"%s\", %d, %d, %s, %s),", ssid, eaddr, channel, percent, wep, adhoc);
				else
					ret += websWrite(wp, "new Site(\"%s\", \"%s\", %d, %d, %s, %s)", ssid, eaddr, channel, percent, wep, adhoc);
				pbss_info = (wl_bss_info_t *)(((char *) pbss_info) + pbss_info->length);
  			}
		}
		else{
			cprintf("Found no AP!");
		}
	}
	else{
		cprintf("Scan process failed!");
	}
	
	return ret;
}
#endif

#ifdef BRIDGE
/*USR Return BSSID of associated network */
static int
ej_wl_cur_BSSID(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];

	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
		    	unsigned char buf[ETHER_ADDR_LEN];
		        char eaddr[32];
			if (wl_ioctl(name, WLC_GET_BSSID, buf, sizeof(buf)) == 0)
			{				
				if (strcmp(ether_etoa(buf,eaddr),"00:00:00:00:00:00"))
					return websWrite(wp, "%s", ether_etoa(buf,eaddr));
				else
					return websWrite(wp, "No connection");
			}
		}
	}
	return websWrite(wp, "No connection");
}

/* USR get signal strength */
static int
ej_wl_cur_signal(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];
	int rssi = 0;
	int percent = 0;
		
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
		    	if (wl_ioctl(name, WLC_GET_RSSI, &rssi, sizeof(rssi)) == 0) {
				//printf("rssi: %d, \n", rssi);
				if (rssi < SIGNAL_MIN)
					rssi = SIGNAL_MIN;
				if (rssi > SIGNAL_MAX)
					rssi = SIGNAL_MAX;
				percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
				//printf("RSSI: %d, \n", rssi);
				return websWrite(wp, "%d%%", percent);
			}
		}
	}
	return websWrite(wp, "0%%");
}

/* USR get signal strength */
static int
ej_wl_cur_rate(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit, unit2;
	char name[IFNAMSIZ], *next;
	char ifnames[256];
	int rate;
	char buf[WLC_IOCTL_MAXLEN];
	
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {
			if (wl_ioctl(name, WLC_GET_RATE, &rate, sizeof(rate)) == 0)
			{
				//printf("rate in 500kbps: %d, \n", rate);
				if (rate >= 2) {
					rate = rate * 500/1000;
					//get afterburner value
					memset(buf, 0, WLC_IOCTL_MAXLEN);
					strcpy(buf, "afterburner");
					if (wl_ioctl(name, WLC_GET_VAR, buf, WLC_IOCTL_MAXLEN) < 0)
						printf("afterburner error!\n");
					else {	
						printf("afterbuner rate: %d\n", rate);
						int * tm = (int *)buf;
						printf("afterburner: %d\n", *tm);
						if ((rate == 54) && (*tm == 1)) 
							return websWrite(wp, "125 Mbps");
					}
					return websWrite(wp, "%d Mbps", rate);
				}
				else {
					rate = rate * 500;
					//printf("rate in kbps: %d, \n", rate);
					return websWrite(wp, "%d Kbps", rate);
				}
			}
		}
	}
	return websWrite(wp, "0 Mbps");
}

/* USR implementation of wl scan for WET*/
static int
ej_wl_scan(int eid, webs_t wp, int argc, char_t **argv)
{
	unsigned int i;
	char eaddr[40];
	uchar ssid[40];
	char *wep;
	int ret = 0;
	char ifnames[256];
	char name[IFNAMSIZ];
	char *next;
	int unit, unit2;
	struct scan_results {
		uint32 buflen;
		uint32 version;
		uint32 count;
		unsigned buf[4096]; /*arbitrary- enough to hold 30+ bss info.IOCTL buffer's size is limited to  WLC_IOCTL_MAXLEN(8192)*/
	}scan_results;
	wl_bss_info_t *pbss_info;
	int channel = 0;
	int rssi = 0;
	int percent = 0;
	char * adhoc;
	wlc_ssid_t ssid_temp;
	brcm_ie_t *ptr_ie;
	uint32 ie_len;
	uint8 len, id, flags;
	uchar oui[10];
	char *wds;
	static char temp[] = {0x0, 0x90, 0x4C};
		
	#define BRF_LZYWDS 0x04
	
	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
                return -1;

        snprintf(ifnames, sizeof(ifnames), "%s %s",
                nvram_safe_get("lan_ifnames"),
                nvram_safe_get("wan_ifnames"));
        foreach(name, ifnames, next) {
                if (wl_probe(name) == 0 &&
                    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 && unit == unit2)
			break;
	}

	if(*name == 0) {//shoudn't happen
		cprintf("Found No Wireless Interface.");
		return 0;
	}
	//scan for infra and adhoc mode
	char *mode = websGetVar(wp, "wl_mode", NULL);
	char tmp[30];
	if (!mode) {
		cprintf("mode is null\n");
		sprintf(tmp, "%s", nvram_safe_get("wl_mode")); 
	}
	else 
		sprintf(tmp, "%s", mode);
	cprintf("mode is: %s\n", tmp);
	if(strcmp(tmp,"wds"))
	{
	cprintf("Client defined! \n");
	usleep(1000000);
	if(!wl_ioctl(name, WLC_SCAN, &channel, sizeof(channel))){
		usleep(1000000);
		memset((char *)&scan_results, 0, sizeof(scan_results));
		scan_results.buflen = sizeof(scan_results);
		scan_results.version = WL_BSS_INFO_VERSION;
		if((! wl_ioctl(name, WLC_SCAN_RESULTS, &scan_results, sizeof(scan_results))) ||
			(scan_results.count != 0)) {
			//printf("scan_results.count: %d, \n", scan_results.count);
			pbss_info = (wl_bss_info_t *)scan_results.buf;
			for (i = 0; i < scan_results.count; i++) {
				memset(ssid,0,sizeof(ssid));
			 	strncpy(ssid,(char *)pbss_info->SSID,pbss_info->SSID_len);
				ether_etoa(pbss_info->BSSID.octet,eaddr);
				rssi = pbss_info->RSSI;
				if (rssi < SIGNAL_MIN)
					rssi = SIGNAL_MIN;
				if (rssi > SIGNAL_MAX)
					rssi = SIGNAL_MAX;
				percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
				//printf("RSSI: %d, \n", rssi);
				channel = pbss_info->channel;
				//printf("RSSI: %d, \n", rssi);
				if(pbss_info->capability & DOT11_CAP_PRIVACY)
					wep = "true";
				else
					wep = "false";
			       	
				if(pbss_info->capability & DOT11_CAP_IBSS)
					adhoc = "true";
				else
					adhoc = "false";
				//cprintf("adhoc is %s\n", adhoc);	
				ie_len = pbss_info->ie_length;
				//cprintf("ie_len: %d \n", ie_len);
				ptr_ie = (brcm_ie_t *)(((char *) pbss_info) + sizeof(wl_bss_info_t));
				wds = "false";
				while (ie_len > 0) {
					id = ptr_ie->id;
					len = ptr_ie->len; 
					if (id == 221) {
						memset(oui,0,10);
						strncpy(oui,(char *)ptr_ie->oui,3);
						if (!strncmp(oui,temp,3)) {
							flags = ptr_ie->flags;
							if (flags & BRF_LZYWDS)
								wds = "true";
							else
								;
						}
					}
					ie_len = ie_len - len - 2;
					ptr_ie = (brcm_ie_t *)(((char *)ptr_ie) + len + 2);
				}//while
				if (i < scan_results.count - 1) 					
					ret += websWrite(wp, "new Site(\"%s\", \"%s\", %d, %d, %s, %s, %s),", ssid, eaddr, channel, percent, wep, adhoc, wds);
				else
					ret += websWrite(wp, "new Site(\"%s\", \"%s\", %d, %d, %s, %s, %s)", ssid, eaddr, channel, percent, wep, adhoc, wds);
				pbss_info = (wl_bss_info_t *)(((char *) pbss_info) + pbss_info->length);
  			}
		}
		else{
			cprintf("Found no AP!");
		}
	}
	else{
		cprintf("Scan process failed!");
	}
	}
	else 	//wds scan
	{
	/* to be able to start the scanning process we have to kill SSID*/
	/* Nuke SSID  */
	cprintf("Repeater defined! \n");
	ssid_temp.SSID_len = 0;
	ssid_temp.SSID[0] = '\0';
	//cprintf("name_scan: %s \n", name);
	if (wl_ioctl(name, WLC_SET_SSID, &ssid_temp, sizeof(ssid_temp))) {
		cprintf("SET SSID failed!\n");
		cprintf("ssid: %s\n", ssid_temp.SSID);
	}
	wl_ioctl(name, WLC_DOWN, NULL, 0);
	wl_ioctl(name, WLC_UP, NULL, 0);
	usleep(2000000);
	//cprintf("start scan!\n");
	if (wl_ioctl(name, WLC_SCAN, &channel, sizeof(channel)))
	{
		cprintf("Scan failed!\n");
		//restore SSID
		ssid_temp.SSID_len = strlen(nvram_safe_get("wl0_ssid"));
		strncpy(ssid_temp.SSID, nvram_safe_get("wl0_ssid"), ssid_temp.SSID_len);
		wl_ioctl(name, WLC_SET_SSID, &ssid_temp, sizeof(ssid_temp));
		return ret;	
	}
	usleep(1000000);
	//cprintf("wait 1 sec\n");
	memset((char *)&scan_results, 0, sizeof(scan_results));
	scan_results.buflen = sizeof(scan_results);
	scan_results.version = WL_BSS_INFO_VERSION;
	if((! wl_ioctl(name, WLC_SCAN_RESULTS, &scan_results, sizeof(scan_results))) ||
		(scan_results.count != 0)) {
		printf("scan_results.count: %d, \n", scan_results.count);
		pbss_info = (wl_bss_info_t *)scan_results.buf;
		for (i = 0; i < scan_results.count; i++) {
			memset(ssid,0,sizeof(ssid));
		 	strncpy(ssid,(char *)pbss_info->SSID,pbss_info->SSID_len);
			ether_etoa(pbss_info->BSSID.octet,eaddr);
			rssi = pbss_info->RSSI;
			if (rssi < SIGNAL_MIN)
				rssi = SIGNAL_MIN;
			if (rssi > SIGNAL_MAX)
				rssi = SIGNAL_MAX;
			percent = (int)((((float)(rssi - SIGNAL_MIN))/(float)(SIGNAL_MAX - SIGNAL_MIN))*100.0);
			//cprintf("RSSI: %d, \n", rssi);
			channel = pbss_info->channel;
			//cprintf("channel: %d, \n", channel);
			if(pbss_info->capability & DOT11_CAP_PRIVACY)
				wep = "true";
			else
				wep = "false";
			//cprintf("wep: %s, \n", wep);
			if(pbss_info->capability & DOT11_CAP_IBSS)
					adhoc = "true";
				else
					adhoc = "false";
			
			ie_len = pbss_info->ie_length;
			//cprintf("ie_len: %d \n", ie_len);
			ptr_ie = (brcm_ie_t *)(((char *) pbss_info) + sizeof(wl_bss_info_t));
			wds = "false";
			while (ie_len > 0) {
				id = ptr_ie->id;
				//cprintf("id: %d ", id);
				len = ptr_ie->len; 
				//cprintf("length: %d ", len);
				if (id == 221) {
					memset(oui,0,10);
					strncpy(oui,(char *)ptr_ie->oui,3);
					if (!strncmp(oui,temp,3)) {
						flags = ptr_ie->flags;
						//cprintf("flags: %d\n", flags);
						if (flags & BRF_LZYWDS)
						{
							//cprintf("LAZYWDS = 1\n");
							wds = "true";
						}
						else
						{	//cprintf("LAZYWDS = 0\n");
							;//wds = "false";
							//break;
						}
					}
				}
				ie_len = ie_len - len - 2;
				ptr_ie = (brcm_ie_t *)(((char *)ptr_ie) + len + 2);
			}//while
			if (i < scan_results.count - 1) 					
				ret += websWrite(wp, "new Site(\"%s\", \"%s\", %d, %d, %s, %s, %s),", ssid, eaddr, channel, percent, wep, adhoc, wds);
			else
				ret += websWrite(wp, "new Site(\"%s\", \"%s\", %d, %d, %s, %s, %s)", ssid, eaddr, channel, percent, wep, adhoc, wds);
			pbss_info = (wl_bss_info_t *)(((char *) pbss_info) + pbss_info->length);
  		}
	}
	else{
		cprintf("Found no AP!");
	}
	
	//restore SSID
	ssid_temp.SSID_len = strlen(nvram_safe_get("wl0_ssid"));
	strncpy(ssid_temp.SSID, nvram_safe_get("wl0_ssid"), ssid_temp.SSID_len);
	wl_ioctl(name, WLC_SET_SSID, &ssid_temp, sizeof(ssid_temp));
	}
	
	return ret;
}

static int
ej_wl_wds_status(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	char *macs, *next, *name;
	char mac[100];
	int i, len;
	sta_info_t *sta;
	char buf[300];
	int ret = 0;
	
	if (ejArgs(argc, argv, "%d", &i) < 1) {
		websError(wp, 400, "Insufficient args\n");
		return -1;
	}

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));
	macs = nvram_safe_get(strcat_r(prefix, "wds", tmp));
	
	foreach(mac, macs, next) {
		//if (i-- == 0) {
			len = sprintf(buf, "sta_info");
			ether_atoe(mac, (unsigned char *)&buf[len + 1]);
			//if (atoi(nvram_safe_get(strcat_r(prefix, "wds_timeout", tmp))) &&
			if(!wl_ioctl(name, WLC_GET_VAR, buf, sizeof(buf))) {
				sta = (sta_info_t *)buf;
				ret += websWrite(wp, "<tr><td class=\"clsStatic\">%s</td> %s</tr>", mac, (sta->flags & WL_WDS_LINKUP) ? "<td><img src=\"check.gif\" border=\"0\" /></td>" : "");
			}
		//}
	}	

	return ret;
}
#endif

/* should #include <d11.h> if allowed */
/* phy types */
#define	PHY_TYPE_A		0
#define	PHY_TYPE_B		1
#define	PHY_TYPE_G		2
#define	PHY_TYPE_NULL		0xf

/* Return current phytype 
static int
ej_wl_cur_phytype(int eid, webs_t wp, int argc, char_t **argv)
{
	int unit;
	char *name;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";
	int phytype;

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));

	// Get configured phy type 
	wl_ioctl(name, WLC_GET_PHYTYPE, &phytype, sizeof(phytype));

	return websWrite(wp, "Current: 802.11%s", phytype == PHY_TYPE_A ? "a" :
		phytype == PHY_TYPE_B ? "b" : "g");
}
*/

#ifdef __CONFIG_NAT__
#ifdef ROUTER
static char * 
wan_name(int unit, char *prefix, char *name, int len)
{
	char tmp[NVRAM_BUFSIZE], *desc;
	desc = nvram_safe_get(strcat_r(prefix, "desc", tmp));
	snprintf(tmp, sizeof(tmp), "Connection %d", unit + 1);
	snprintf(name, len, "%s", !strcmp(desc, "") ? tmp : desc);
	return name;
}

/* Return a list of wan connections (Connection <N>/<Connection Name>) 
static int
ej_wan_list(int eid, webs_t wp, int argc, char_t **argv)
{
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";
	int unit, ret = 0;

	// build wan connection name list 
	for (unit = 0; unit < MAX_NVPARSE; unit ++) {
		wan_prefix(unit, prefix);
		if (!nvram_get(strcat_r(prefix, "unit", tmp)))
			continue;
		ret += websWrite(wp, "<option value=\"%d\" %s>%s</option>", unit,
				unit == atoi(nvram_safe_get("wan_unit")) ? "selected" : "",
				wan_name(unit, prefix, tmp, sizeof(tmp)));
	}

	return ret;
} */
#endif   /* ROUTER */
#endif	/* __CONFIG_NAT__ */

char *webs_buf=NULL;
int webs_buf_offset=0;

static void
validate_list(webs_t wp, char *value, struct variable *v,
	      int (*valid)(webs_t, char *, struct variable *))
{
	int n, i;
	char name[100];
	char buf[1000] = "", *cur = buf;

	n = atoi(value);

	for (i = 0; i < n; i++) {
		snprintf(name, sizeof(name), "%s%d", v->name, i);
		if (!(value = websGetVar(wp, name, NULL)))
			return;
		if (!*value && v->nullok)
			continue;
		if (!valid(wp, value, v))
			continue;
		cur += snprintf(cur, buf + sizeof(buf) - cur, "%s%s",
				cur == buf ? "" : " ", value);
	}

	nvram_set(v->name, buf);
}	

static int
valid_ipaddr(webs_t wp, char *value, struct variable *v)
{
	unsigned int buf[5];
	struct in_addr ipaddr, netaddr, broadaddr, netmask;
	
	//printf("validate ipaddress!\n");
	if (sscanf(value, "%d.%d.%d.%d.%d", &buf[0], &buf[1], &buf[2], &buf[3], &buf[5]) != 4) {
	//USR validation
		if (!flag)
			validation_page(wp);
		error = 1;
		websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. %s is not a valid IP address.</p>",
			  v->longname, value);
		return FALSE;
	}

	ipaddr.s_addr = htonl((buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]);

	if (v->argv) {
		(void) inet_aton(nvram_safe_get(v->argv[0]), &netaddr);
		(void) inet_aton(nvram_safe_get(v->argv[1]), &netmask);
		netaddr.s_addr &= netmask.s_addr;
		broadaddr.s_addr = netaddr.s_addr | ~netmask.s_addr;
		if (netaddr.s_addr != (ipaddr.s_addr & netmask.s_addr)) {
			//USR validation
			if (!flag)
				validation_page(wp);
			websBufferWrite(wp, "<p>Invalid value for <b>%s</b>.  %s is not in the %s/",
				  v->longname, value, inet_ntoa(netaddr));

			websBufferWrite(wp, "%s network.</p>", inet_ntoa(netmask));
			error = 1;
			return FALSE;
		}
		if (ipaddr.s_addr == netaddr.s_addr) {
			//USR validation
			if (!flag)
				validation_page(wp);
			websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. %s cannot be the network address.</p>",
				  v->longname, value);
			error = 1;
			return FALSE;
		}
		if (ipaddr.s_addr == broadaddr.s_addr) {
			//USR validation
			if (!flag)
				validation_page(wp);
			websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. %s cannot be the broadcast address.</p>",
				  v->longname, value);
			error = 1;
			return FALSE;
		}
	}

	return TRUE;
}

static void
validate_ipaddr(webs_t wp, char *value, struct variable *v)
{
	if (valid_ipaddr(wp, value, v))
		nvram_set(v->name, value);
}

static void
validate_ipaddrs(webs_t wp, char *value, struct variable *v)
{
	validate_list(wp, value, v, valid_ipaddr);
}

static int
valid_choice(webs_t wp, char *value, struct variable *v)
{
	char **choice=NULL;
	
	assert(v);
	for (choice = v->argv; *choice; choice++) {
		if (!strcmp(value, *choice))
			return TRUE;
	}
	//USR validation
	if (!flag)
		validation_page(wp);
	websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is not one of ", v->longname, value);
	for (choice = v->argv; *choice; choice++)
		websBufferWrite(wp, "%s%s", choice == v->argv ? "" : "/", *choice);
	websBufferWrite(wp, ".</p>");
	error = 1;
	return FALSE;
}

static void
validate_choice(webs_t wp, char *value, struct variable *v)
{
	if (valid_choice(wp, value, v)) 
		nvram_set(v->name, value);
}

static int
valid_range(webs_t wp, char *value, struct variable *v)
{
	int n, start, end;

	n = atoi(value);
	start = atoi(v->argv[0]);
	end = atoi(v->argv[1]);

	if (n < start || n > end) {
		//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is out of the range %d-%d.</p>",
			  v->longname, value, start, end);
		error = 1;
		return FALSE;
	}

	return TRUE;
}

static void
validate_range(webs_t wp, char *value, struct variable *v)
{
	if (valid_range(wp, value, v))
		nvram_set(v->name, value);
}

static int
valid_name(webs_t wp, char *value, struct variable *v)
{
	int n, min, max;

	n = strlen(value);
	min = atoi(v->argv[0]);
	max = atoi(v->argv[1]);

	if (n > max) {
		//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is longer than %d characters.</p>",
			  v->longname, value, max);
		error = 1;
		return FALSE;
	}
	else if (n < min) {
		//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is shorter than %d characters.</p>",
			  v->longname, value, min);
		error = 1;
		return FALSE;
	}

	return TRUE;
}

static void
validate_name(webs_t wp, char *value, struct variable *v)
{
	if (valid_name(wp, value, v))
		nvram_set(v->name, value);
}

static int
valid_hwaddr(webs_t wp, char *value, struct variable *v)
{
	unsigned char hwaddr[6];

	/* Make exception for "NOT IMPLELEMENTED" string */
	if (!strcmp(value,"NOT_IMPLEMENTED"))
		return(TRUE);

	/* Check for bad, multicast, broadcast, or null address */
	if (!ether_atoe(value, hwaddr) ||
	    (hwaddr[0] & 1) ||
	    (hwaddr[0] & hwaddr[1] & hwaddr[2] & hwaddr[3] & hwaddr[4] & hwaddr[5]) == 0xff ||
	    (hwaddr[0] | hwaddr[1] | hwaddr[2] | hwaddr[3] | hwaddr[4] | hwaddr[5]) == 0x00) {
	    	//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid MAC address. %s is not a valid MAC address.</p>", value);
		error = 1;
		return FALSE;
	}

	return TRUE;
}

#ifdef __CONFIG_NAT__
static void
validate_hwaddr(webs_t wp, char *value, struct variable *v)
{
	if (valid_hwaddr(wp, value, v))
		nvram_set(v->name, value);
}
#endif	/* __CONFIG_NAT__ */

static void
validate_hwaddrs(webs_t wp, char *value, struct variable *v)
{
	validate_list(wp, value, v, valid_hwaddr);
}

/*
static void
validate_country(webs_t wp, char *value, struct variable *v)
{
	country_name_t *country;
	for(country = country_names; country->name; country++)
		if (!strcmp(value, country->abbrev))
			nvram_set(v->name, value);
}
*/

static void
validate_dhcp(webs_t wp, char *value, struct variable *v)
{
	struct variable dhcp_variables[] = {
		{ longname: "DHCP Server Starting LAN IP Address", argv: ARGV("lan_ipaddr", "lan_netmask") },
		{ longname: "DHCP Server Ending LAN IP Address", argv: ARGV("lan_ipaddr", "lan_netmask") },
	};
	char *start, *end;

	if (!(start = websGetVar(wp, "dhcp_start", NULL)) ||
	    !(end = websGetVar(wp, "dhcp_end", NULL)))
		return;
	if (!*start) start = end;
	if (!*end) end = start;
	if (!*start && !*end && !strcmp(nvram_safe_get("lan_proto"), "dhcp")) {
		//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please specify a range.</p>", v->longname);
		error = 1;
		return;
	}
	if (!valid_ipaddr(wp, start, &dhcp_variables[0]) ||
	    !valid_ipaddr(wp, end, &dhcp_variables[1]))
		return;
	if (ntohl(inet_addr(start)) > ntohl(inet_addr(end))) {
		//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
			  dhcp_variables[0].longname, start, dhcp_variables[1].longname, end);
		error = 1;
		return;
	}

	nvram_set("dhcp_start", start);
	nvram_set("dhcp_end", end);
}

static void
validate_lan_ipaddr(webs_t wp, char *value, struct variable *v)
{
	struct variable fields[] = {
		{ name: "lan_ipaddr", longname: "LAN IP Address" },
		{ name: "lan_netmask", longname: "LAN Subnet Mask" },
	};
	char *lan_ipaddr, *lan_netmask;
	struct in_addr ipaddr, netmask, netaddr, broadaddr;
	char *lan_ipaddrs[] = { "dhcp_start", "dhcp_end", "dmz_ipaddr" };
#ifdef __CONFIG_NAT__
#ifdef ROUTER	
	netconf_filter_t start, end;
	netconf_nat_t nat;
	bool valid;
#endif
#endif	/* __CONFIG_NAT__ */
	int i;

	/* Basic validation */
	if (!(lan_ipaddr = websGetVar(wp, fields[0].name, NULL)) ||
	    !(lan_netmask = websGetVar(wp, fields[1].name, NULL)) ||
	    !valid_ipaddr(wp, lan_ipaddr, &fields[0]) ||
	    !valid_ipaddr(wp, lan_netmask, &fields[1]))
		return;

	/* Check for broadcast or network address */
	(void) inet_aton(lan_ipaddr, &ipaddr);
	(void) inet_aton(lan_netmask, &netmask);
	netaddr.s_addr = ipaddr.s_addr & netmask.s_addr;
	broadaddr.s_addr = netaddr.s_addr | ~netmask.s_addr;
	if (ipaddr.s_addr == netaddr.s_addr) {
		//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. %s cannot be the network address.</p>",
			  fields[0].longname, lan_ipaddr);
		error = 1;
		return;
	}
	if (ipaddr.s_addr == broadaddr.s_addr) {
		//USR validation
		if (!flag)
			validation_page(wp);
		websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. %s cannot be the broadcast address.</p>",
			  fields[0].longname, lan_ipaddr);
		error = 1;
		return;
	}

	nvram_set("lan_ipaddr", lan_ipaddr);
	nvram_set("lan_netmask", lan_netmask);

	/* Fix up LAN IP addresses */
	for (i = 0; i < ARRAYSIZE(lan_ipaddrs); i++) {
		value = nvram_get(lan_ipaddrs[i]);
		if (value && *value) {
			(void) inet_aton(value, &ipaddr);
			ipaddr.s_addr &= ~netmask.s_addr;
			ipaddr.s_addr |= netaddr.s_addr;
			nvram_set(lan_ipaddrs[i], inet_ntoa(ipaddr));
		}
	}				

#ifdef __CONFIG_NAT__
#ifdef ROUTER
	/* Fix up client filters and port forwards */
	for (i = 0; i < MAX_NVPARSE; i++) {
		if (get_filter_client(i, &start, &end)) {
			start.match.src.ipaddr.s_addr &= ~netmask.s_addr;
			start.match.src.ipaddr.s_addr |= netaddr.s_addr;
			end.match.src.ipaddr.s_addr &= ~netmask.s_addr;
			end.match.src.ipaddr.s_addr |= netaddr.s_addr;
			valid = set_filter_client(i, &start, &end);
			a_assert(valid);
		}
		if (get_forward_port(i, &nat)) {
			nat.ipaddr.s_addr &= ~netmask.s_addr;
			nat.ipaddr.s_addr |= netaddr.s_addr;
			valid = set_forward_port(i, &nat);
			a_assert(valid);
		}
	}
#endif   /* ROUTER */
#endif	/* __CONFIG_NAT__ */
}	

#ifdef __CONFIG_NAT__
#ifdef ROUTER
static void
validate_filter_client(webs_t wp, char *value, struct variable *v)
{
	int n, i, j;
	bool valid;
	struct variable fields[] = {
		{ name: "filter_client_from_start%d", longname: "LAN Client Filter Starting IP Address", argv: ARGV("lan_ipaddr", "lan_netmask") },
		{ name: "filter_client_from_end%d", longname: "LAN Client Filter Ending IP Address", argv: ARGV("lan_ipaddr", "lan_netmask") },
		{ name: "filter_client_proto%d", longname: "LAN Client Filter Protocol", argv: ARGV("tcp", "udp", "tcp/udp") },
		{ name: "filter_client_to_start%d", longname: "LAN Client Filter Starting Destination Port", argv: ARGV("0", "65535") },
		{ name: "filter_client_to_end%d", longname: "LAN Client Filter Ending Destination Port", argv: ARGV("0", "65535") },
		{ name: "filter_client_from_day%d", longname: "LAN Client Filter Starting Day", argv: ARGV("0", "6") },
		{ name: "filter_client_to_day%d", longname: "LAN Client Filter Ending Day", argv: ARGV("0", "6") },
		{ name: "filter_client_from_sec%d", longname: "LAN Client Filter Starting Second", argv: ARGV("0", "86400") },
		{ name: "filter_client_to_sec%d", longname: "LAN Client Filter Ending Second", argv: ARGV("0", "86400") },
		{ name: "filter_client_enable%d", longname: "Enable", argv: ARGV("on", "off") },
	};
	char *from_start, *from_end, *proto, *to_start, *to_end, *from_day, *to_day, *from_sec, *to_sec, *enable;
	char **locals[] = { &from_start, &from_end, &proto, &to_start, &to_end, &from_day, &to_day, &from_sec, &to_sec, &enable };
	char name[1000];
	netconf_filter_t start, end;

	/* filter_client indicates how many to expect */
	if (!valid_range(wp, value, v))
		return;
	n = atoi(value);
	
	/* USR delete unused fields */
	int max_number = 30;
	for (i = n; i < max_number; i++)
		del_filter_client(i);
	/* validate used fields */
	for (i = 0; i < n; i++) {
		/* Set up field names */
		for (j = 0; j < ARRAYSIZE(fields); j++) {
			snprintf(name, sizeof(name), fields[j].name, i);
			if (!(*locals[j] = websGetVar(wp, name, NULL)))
				break;
		}
		/* Incomplete web page */
		if (j < ARRAYSIZE(fields))
			continue;
		/* Enable is a checkbox 
		snprintf(name, sizeof(name), "filter_client_enable%d", i);
		if (websGetVar(wp, name, NULL))
			enable = "on";
		else
			enable = "off"; */
		/* Delete entry if all fields are blank */
		if (!*from_start && !*from_end && !*to_start && !*to_end) {
			del_filter_client(i);
			continue;
		}
		/* Fill in empty fields with default values */
		if (!*from_start) from_start = from_end;
		if (!*from_end) from_end = from_start;
		if (!*to_start) to_start = to_end;
		if (!*to_end) to_end = to_start;
		if (!*from_start || !*from_end) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter a LAN IP Address Range.</p>", v->longname);
			continue;
		}
		if (!*to_start || !*to_end) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;	
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter a Destination Port Range.</p>", v->longname);
			continue;
		}
		/* Check individual fields */
		if (!valid_ipaddr(wp, from_start, &fields[0]) ||
		    !valid_ipaddr(wp, from_end, &fields[1]) ||
		    !valid_choice(wp, proto, &fields[2]) ||
		    !valid_range(wp, to_start, &fields[3]) ||
		    !valid_range(wp, to_end, &fields[4]) ||
		    !valid_range(wp, from_day, &fields[5]) ||
		    !valid_range(wp, to_day, &fields[6]) ||
		    !valid_range(wp, from_sec, &fields[7]) ||
		    !valid_range(wp, to_sec, &fields[8]) ||
		    !valid_choice(wp, enable, &fields[9]))
			continue;
		/* Check dependencies between fields */
		if (ntohl(inet_addr(from_start)) > ntohl(inet_addr(from_end))) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;	
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
				  fields[0].longname, from_start, fields[1].longname, from_end);
			continue;
		}
		if (atoi(to_start) > atoi(to_end)) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;	
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
				  fields[3].longname, to_start, fields[4].longname, to_end);
			continue;
		}

		/* Set up parameters */
		memset(&start, 0, sizeof(netconf_filter_t));
		if (!strcmp(proto, "tcp"))
			start.match.ipproto = IPPROTO_TCP;
		else if (!strcmp(proto, "udp"))
			start.match.ipproto = IPPROTO_UDP;
		(void) inet_aton(from_start, &start.match.src.ipaddr);
		start.match.src.netmask.s_addr = htonl(0xffffffff);
		start.match.dst.ports[0] = htons(atoi(to_start));
		start.match.dst.ports[1] = htons(atoi(to_end));
		start.match.days[0] = atoi(from_day);
		start.match.days[1] = atoi(to_day);
		start.match.secs[0] = atoi(from_sec);
		start.match.secs[1] = atoi(to_sec);
		if (!strcmp(enable, "off"))
			start.match.flags |= NETCONF_DISABLED;
		/* solve tcp/udp */
		if (!strcmp(proto, "tcp/udp")) {
			start.match.ipproto = IPPROTO_TCP;
			memcpy(&end, &start, sizeof(netconf_filter_t));
			(void) inet_aton(from_end, &end.match.src.ipaddr);
			valid = set_filter_client(i, &start, &end);
			a_assert(valid);
			i++;
			n++;
			start.match.ipproto = IPPROTO_UDP;
			memcpy(&end, &start, sizeof(netconf_filter_t));
			(void) inet_aton(from_end, &end.match.src.ipaddr);
			valid = set_filter_client(i, &start, &end);
			a_assert(valid);
		}
		else {
			/* Do it */
			memcpy(&end, &start, sizeof(netconf_filter_t));
			(void) inet_aton(from_end, &end.match.src.ipaddr);
			valid = set_filter_client(i, &start, &end);
			a_assert(valid);
		}
	}
}

static void
validate_forward_port(webs_t wp, char *value, struct variable *v)
{
	int n, i, j;
	bool valid;
	struct variable fields[] = {
		{ name: "forward_port_proto%d", longname: "Port Forward Protocol", argv: ARGV("tcp", "udp", "tcp/udp") },
		{ name: "forward_port_from_start%d", longname: "Port Forward Starting WAN Port", argv: ARGV("0", "65535") },
		{ name: "forward_port_from_end%d", longname: "Port Forward Ending WAN Port", argv: ARGV("0", "65535") },
		{ name: "forward_port_to_ip%d", longname: "Port Forward LAN IP Address", argv: ARGV("lan_ipaddr", "lan_netmask") },
		{ name: "forward_port_to_start%d", longname: "Port Forward Starting LAN Port", argv: ARGV("0", "65535") },
		{ name: "forward_port_to_end%d", longname: "Port Forward Ending LAN Port", argv: ARGV("0", "65535") },
		{ name: "forward_port_enable%d", longname: "Enable", argv: ARGV("on", "off") },
	};
	char *proto, *from_start, *from_end, *to_ip, *to_start, *to_end, *enable;
	char **locals[] = { &proto, &from_start, &from_end, &to_ip, &to_start, &to_end, &enable };
	char name[1000];
	netconf_nat_t nat;

	/* forward_port indicates how many to expect */
	if (!valid_range(wp, value, v))
		return;
	n = atoi(value);
	
	/* USR delete unused fields */
	int max_number = 30;
	for (i = n; i < max_number; i++)
		del_forward_port(i);
	/* validate used fields */
	for (i = 0; i < n; i++) {
		/* Set up field names */
		for (j = 0; j < ARRAYSIZE(fields); j++) {
			snprintf(name, sizeof(name), fields[j].name, i);
			if (!(*locals[j] = websGetVar(wp, name, NULL)))
				break;
		}
		/* Incomplete web page */
		if (j < ARRAYSIZE(fields))
			continue;
		/* Enable is a checkbox 
		snprintf(name, sizeof(name), "forward_port_enable%d", i);
		if (websGetVar(wp, name, NULL))
			enable = "on";
		else
			enable = "off";*/
		/* Delete entry if all fields are blank */
		if (!*from_start && !*from_end && !*to_ip && !*to_start && !*to_end) {
			del_forward_port(i);
			continue;
		}
		/* Fill in empty fields with default values */
		if (!*from_start) from_start = from_end;
		if (!*from_end) from_end = from_start;
		if (!*to_start && !*to_end)
			to_start = from_start;
		if (!*to_start) to_start = to_end;
		if (!*to_end) to_end = to_start;
		if (!*from_start || !*from_end) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter a LAN IP Address Range.</p>", v->longname);
			continue;
		}
		if (!*to_ip) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter a LAN IP Address.</p>", v->longname);
			continue;
		}
		if (!*to_start || !*to_end) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter a Destination Port Range.</p>", v->longname);
			continue;
		}
		/* Check individual fields */
		if (!valid_choice(wp, proto, &fields[0]) ||
		    !valid_range(wp, from_start, &fields[1]) ||
		    !valid_range(wp, from_end, &fields[2]) ||
		    !valid_ipaddr(wp, to_ip, &fields[3]) ||
		    !valid_range(wp, to_start, &fields[4]) ||
		    !valid_range(wp, to_end, &fields[5]) ||
		    !valid_choice(wp, enable, &fields[6]))
			continue;
		if (atoi(from_start) > atoi(from_end)) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
				  fields[1].longname, from_start, fields[2].longname, from_end);
			continue;
		}
		if (atoi(to_start) > atoi(to_end)) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
				  fields[4].longname, to_start, fields[5].longname, to_end);
			continue;
		}
		if ((atoi(from_end) - atoi(from_start)) != (atoi(to_end) - atoi(to_start))) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. WAN Port Range and LAN Port Range must be the same size.</p>", v->longname);
			continue;
		}

		/* Set up parameters */
		memset(&nat, 0, sizeof(netconf_nat_t));
		if (!strcmp(proto, "tcp"))
			nat.match.ipproto = IPPROTO_TCP;
		else if (!strcmp(proto, "udp"))
			nat.match.ipproto = IPPROTO_UDP;
		nat.match.dst.ports[0] = htons(atoi(from_start));
		nat.match.dst.ports[1] = htons(atoi(from_end));
		(void) inet_aton(to_ip, &nat.ipaddr);
		nat.ports[0] = htons(atoi(to_start));
		nat.ports[1] = htons(atoi(to_end));
		if (!strcmp(enable, "off"))
			nat.match.flags |= NETCONF_DISABLED;
		/* solve tcp/udp */
		if (!strcmp(proto, "tcp/udp")) {
			nat.match.ipproto = IPPROTO_TCP;
			valid = set_forward_port(i, &nat);
			a_assert(valid);
			i++;
			n++;
			nat.match.ipproto = IPPROTO_UDP;
			valid = set_forward_port(i, &nat);
			a_assert(valid);
		}
		else {
			/* Do it */
			valid = set_forward_port(i, &nat);
			a_assert(valid);
		}
	}
}

static void
validate_autofw_port(webs_t wp, char *value, struct variable *v)
{
	int n, i, j;
	bool valid;
	struct variable fields[] = {
		{ name: "autofw_port_out_proto%d", longname: "Outbound Protocol", argv: ARGV("tcp", "udp", "tcp/udp") },
		{ name: "autofw_port_out_start%d", longname: "Outbound Port Start", argv: ARGV("0", "65535") },
		{ name: "autofw_port_out_end%d", longname: "Outbound Port End", argv: ARGV("0", "65535") },
		{ name: "autofw_port_in_proto%d", longname: "Inbound Protocol", argv: ARGV("tcp", "udp", "tcp/udp") },
		{ name: "autofw_port_in_start%d", longname: "Inbound Port Start", argv: ARGV("0", "65535") },
		{ name: "autofw_port_in_end%d", longname: "Inbound Port End", argv: ARGV("0", "65535") },
 		{ name: "autofw_port_to_start%d", longname: "To Port Start", argv: ARGV("0", "65535") },
 		{ name: "autofw_port_to_end%d", longname: "To Port End", argv: ARGV("0", "65535") },
		{ name: "autofw_port_enable%d", longname: "Enable", argv: ARGV("on", "off") },
	};
	char *out_proto, *out_start, *out_end, *in_proto, *in_start, *in_end, *to_start, *to_end, *enable;
	char **locals[] = { &out_proto, &out_start, &out_end, &in_proto, &in_start, &in_end, &to_start, &to_end, &enable };
	char name[1000];
	netconf_app_t app;

	/* autofw_port indicates how many to expect */
	if (!valid_range(wp, value, v))
		return;
	n = atoi(value);
	/* USR delete unused fields */
	int max_number = 30;
	for (i = n; i < max_number; i++)
		del_autofw_port(i);
	/* validate used fields */	
	for (i = 0; i < n; i++) {
		/* Set up field names */
		for (j = 0; j < ARRAYSIZE(fields); j++) {
			snprintf(name, sizeof(name), fields[j].name, i);
			if (!(*locals[j] = websGetVar(wp, name, NULL)))
				break;
		}
		/* Incomplete web page */
		if (j < ARRAYSIZE(fields))
			continue;
		/* Enable is a checkbox 
		snprintf(name, sizeof(name), "autofw_port_enable%d", i);
		if (websGetVar(wp, name, NULL))
			enable = "on";
		else
			enable = "off";*/
		/* Delete entry if all fields are blank */
		if (!*out_start && !*out_end && !*in_start && !*in_end && !*to_start && !*to_end) {
			del_autofw_port(i);
			continue;
		}
		/* Fill in empty fields with default values */
		if (!*out_start) out_start = out_end;
		if (!*out_end) out_end = out_start;
		if (!*in_start) in_start = in_end;
		if (!*in_end) in_end = in_start;
		if (!*to_start && !*to_end)
			to_start = in_start;
		if (!*to_start) to_start = to_end;
		if (!*to_end) to_end = to_start;
		if (!*out_start || !*out_end) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter an Outbound Port Range.</p>", v->longname);
			continue;
		}
		if (!*in_start || !*in_end) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter an Inbound Port Range.</p>", v->longname);
			continue;
		}
		if (!*to_start || !*to_end) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Please enter a To Port Range.</p>", v->longname);
			continue;
		}
		/* Check individual fields */
		if (!valid_choice(wp, out_proto, &fields[0]) ||
		    !valid_range(wp, out_start, &fields[1]) ||
		    !valid_range(wp, out_end, &fields[2]) ||
		    !valid_choice(wp, in_proto, &fields[3]) ||
		    !valid_range(wp, in_start, &fields[4]) ||
		    !valid_range(wp, in_end, &fields[5]) ||
		    !valid_range(wp, to_start, &fields[6]) ||
		    !valid_range(wp, to_end, &fields[7]) ||
		    !valid_choice(wp, enable, &fields[8]))
			continue;
		/* Check dependencies between fields */
		if (atoi(out_start) > atoi(out_end)) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
				  fields[1].longname, out_start, fields[2].longname, out_end);
			continue;
		}
		if (atoi(in_start) > atoi(in_end)) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
				  fields[4].longname, in_start, fields[5].longname, in_end);
			continue;
		}
		if (atoi(to_start) > atoi(to_end)) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s is greater than <b>%s</b> %s.</p>",
				  fields[6].longname, in_start, fields[7].longname, in_end);
			continue;
		}
		if ((atoi(in_end) - atoi(in_start)) != (atoi(to_end) - atoi(to_start))) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Inbound Port Range and To Port Range must be the same size.</p>", v->longname);
			continue;
		}
#ifdef NEW_PORT_TRIG
		if ((atoi(in_end) - atoi(in_start)) > 100) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. Inbound Port Range must be less than 100.</p>", v->longname);
			continue;
		}
#endif

		/* Set up parameters */
		memset(&app, 0, sizeof(netconf_app_t));
		if (!strcmp(out_proto, "tcp"))
			app.match.ipproto = IPPROTO_TCP;
		else if (!strcmp(out_proto, "udp"))
			app.match.ipproto = IPPROTO_UDP;
		app.match.dst.ports[0] = htons(atoi(out_start));
		app.match.dst.ports[1] = htons(atoi(out_end));
		if (!strcmp(in_proto, "tcp"))
			app.proto = IPPROTO_TCP;
		else if (!strcmp(in_proto, "udp"))
			app.proto = IPPROTO_UDP;
		app.dport[0] = htons(atoi(in_start));
		app.dport[1] = htons(atoi(in_end));
		app.to[0] = htons(atoi(to_start));
		app.to[1] = htons(atoi(to_end));
		if (!strcmp(enable, "off"))
			app.match.flags |= NETCONF_DISABLED;
		/* solve tcp/udp proto */
		if (!strcmp(out_proto, "tcp/udp")) {
			app.match.ipproto = IPPROTO_TCP;
			if (!strcmp(in_proto, "tcp/udp")) {
				app.proto = IPPROTO_TCP;
				valid = set_autofw_port(i, &app);
				a_assert(valid);
				i++;
				n++;
				app.match.ipproto = IPPROTO_UDP;
				app.proto = IPPROTO_UDP;
				valid = set_autofw_port(i, &app);
				a_assert(valid);
			}
			else {
				valid = set_autofw_port(i, &app);
				a_assert(valid);
				i++;
				n++;
				app.match.ipproto = IPPROTO_UDP;
				valid = set_autofw_port(i, &app);
				a_assert(valid);
			}
		}
		else if (!strcmp(in_proto, "tcp/udp")) {
			app.proto = IPPROTO_TCP;
			valid = set_autofw_port(i, &app);
			a_assert(valid);
			i++;
			n++;
			app.proto = IPPROTO_UDP;
			valid = set_autofw_port(i, &app);
			a_assert(valid);
		}
		else {
			/* Do it */
			valid = set_autofw_port(i, &app);
			a_assert(valid);
		}
	}
}
#endif /* ROUTER */
#endif	/* __CONFIG_NAT__ */

static void
validate_lan_route(webs_t wp, char *value, struct variable *v)
{
	int n, i;
	char buf[1000] = "", *cur = buf;
	struct variable lan_route_variables[] = {
		{ longname: "Route IP Address", argv: NULL },
		{ longname: "Route Subnet Mask", argv: NULL },
		{ longname: "Route Gateway", argv: NULL },
		{ longname: "Route Metric", argv: ARGV("0", "15") },
	};

	n = atoi(value);

	for (i = 0; i < n; i++) {
		char lan_route_ipaddr[] = "lan_route_ipaddrXXX";
		char lan_route_netmask[] = "lan_route_netmaskXXX";
		char lan_route_gateway[] = "lan_route_gatewayXXX";
		char lan_route_metric[] = "lan_route_metricXXX";
		char *ipaddr, *netmask, *gateway, *metric;

 		snprintf(lan_route_ipaddr, sizeof(lan_route_ipaddr), "%s_ipaddr%d", v->name, i);
		snprintf(lan_route_netmask, sizeof(lan_route_netmask), "%s_netmask%d", v->name, i);
 		snprintf(lan_route_gateway, sizeof(lan_route_gateway), "%s_gateway%d", v->name, i);
 		snprintf(lan_route_metric, sizeof(lan_route_metric), "%s_metric%d", v->name, i);
		if (!(ipaddr = websGetVar(wp, lan_route_ipaddr, NULL)) ||
		    !(netmask = websGetVar(wp, lan_route_netmask, NULL)) ||
		    !(gateway = websGetVar(wp, lan_route_gateway, NULL)) ||
		    !(metric = websGetVar(wp, lan_route_metric, NULL)))
			return;
		if (!*ipaddr && !*netmask && !*gateway && !*metric)
			continue;
		if (!*ipaddr && !*netmask && *gateway) {
			ipaddr = "0.0.0.0";
			netmask = "0.0.0.0";
		}
		if (!*gateway)
			gateway = "0.0.0.0";
		if (!*metric)
			metric = "0";
		if (!*ipaddr) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. Please enter an IP Address.</p>", v->longname);
			continue;
		}
		if (!*netmask) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. Please enter a Subnet Mask.</p>", v->longname);
			continue;
		}
		lan_route_variables[2].argv = ARGV("lan_ipaddr", "lan_netmask");
		if (!valid_ipaddr(wp, ipaddr, &lan_route_variables[0]) ||
		    !valid_ipaddr(wp, netmask, &lan_route_variables[1]) ||
		    !valid_ipaddr(wp, gateway, &lan_route_variables[2]) ||
		    !valid_range(wp, metric, &lan_route_variables[3]))
			continue;
		cur += snprintf(cur, buf + sizeof(buf) - cur, "%s%s:%s:%s:%s",
				cur == buf ? "" : " ", ipaddr, netmask, gateway, metric);
	}

	nvram_set(v->name, buf);
}

#ifdef __CONFIG_NAT__
#ifdef ROUTER
static void
validate_wan_route(webs_t wp, char *value, struct variable *v)
{
	int n, i;
	char buf[1000] = "", *cur = buf;
	struct variable wan_route_variables[] = {
		{ longname: "Route IP Address", argv: NULL },
		{ longname: "Route Subnet Mask", argv: NULL },
		{ longname: "Route Gateway", argv: NULL },
		{ longname: "Route Metric", argv: ARGV("0", "15") },
	};

	n = atoi(value);

	for (i = 0; i < n; i++) {
		char wan_route_ipaddr[] = "wan_route_ipaddrXXX";
		char wan_route_netmask[] = "wan_route_netmaskXXX";
		char wan_route_gateway[] = "wan_route_gatewayXXX";
		char wan_route_metric[] = "wan_route_metricXXX";
		char *ipaddr, *netmask, *gateway, *metric;

 		snprintf(wan_route_ipaddr, sizeof(wan_route_ipaddr), "%s_ipaddr%d", v->name, i);
		snprintf(wan_route_netmask, sizeof(wan_route_netmask), "%s_netmask%d", v->name, i);
 		snprintf(wan_route_gateway, sizeof(wan_route_gateway), "%s_gateway%d", v->name, i);
 		snprintf(wan_route_metric, sizeof(wan_route_metric), "%s_metric%d", v->name, i);
		if (!(ipaddr = websGetVar(wp, wan_route_ipaddr, NULL)) ||
		    !(netmask = websGetVar(wp, wan_route_netmask, NULL)) ||
		    !(gateway = websGetVar(wp, wan_route_gateway, NULL)) ||
		    !(metric = websGetVar(wp, wan_route_metric, NULL)))
			continue;
		if (!*ipaddr && !*netmask && !*gateway && !*metric)
			continue;
		if (!*ipaddr && !*netmask && *gateway) {
			ipaddr = "0.0.0.0";
			netmask = "0.0.0.0";
		}
		if (!*gateway)
			gateway = "0.0.0.0";
		if (!*metric)
			metric = "0";
		if (!*ipaddr) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. Please enter an IP Address.</p>", v->longname);
			continue;
		}
		if (!*netmask) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>%s</b>. Please enter a Subnet Mask.</p>", v->longname);
			continue;
		}
		if (!valid_ipaddr(wp, ipaddr, &wan_route_variables[0]) ||
		    !valid_ipaddr(wp, netmask, &wan_route_variables[1]) ||
		    !valid_ipaddr(wp, gateway, &wan_route_variables[2]) ||
		    !valid_range(wp, metric, &wan_route_variables[3]))
			continue;
		cur += snprintf(cur, buf + sizeof(buf) - cur, "%s%s:%s:%s:%s",
				cur == buf ? "" : " ", ipaddr, netmask, gateway, metric);
	}

	nvram_set(v->name, buf);
}
#endif   /* ROUTER */
#endif	/* __CONFIG_NAT__ */

static void
validate_wl_preauth(webs_t wp, char *value, struct variable *v)
{
	if (!strcmp(value, "disabled"))
		nvram_set(v->name, "0");
	else 
		nvram_set(v->name, "1");
	return;
}

#ifdef CLIENT
static void
validate_cbAuthEncrypt(webs_t wp, char *value, struct variable *v)
{
	if (!valid_choice(wp, value, v))
		return;
		
	if (!strcmp(value, "wpa") || !strcmp(value, "wpa2")) {
		char *key = websGetVar(wp, "wl_wpa_psk", "");
		if (!key || !strcmp(key, "")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>Network Key</b>. Please enter a new key.</p>");
			return;
		}
		nvram_set("cbAuthEncrypt", value);
		nvram_set("wl_wep", "disabled");
		nvram_set("wl_auth", "0");
		nvram_set("wl_auth_mode", "none");
		if (!strcmp(value, "wpa")) {
			nvram_set("wl_akm", "psk ");
			return;
		}
				
		if (!strcmp(value, "wpa2")) {
			nvram_set("wl_akm", "psk2 ");
			return;
		}
	}
		
	if (!strcmp(value, "open")) {
		char *wep = websGetVar(wp, "wl_wep", "");
		
		if(!strcmp(wep,"enabled")) {
			char *number = websGetVar(wp, "wl_key", "");
			char *key1 = websGetVar(wp, "wl_key1", "");
			char *key2 = websGetVar(wp, "wl_key2", "");
			char *key3 = websGetVar(wp, "wl_key3", "");
			char *key4 = websGetVar(wp, "wl_key4", "");
			char wl_key[] = "wl_keyXXX";
			char *key = NULL;
				
			snprintf(wl_key, sizeof(wl_key), "wl_key%s", number);
		
			if (!strcmp(number, "1"))
				key = key1;
			else if (!strcmp(number, "2"))
				key = key2;
			else if (!strcmp(number, "3"))
				key = key3;
			else
				key = key4;	
		
			if (!key || !strcmp(key, "")) {
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				websBufferWrite(wp, "<p>Invalid value for <b>Network Key %s</b>. Please enter a new key.</p>", number);
				return;
			}
		}
		nvram_set("cbAuthEncrypt", value);
		nvram_set("wl_auth_mode", "none");
		nvram_set("wl_akm", "");
		nvram_set("wl_auth", "0");
		return;
	}
	if (!strcmp(value, "shared")) {
		char *wep = websGetVar(wp, "wl_wep", "");
		
		if (!wep  || !strcmp(wep,"disabled")) {
		//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid settings. <b>WEP Encryption</b> must be <b>Enabled</b> when <b>Authentication</b> is <b>Shared Key</b>.</p>"); 
			return;
		}
		
		char *number = websGetVar(wp, "wl_key", "");
		char *key1 = websGetVar(wp, "wl_key1", "");
		char *key2 = websGetVar(wp, "wl_key2", "");
		char *key3 = websGetVar(wp, "wl_key3", "");
		char *key4 = websGetVar(wp, "wl_key4", "");
		char wl_key[] = "wl_keyXXX";
		char *key = NULL;
				
		snprintf(wl_key, sizeof(wl_key), "wl_key%s", number);
		
		if (!strcmp(number, "1"))
			key = key1;
		else if (!strcmp(number, "2"))
			key = key2;
		else if (!strcmp(number, "3"))
			key = key3;
		else
			key = key4;	
		//printf("key: %s\n", key);
		if (!key || !strcmp(key, "")) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>Network Key %s</b>. Please enter a new key.</p>", number);
			return;
		}
		nvram_set("cbAuthEncrypt", value);
		nvram_set("wl_auth_mode", "none");
		nvram_set("wl_akm", "");
		nvram_set("wl_wep", "enabled");
		nvram_set("wl_auth", "1");
		return;
	}
}
#else
static void
validate_cbAuthEncrypt(webs_t wp, char *value, struct variable *v)
{
	if (!valid_choice(wp, value, v))
		return;
	
	char *cbEncryption = websGetVar(wp, "cbEncryption", "");
	//printf("cbEncrypt: %s\n", cbEncryption);	
	if (!strcmp(value, "none"))
	{ 
		nvram_set("cbAuthEncrypt", value);
		nvram_set("wl_auth_mode", "none");
		nvram_set("wl_auth", "0");
		nvram_set("wl_wep", "disabled");
		nvram_set("wl_akm", "");
		return;
	}
	
	if (!strcmp(value, "wpa12") || !strcmp(value, "wpa2") || !strcmp(value, "wpa")) {
		char *key = websGetVar(wp, "wl_wpa_psk", "");
		//printf("cbEncrypt: %s\n", cbEncryption);
		if ( !cbEncryption || !strcmp(cbEncryption, "wl_wep_open") || !strcmp(cbEncryption, "wl_wep_shared")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			//nvram_unset("cbEncryption");
			websWrite(wp, "<p>Error, <b>TKIP, AES or TKIP & AES Encryption</b> must be selected when <b>WPA and/or WPA2</b> method is chosen.</p>");
			return;	
		}
		if (!key || !strcmp(key, "")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>Pass phrase</b>. Please enter a new Pass phrase.</p>");
			return;
		}
		nvram_set("cbAuthEncrypt", value);
		nvram_set("wl_wep", "disabled");
		nvram_set("wl_auth", "0");
		nvram_set("wl_auth_mode", "none");
		if (!strcmp(value, "wpa12")) {
			nvram_set("wl_akm", "psk psk2 ");
			return;
		}
		
		if (!strcmp(value, "wpa2")) {
			nvram_set("wl_akm", "psk2 ");
			return;
		}
	
		if (!strcmp(value, "wpa")) {
			nvram_set("wl_akm", "psk ");
			return;
		}
	}
		
	if (!strcmp(value, "open") || !strcmp(value, "shared")) {
		char *number = websGetVar(wp, "wl_key", "");
		char *key1 = websGetVar(wp, "wl_key1", "");
		char *key2 = websGetVar(wp, "wl_key2", "");
		char *key3 = websGetVar(wp, "wl_key3", "");
		char *key4 = websGetVar(wp, "wl_key4", "");
		char wl_key[] = "wl_keyXXX";
		char *key = NULL;
				
		snprintf(wl_key, sizeof(wl_key), "wl_key%s", number);
		
		if (!strcmp(number, "1"))
			key = key1;
		else if (!strcmp(number, "2"))
			key = key2;
		else if (!strcmp(number, "3"))
			key = key3;
		else
			key = key4;	
		//printf("key: %s\n", key);
		if (!key || !strcmp(key, "")) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
#ifdef REPEATER
			websBufferWrite(wp, "<p>Invalid value for <b>Key</b>. Please enter a new key.</p>");
#else		
			websBufferWrite(wp, "<p>Invalid value for <b>Key %s</b>. Please enter a new key.</p>", number);
#endif			
			return;
		}
		nvram_set("cbAuthEncrypt", value);
		nvram_set("wl_auth_mode", "none");
		nvram_set("wl_akm", "");
		nvram_set("wl_wep", "enabled");
		if (!strcmp(value, "open")) {
			nvram_set("wl_auth", "0");
			return;
		}
		if (!strcmp(value, "shared")) {
			nvram_set("wl_auth", "1");
			return;
		}
	}
	
	if (!strcmp(value, "radius-wpa12") || !strcmp(value, "radius-wpa2") || !strcmp(value, "radius-wpa")) {
		char *ipaddr = websGetVar(wp, "wl_radius_ipaddr", "");
		if (!ipaddr || !strcmp(ipaddr, "")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websWrite(wp, "<p>Invalid value for <b>RADIUS Server</b>. Please enter a new RADIUS Server.</p>");
			return;
		}
		nvram_set("cbAuthEncrypt", value);	
		nvram_set("wl_auth_mode", "radius");
		nvram_set("wl_auth", "0");
		if ( !strcmp(cbEncryption, "tkip+aes") || !strcmp(cbEncryption, "aes") || !strcmp(cbEncryption, "tkip")) {
			nvram_set("wl_wep", "disabled");
			if (!strcmp(value, "radius-wpa12"))
			 	nvram_set("wl_akm", "wpa wpa2 ");
			else if (!strcmp(value, "radius-wpa2"))
				nvram_set("wl_akm", "wpa2 ");
			else if (!strcmp(value, "radius-wpa"))
				nvram_set("wl_akm", "wpa ");
			return;
		}
		if ( !strcmp(cbEncryption, "wl_wep_open") || !strcmp(cbEncryption, "wl_wep_shared")) {
			nvram_set("wl_wep", "enabled");
			nvram_set("wl_akm", "");
			return;
		}
	}
}
#endif

static void
validate_cbEncryption(webs_t wp, char *value, struct variable *v)
{
	if (!valid_choice(wp, value, v))
		return;
	
	nvram_set("cbEncryption", value);
	if (!strcmp(value, "tkip")) {
		nvram_set("wl_crypto", "tkip");
		return;
	}
	
	if (!strcmp(value, "aes")) {
		nvram_set("wl_crypto", "aes");
		return;
	}
	
	if (!strcmp(value, "tkip+aes")) {
		nvram_set("wl_crypto", "tkip+aes");
		return;
	}
}

static void
validate_connection(webs_t wp, char *value, struct variable *v)
{
	if (!valid_choice(wp, value, v))
		return;
	
	if (!strcmp(value, "keepalive")) {
		nvram_set("connection", "keepalive");
		nvram_set("wan_pppoe_keepalive", "1");
		nvram_set("wan0_pppoe_keepalive", "1");
		nvram_set("wan_pppoe_demand", "0");
		nvram_set("wan0_pppoe_demand", "0");
		return;
	}
	
	if (!strcmp(value, "demand")) {
		nvram_set("connection", "demand");
		nvram_set("wan_pppoe_keepalive", "0");
		nvram_set("wan0_pppoe_keepalive", "0");
		nvram_set("wan_pppoe_demand", "1");
		nvram_set("wan0_pppoe_demand", "1");
		return;
	}
	
	if (!strcmp(value, "timeout"))
	{ 
		
		nvram_set("connection", "timeout");
		nvram_set("wan_pppoe_keepalive", "0");
		nvram_set("wan0_pppoe_keepalive", "0");
		nvram_set("wan_pppoe_demand", "0");
		nvram_set("wan0_pppoe_demand", "0");	
		return;
	}
}		


static void
validate_wl_auth(webs_t wp, char *value, struct variable *v)
{
	if (!valid_choice(wp, value, v))
		return;

	if (!strcmp(value, "1")) {
		char *wep = websGetVar(wp, "wl_wep", "");
		if (!wep || strcmp(wep, "enabled")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid settings. <b>WEP Encryption</b> must be <b>Enabled</b> when <b>802.11 Authentication</b> is <b>%s</b>.</p>", value);
			return;
		}
	}

	nvram_set(v->name, value);
}

static void
validate_wl_auth_mode(webs_t wp, char *value, struct variable *v)
{
	if (!valid_choice(wp, value, v))
		return;

	if (!strcmp(value, "radius")) {
		char *wep = websGetVar(wp, "wl_wep", "");
		char *ipaddr = websGetVar(wp, "wl_radius_ipaddr", "");
		if (!wep || strcmp(wep, "enabled")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Error, <b>WEP Encryption</b> must be <b>Enabled</b> when <b>Network Authentication</b> is <b>%s</b>.</p>", value);
			return;
		}
		if (!ipaddr || !strcmp(ipaddr, "")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>RADIUS Server</b>. Please enter a new RADIUS Server.</p>");
			return;
		}
	}
	
	nvram_set(v->name, value);
}


static void 
validate_wl_akm(webs_t wp, char *value, struct variable *v) 
{
	char akms[WLC_IOCTL_SMLEN] = "";
	char *wpa, *psk;
	char *wpa2, *psk2;

	wpa = websGetVar(wp, "wl_akm_wpa", NULL);
	psk = websGetVar(wp, "wl_akm_psk", NULL);
	wpa2 = websGetVar(wp, "wl_akm_wpa2", NULL);
	psk2 = websGetVar(wp, "wl_akm_psk2", NULL);
	if (!wpa || !psk
	    || !wpa2 || !psk2
	    ) {
		return;
	}
	
	if (!strcmp(wpa, "enabled") 
	    || !strcmp(wpa2, "enabled")
	    ) {
		char *ipaddr = websGetVar(wp, "wl_radius_ipaddr", "");
		if (!ipaddr || !strcmp(ipaddr, "")) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid value for <b>RADIUS Server</b>. Please enter a new RADIUS Server.</p>");
			return;
		}
	}

	if (!strcmp(psk, "enabled") 
	    || !strcmp(psk2, "enabled")
	    ) {
		char *key = websGetVar(wp, "wl_wpa_psk", "");
		if (!key || !strcmp(key, "")) {
			if (!flag)
				validation_page(wp);
			error = 1;
#ifdef CLIENT
			websBufferWrite(wp, "<p>Invalid value for <b>WPA Pre-Shared Key</b>. Please enter a new Key.</p>");
#else
			websBufferWrite(wp, "<p>Invalid value for <b>Pass phrase</b>. Please enter a new Pass phrase.</p>");			
#endif			
			return;
		}
	} 

	if (!strcmp(wpa, "enabled") || !strcmp(psk, "enabled") 
	    || !strcmp(wpa2, "enabled") || !strcmp(psk2, "enabled")
	    ) {
		char *crypto = websGetVar(wp, "wl_crypto", "");
		if (!crypto || (strcmp(crypto, "tkip") && strcmp(crypto, "aes") &&
			strcmp(crypto, "tkip+aes"))) {
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p>Invalid <b>%s</b>. <b>Encryption</b> mode must be TKIP or AES or TKIP & AES.</p>", v->longname);
			return;
		}
	}

	if (!strcmp(wpa, "enabled"))
		strcat(akms, "wpa ");
	if (!strcmp(psk, "enabled"))
		strcat(akms, "psk ");
	if (!strcmp(wpa2, "enabled"))
		strcat(akms, "wpa2 ");
	if (!strcmp(psk2, "enabled"))
		strcat(akms, "psk2 ");

	nvram_set("wl_akm", akms);
}


static void
validate_wl_wpa_psk(webs_t wp, char *value, struct variable *v)
{
	int len = strlen(value);
	char *c;

	if (len == 64) {
		for (c = value; *c; c++) {
			if (!isxdigit((int) *c)) {
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				websBufferWrite(wp, "<p>Invalid <b>%s</b>. The character %c is not a hexadecimal digit.</p>", v->longname, *c);
				return;
			}
		}
	} else if (len < 8 || len > 63) {
		//USR validation
		if (!flag)
			validation_page(wp);
		error = 1;
#ifdef CLIENT		
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. The key must be between 8 and 63 ASCII characters or 64 hexadecimal digits.</p>", v->longname);
#else
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. The Pass phrase must be between 8 and 63 ASCII characters or 64 hexadecimal digits.</p>", v->longname);		
#endif		
		return;
	}

	nvram_set(v->name, value);
}

static void
validate_key(webs_t wp, char *value, struct variable *v)
{
	char *cbAuthEncrypt = websGetVar(wp, "cbAuthEncrypt", "");
	
	if (!cbAuthEncrypt || (strcmp(cbAuthEncrypt, "open") && strcmp(cbAuthEncrypt, "shared")))
		return;
	
	char *key1 = websGetVar(wp, "wl_key1", "");
	char *key2 = websGetVar(wp, "wl_key2", "");
	char *key3 = websGetVar(wp, "wl_key3", "");
	char *key4 = websGetVar(wp, "wl_key4", "");
	char wl_key[] = "wl_keyXXX";
	char *key = NULL;
				
	snprintf(wl_key, sizeof(wl_key), "wl_key%s", value);
		
	if (!strcmp(value, "1"))
		key = key1;
	else if (!strcmp(value, "2"))
		key = key2;
	else if (!strcmp(value, "3"))
		key = key3;
	else
		key = key4;
	
	if (!key || !strcmp(key, "")) 
		return;
#ifdef CLIENT	
	char *c;
	switch (strlen(key)) {
	case 5:
	case 13:	
		break;
	case 10:
	case 26:
		for (c = key; *c; c++) {
			if (!isxdigit((int) *c)) 
				return;
		}
		break;
	default:
		return;
	}
#endif
	nvram_set(v->name, value);
}

static void
validate_wl_key(webs_t wp, char *value, struct variable *v)
{
	char *c;
	char *cbAuthEncrypt = websGetVar(wp, "cbAuthEncrypt", "");
	
	if (!cbAuthEncrypt || (strcmp(cbAuthEncrypt, "open") && strcmp(cbAuthEncrypt, "shared")))
		return;
#ifdef CLIENT
	;
#else	
	char *number = websGetVar(wp, "wl_key", "");
	char *key1 = websGetVar(wp, "wl_key1", "");
	char *key2 = websGetVar(wp, "wl_key2", "");
	char *key3 = websGetVar(wp, "wl_key3", "");
	char *key4 = websGetVar(wp, "wl_key4", "");
	char wl_key[] = "wl_keyXXX";
	char *key = NULL;
				
	snprintf(wl_key, sizeof(wl_key), "wl_key%s", number);
		
	if (!strcmp(number, "1"))
		key = key1;
	else if (!strcmp(number, "2"))
		key = key2;
	else if (!strcmp(number, "3"))
		key = key3;
	else
		key = key4;	
	
	if (!key || !strcmp(key, "")) 
		return;
	if (strcmp(value,key) && !strcmp(value, "")) {
		nvram_set(v->name, value);
		return;
	}
#endif	
	switch (strlen(value)) {
	case 5:
	case 13:			
		break;
	case 10:
	case 26:
		for (c = value; *c; c++) {
			if (!isxdigit((int) *c)) {
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				websBufferWrite(wp, "<p>Invalid <b>%s</b>. The character %c is not a hexadecimal digit.</p>", v->longname, *c);
				return;
			}
		}
		break;
	default:
		//USR validation
		if (!flag)
			validation_page(wp);
		error = 1;
		websBufferWrite(wp, "<p>Invalid <b>%s</b>. The key must be 5 or 13 ASCII characters or 10 or 26 hexadecimal digits.</p>", v->longname);
		return;
	}
	
	nvram_set(v->name, value);
}


static void
validate_wl_wep(webs_t wp, char *value, struct variable *v)
{
	char *auth;
	
	if (!valid_choice(wp, value, v))
		return;

	auth = websGetVar(wp, "cbAuthEncrypt", NULL);
	//auth_mode = websGetVar(wp, "wl_auth_mode", NULL);
	if (!strcmp(value, "enabled")) {
		if (!auth || !strcmp(auth,"open") || !strcmp(auth, "shared")) {
			char wl_key[] = "wl_keyXXX";

			snprintf(wl_key, sizeof(wl_key), "wl_key%s", nvram_safe_get("wl_key"));
			if (!strlen(nvram_safe_get(wl_key))) {
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				//websBufferWrite(wp, "<p>Invalid value for <b>Network Key %s</b>. Please enter a new key.</p>", nvram_safe_get("wl_key"));
				return;
			}
		}
		else {
			nvram_set(v->name, "disabled");
			return;
		}
	}		
	else {
	    	if (!auth || !strcmp(auth, "shared")) {
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			//websBufferWrite(wp, "<p>Invalid settings. <b>WEP Encryption</b> must be <b>Enabled</b> when <b>Authentication</b> is <b>Shared Key</b>.</p>"); 
			return;
	    	}
	}
	nvram_set(v->name, value);
}


static void
validate_wl_crypto(webs_t wp, char *value, struct variable *v)
{
	if (!valid_choice(wp, value, v))
		return;
	nvram_set(v->name, value);
}

#ifdef __CONFIG_NAT__
#ifdef ROUTER
static void
validate_wan_ifname(webs_t wp, char *value, struct variable *v)
{
	char ifname[64], *next;
	foreach (ifname, nvram_safe_get("wan_ifnames"), next)
		if (!strcmp(ifname, value)) {
			nvram_set(v->name, value);
			return;
		}
	//USR validation
	if (!flag)
		validation_page(wp);
	error = 1;	
	websBufferWrite(wp, "<p>Invalid <b>%s</b>, must be one of <b>%s</b>.</p>", v->longname, nvram_safe_get("wan_ifnames"));
}
#endif
#endif	/* __CONFIG_NAT__ */
/*
static void
validate_wl_afterburner(webs_t wp, char *value, struct variable *v)
{
	char *mode, *lazy;
	
	if (!valid_choice(wp, value, v))
		return;
	
	mode = websGetVar(wp, "wl_mode", NULL);
	lazy = websGetVar(wp, "wl_lazywds", NULL);
		
	// force certain wireless variables to fixed values 
	if (!strcmp(value, "auto")) {
		if ( nvram_invmatch("wl_mode", "ap")) 
		{
			// notify the user 
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "Invalid <b>%s</b>: MAXg acceleration requires:"
				"<br><b>Bridge Mode</b> to be unchecked</b>"
				"<br><b>Fragmentation threshold</b> set to <b>2346(disable fragmentation)</b>"
				"<br><b>Please, correct the above listed settings and try again.</b>"
				"<br>", v->longname);
			return;
		}
	}
	nvram_set("cbAcceleration", "afterburner");
	nvram_set("wl_frameburst", "off");
	nvram_set("wl_frag", "2346");
	
	nvram_set(v->name, value);
}
*/
/*USR - accomodate new web design for acceleration */
static void
validate_wl_acceleration(webs_t wp, char *value, struct variable *v)
{	
	if (!valid_choice(wp, value, v))
		return;

	if (!strcmp(value, "none")) {
		nvram_set("cbAcceleration", "none");
		nvram_set("wl_frameburst", "off");
		nvram_set("wl_afterburner", "off");
		return;
	}

	if (!strcmp(value, "frameburst")) {
		nvram_set("cbAcceleration", "frameburst");
		nvram_set("wl_frameburst", "on");
		nvram_set("wl_afterburner", "off");
		return;
	}
			
	if (!strcmp(value, "afterburner")) {
		char *mode, *wme;
		mode = websGetVar(wp, "wl_mode", NULL);
		wme = websGetVar(wp, "wl_wme", NULL);
		
		if ( !strcmp(wme, "on")) 
		{
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websWrite(wp, "<p>This setting is invalid. MAXg acceleration requires the following:</p>");
			websWrite(wp, "<ul><li>WMM disabled</li><li>Fragmentation threshold set to 2346</li></ul>");
			websWrite(wp, "<p>Please, correct these settings and try again.</p>");
			return;
		}
		nvram_set("wl_afterburner", "auto");
		nvram_set("cbAcceleration", "afterburner");
		nvram_set("wl_frameburst", "on");
		nvram_set("wl_frag", "2346");
	}
}

/* USR */
static void
validate_mode(webs_t wp, char *value, struct variable *v)
{
	if (valid_choice(wp, value, v)) {
	 	if (!strcmp(value,"adhoc")) {
			nvram_set("wl_infra","0");
			nvram_set(v->name, "wet");
		}
		else {
			nvram_set("wl_infra","1");
			nvram_set(v->name, value);
		}
	}
	//printf("mode validation done\n");
}


static void
validate_wl_lazywds(webs_t wp, char *value, struct variable *v)
{
	validate_choice(wp, value, v);
}
	
static void
validate_wl_wds_hwaddrs(webs_t wp, char *value, struct variable *v)
{
	validate_list(wp, value, v, valid_hwaddr);
}

/*
static void
validate_wl_mode(webs_t wp, char *value, struct variable *v)
{
	char *afterburner;
	if (strcmp(value, "ap") &&
	    (afterburner = websGetVar(wp, "cbAcceleration", NULL)) && (!strcmp(afterburner, "auto"))) {
		//USR validation
		if (!flag)
			validation_page(wp);
		error = 1;
		websBufferWrite(wp, "Invalid <b>%s</b>: must be set to <b>Access Point</b> when AfterBurner is enabled.<br>", v->longname);
		return;
	}
	validate_choice(wp, value, v);
}
*/
/* USR added validation for WME */
static void
validate_wl_wme(webs_t wp, char *value, struct variable *v)
{
	char *afterburner;
	if (strcmp(value, "off") &&
	    (afterburner = websGetVar(wp, "cbAcceleration", NULL)) && (!strcmp(afterburner, "afterburner"))) {
		//USR: Invalid settings. WME must be set to <b>Disabled</b> when Acceleration is Afterburner.
		return;
	}
	validate_choice(wp, value, v);
}

static void
validate_noack(webs_t wp, char *value, struct variable *v)
{
	char *wme;

	/* return if wme is not enabled */
	if (!(wme = websGetVar(wp, "wl_wme", NULL)))
		return;
	else if (strcmp(wme, "on"))
		return;

	validate_choice(wp, value, v);
}
/*
static void
validate_wl_wme_params(webs_t wp, char *value, struct variable *v)
{
	int n, i;
	int cwmin = 0, cwmax = 0;
	char *wme, *afterburner;
	char name[100];
	char buf[1000] = "", *cur = buf;
	struct {
		char *name;
		int range;
		char *arg1;
		char *arg2;
	} field_attrib[] = {
		{ "WME AC CWmin", 1, "0", "32767" },
		{ "WME AC CWmax", 1, "0", "32767" },
		{ "WME AC AIFSN", 1, "1", "15" },
		{ "WME AC TXOP(b)", 1, "0", "65504" },
		{ "WME AC TXOP(a/g)", 1, "0", "65504" },
		{ "WME AC Admin Forced", 0, "on", "off" }
	};

	// return if wme is not enabled
	if (!(wme = websGetVar(wp, "wl_wme", NULL)))
		return;
	else if (strcmp(wme, "on"))
		return;

	// return if afterburner enabled 
	if ((afterburner = websGetVar(wp, "cbAcceleration", NULL)) && (!strcmp(afterburner, "afterburner")))
		return;

	n = atoi(value) + 1;

	for (i = 0; i < n; i++) {
		snprintf(name, sizeof(name), "%s%d", v->name, i);
		if (!(value = websGetVar(wp, name, NULL)))
			return;
		if (!*value && v->nullok)
			continue;

		if (i == 0)
			cwmin = atoi(value);
		else if (i == 1) {
			cwmax = atoi(value);
			if (cwmax < cwmin) {
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				websBufferWrite(wp, "<p>Invalid <b>%s</b>. %d is greater than <b>%s</b> %d.</p>",
					field_attrib[0].name, cwmin, field_attrib[i].name, cwmax);
				return;
			}
		}
		if (field_attrib[i].range) {
			if (atoi(value) < atoi(field_attrib[i].arg1) || atoi(value) > atoi(field_attrib[i].arg2)) {
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				websBufferWrite(wp, "<p>Invalid <b>%s</b>. %d should be in the range %s to %s.</p>",
					field_attrib[i].name, atoi(value), field_attrib[i].arg1, field_attrib[i].arg2);
				return;
			}
		} else {
			if (strcmp(value, field_attrib[i].arg1) && strcmp(value, field_attrib[i].arg2)) {
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				websBufferWrite(wp, "<p>Invalid <b>%s</b>. %s should be %s or %s.</p>",
					field_attrib[i].name, value, field_attrib[i].arg1, field_attrib[i].arg2);
			}
		}

		cur += snprintf(cur, buf + sizeof(buf) - cur, "%s%s",
				cur == buf ? "" : " ", value);
	}

	nvram_set(v->name, buf);
}
*/
static void 
validate_power(webs_t wp, char *value, struct variable *v)
{
	int unit;
	char tmp[100], prefix[] = "wlXXXXXXXXXX_";
	char *name;
	char *level;

	if ((unit = atoi(nvram_safe_get("wl_unit"))) < 0) {
		websError(wp, 400, "unit number variable doesn't exist\n");
		return;
	}

	snprintf(prefix, sizeof(prefix), "wl%d_", unit);
	name = nvram_safe_get(strcat_r(prefix, "ifname", tmp));
	level = nvram_safe_get("wl_power");
	printf("power_level: %s\n", level);
	if (strcmp(level, value)) {
		nvram_set(v->name, value);
	}
}
//pptp implementation
#ifdef ROUTER
static void
validate_pptp_activated(webs_t wp, char *value, struct variable *v)
{
	char tmp[128];
    
	if ((value = websGetVar(wp, "chkUsePPTP", NULL)))	
    	{	
		nvram_set("wan_pptp_activated","1");
		
		if ((value = websGetVar(wp, "pptp_pap", NULL))) 
		    strcpy(tmp, "1");
		else 
		    strcpy(tmp, "0");
				
    		if ((value = websGetVar(wp, "pptp_chap", NULL))) 
		    strcat(tmp, " 1");
		else 
		    strcat(tmp, " 0");
				
    		if ((value = websGetVar(wp, "pptp_mschap", NULL))) 
		    strcat(tmp, " 1");
		else 
		    strcat(tmp, " 0");
		
    		if ((value = websGetVar(wp, "pptp_mschapv2", NULL))) 
		    strcat(tmp, " 1");
		else 
		    strcat(tmp, " 0");
		 		
		nvram_set("wan_pptp_auth", tmp);
	} else {
		nvram_set("wan_pptp_activated","0");
		//cprintf("  pptp_Auth=off\n");
	}
}
/*
static void
validate_mppe(webs_t wp, char *value, struct variable *v)
{
	 char tmp[128];
    
    if ((value = websGetVar(wp, "wan_pptp_activated", NULL)))	
    {	
	if ((value = websGetVar(wp, "pptp_mppe", NULL))) 
	{
		cprintf("  pptp_Encrypt=%s\n",value);
    		if ((value = websGetVar(wp, "pptp_mppe40", NULL))) 
		{
		    strcpy(tmp, "1");
		   // cprintf("  pptp_mppe40=%s\n",value);
		} else {
		    strcpy(tmp, "0");
		    //cprintf("  pptp_mppe40=off\n");
		}
    		if ((value = websGetVar(wp, "pptp_mppe128", NULL))) 
		{
		    strcat(tmp, " 1");
		    //cprintf("  pptp_mppe128=%s\n",value);
		} else {
		    strcat(tmp, " 0");
		    //cprintf("  pptp_mppe128=off\n");
		}
		nvram_set("wan_pptp_mppe", tmp);
	    } else {
		nvram_set("wan_pptp_mppe", "0 0");
		cprintf("  pptp_mppe=off\n");
	}
    }
} */
#endif
	
/* Hook to write wl_* default set through to wl%d_* variable set */
static void
wl_unit(webs_t wp, char *value, struct variable *v)
{
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";

	/* Do not write through if no interfaces are present */
	if (atoi(value) < 0)
		return;
	//printf("wl_unit: %s\n", value);
	/* Set prefix */
	snprintf(prefix, sizeof(prefix), "wl%d_", atoi(value));

	/* Write through to selected variable set */
	for (; v >= variables && !strncmp(v->name, "wl_", 3); v--)
		nvram_set(strcat_r(prefix, &v->name[3], tmp), nvram_safe_get(v->name));
}

#ifdef __CONFIG_NAT__
#ifdef ROUTER
static void
wan_primary(webs_t wp)
{
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";
	int i;
	for (i = 0; i < MAX_NVPARSE; i ++) {
		/* skip non-exist and disabled connection */
		wan_prefix(i, prefix);
		if (!nvram_get(strcat_r(prefix, "unit", tmp))||
		    nvram_match(strcat_r(prefix, "proto", tmp), "disabled"))
			continue;
		/* make connection <i> primary */
		nvram_set(strcat_r(prefix, "primary", tmp), "1");
		/* notify the user */
		//websBufferWrite(wp, "<br><b>%s</b> is set to primary.",
		//	wan_name(i, prefix, tmp, sizeof(tmp)));
		break;
	}
}

/* Hook to write wan_* default set through to wan%d_* variable set */
static void
wan_unit(webs_t wp, char *value, struct variable *v)
{
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";
	char pppx[] = "pppXXXXXXXXXXX";
	int unit, i;
	char *wan_ifname;
	int wan_disabled = 0;
	int wan_prim = 0;
	int wan_wildcard = 0;
	char *wan_pppoe_service;
	char *wan_pppoe_ac;
	char wan_tmp[NVRAM_BUFSIZE];
	int wildcard;
	char *pppoe_service;
	char *pppoe_ac;
	char ea[ETHER_ADDR_LEN], wan_ea[ETHER_ADDR_LEN];

	/* Do not write through if no connections are present */
	if ((unit = atoi(value)) < 0)
		return;

	/* override wan_pppoe_ifname */
	if (nvram_match("wan_proto", "pppoe")) {
		snprintf(pppx, sizeof(pppx), "ppp%d", unit);
		nvram_set("wan_pppoe_ifname", pppx);
	}
	
	//printf("wan_unit: %s \n", value);
	/*
	* Need to make sure this connection can co-exist with others.
	* Disable others if it can't (assuming this is the wanted one).
	* Disabled connection is for sure no problem to co-exist with
	* other connections.
	*/
	if (nvram_match("wan_proto", "disabled")) {
		/* Non primary always go with disabled connection. */
		nvram_set("wan_primary", "0");
		wan_disabled = 1;
	}
	/*
	* PPPoE connection is for sure no problem to co-exist with
	* other PPPoE connections even when they share the same 
	* ethernet interface, but we need to make sure certain 
	* PPPoE parameters are reasonablely different from eatch other 
	* if they share the same ethernet interface.
	*/
	else if (nvram_match("wan_proto", "pppoe")) {
		/* must disable others if this connection is wildcard (any service any ac) */
		wan_pppoe_service = nvram_get("wan_pppoe_service");
		wan_pppoe_ac = nvram_get("wan_pppoe_ac");
		wan_wildcard = (wan_pppoe_service == NULL || *wan_pppoe_service == 0) &&
			(wan_pppoe_ac == NULL || *wan_pppoe_ac == 0);
		wan_ifname = nvram_safe_get("wan_ifname");
		wan_name(unit, "wan_", wan_tmp, sizeof(wan_tmp));
		//printf("wan check: %s\n", wan_tmp);
		/* check all PPPoE connections that share the same interface */
		for (i = 0; i < MAX_NVPARSE; i ++) {
			/* skip the current connection */
			if (i == unit)
				continue;
			/* skip non-exist and connection that does not share the same i/f */
			wan_prefix(i, prefix);
			if (!nvram_get(strcat_r(prefix, "unit", tmp)) ||
			    nvram_match(strcat_r(prefix, "proto", tmp), "disabled") ||
			    nvram_invmatch(strcat_r(prefix, "ifname", tmp), wan_ifname))
				continue;
			/* PPPoE can share the same i/f, but none can be wildcard */
			if (nvram_match(strcat_r(prefix, "proto", tmp), "pppoe")) {
				if (wan_wildcard) {
					/* disable connection <i> */
					nvram_set(strcat_r(prefix, "proto", tmp), "disabled");
					nvram_set(strcat_r(prefix, "primary", tmp), "0");
					/* notify the user */
					//USR validation
					if (!flag)
						validation_page(wp);
					error = 1;
					websBufferWrite(wp, "<br><b>%s</b> is <b>disabled</b> because both "
						"<b>PPPoE Service Name</b> and <b>PPPoE Access Concentrator</b> "
						"in <b>%s</b> are empty.",
						wan_name(i, prefix, tmp, sizeof(tmp)), wan_tmp);
				}
				else {
					pppoe_service = nvram_get(strcat_r(prefix, "pppoe_service", tmp));
					pppoe_ac = nvram_get(strcat_r(prefix, "pppoe_ac", tmp));
					wildcard = (pppoe_service == NULL || *pppoe_service == 0) &&
						(pppoe_ac == NULL || *pppoe_ac == 0);
					/* allow connection <i> if certain pppoe parameters are not all same */
					if (!wildcard &&
					    (nvram_invmatch(strcat_r(prefix, "pppoe_service", tmp), nvram_safe_get("wan_pppoe_service")) ||
					         nvram_invmatch(strcat_r(prefix, "pppoe_ac", tmp), nvram_safe_get("wan_pppoe_ac"))))
						continue;
					/* disable connection <i> */
					nvram_set(strcat_r(prefix, "proto", tmp), "disabled");
					nvram_set(strcat_r(prefix, "primary", tmp), "0");
					/* notify the user */
					//USR validation
					if (!flag)
						validation_page(wp);
					error = 1;
					websBufferWrite(wp, "<br><b>%s</b> is <b>disabled</b> because both its "
						"<b>PPPoE Service Name</b> and <b>PPPoE Access Concentrator</b> "
						"are empty.",
						wan_name(i, prefix, tmp, sizeof(tmp)));
				}
			}
			/* other types can't (?) share the same i/f with PPPoE */
			else {
				/* disable connection <i> */
				nvram_set(strcat_r(prefix, "proto", tmp), "disabled");
				nvram_set(strcat_r(prefix, "primary", tmp), "0");
				/* notify the user */
				//USR validation
				if (!flag)
					validation_page(wp);
				error = 1;
				websBufferWrite(wp, "<br><b>%s</b> is <b>disabled</b> because it can't  "
					"share the same interface with <b>%s</b>.",
					wan_name(i, prefix, tmp, sizeof(tmp)), wan_tmp);
			}
		}
	}
	/* 
	* All other types (now DHCP, Static) can't co-exist with 
	* other connections if they use the same ethernet i/f.
	*/
	else {
		wan_ifname = nvram_safe_get("wan_ifname");
		wan_name(unit, "wan_", wan_tmp, sizeof(wan_tmp));
		/* check all connections that share the same interface */
		for (i = 0; i < MAX_NVPARSE; i ++) {
			/* skip the current connection */
			if (i == unit)
				continue;
			/* check if connection <i> exists and share the same i/f*/
			wan_prefix(i, prefix);
			if (!nvram_get(strcat_r(prefix, "unit", tmp)) ||
			    nvram_match(strcat_r(prefix, "proto", tmp), "disabled") ||
			    nvram_invmatch(strcat_r(prefix, "ifname", tmp), wan_ifname))
				continue;
			/* disable connection <i> */
			nvram_set(strcat_r(prefix, "proto", tmp), "disabled");
			nvram_set(strcat_r(prefix, "primary", tmp), "0");
			/* notify the user */
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<p><b>%s</b> is disabled because it can't share "
				"the ethernet interface with <b>%s</b>.</p>",
				wan_name(i, prefix, tmp, sizeof(tmp)), wan_tmp);
		}
	}

	/*
	* Check if MAC address has been changed. Need to sync it to all connections
	* that share the same i/f if it is changed.
	*/
	wan_prefix(unit, prefix);
	ether_atoe(nvram_safe_get("wan_hwaddr"), wan_ea);
	ether_atoe(nvram_safe_get(strcat_r(prefix, "hwaddr", tmp)), ea);
	if (memcmp(ea, wan_ea, ETHER_ADDR_LEN)) {
		wan_ifname = nvram_safe_get("wan_ifname");
		wan_name(unit, "wan_", wan_tmp, sizeof(wan_tmp));
		/* sync all connections that share the same interface */
		for (i = 0; i < MAX_NVPARSE; i ++) {
			/* skip the current connection */
			if (i == unit)
				continue;
			/* check if connection <i> exists and share the same i/f*/
			wan_prefix(i, prefix);
			if (!nvram_get(strcat_r(prefix, "unit", tmp)) ||
			    nvram_invmatch(strcat_r(prefix, "ifname", tmp), wan_ifname))
				continue;
			/* check if connection <i>'s hardware address is different */
			if (ether_atoe(nvram_safe_get(strcat_r(prefix, "hwaddr", tmp)), ea) &&
			    !memcmp(ea, wan_ea, ETHER_ADDR_LEN))
			    continue;
			/* change connection <i>'s hardware address */
			nvram_set(strcat_r(prefix, "hwaddr", tmp), nvram_safe_get("wan_hwaddr"));
			/* notify the user */
			//USR validation
			if (!flag)
				validation_page(wp);
			error = 1;
			websBufferWrite(wp, "<br><b>MAC Address</b> in <b>%s</b> is changed to "
				"<b>%s</b> because it shares the ethernet interface with <b>%s</b>.",
				wan_name(i, prefix, tmp, sizeof(tmp)), nvram_safe_get("wan_hwaddr"),
				wan_tmp);
		}
	}

	/* Set prefix */
	wan_prefix(unit, prefix);

	/* Write through to selected variable set */
	for (; v >= variables && !strncmp(v->name, "wan_", 4); v--)
		nvram_set(strcat_r(prefix, &v->name[4], tmp), nvram_safe_get(v->name));

	/*
	* There must be one and only one primary connection among all 
	* enabled connections so that traffic can be routed by default 
	* through the primary connection unless they are targetted to
	* a specific connection by means of static routes. (Primary ~=
	* Default Gateway).
	*/
	/* the current connection is primary, set others to non-primary */
	if (!wan_disabled && nvram_match(strcat_r(prefix, "primary", tmp), "1")) {
		/* set other connections to non-primary */
		for (i = 0; i < MAX_NVPARSE; i ++) {
			/* skip the current connection */
			if (i == unit)
				continue;
			/* skip non-exist and disabled connection */
			wan_prefix(i, prefix);
			if (!nvram_get(strcat_r(prefix, "unit", tmp)) ||
			    nvram_match(strcat_r(prefix, "proto", tmp), "disabled"))
				continue;
			/* skip non-primary connection */
			if (nvram_invmatch(strcat_r(prefix, "primary", tmp), "1"))
				continue;
			/* force primary to non-primary */
			nvram_set(strcat_r(prefix, "primary", tmp), "0");
			/* notify the user */
			//websBufferWrite(wp, "<br><b>%s</b> is set to non-primary.",
			//	wan_name(i, prefix, tmp, sizeof(tmp)));
		}
		wan_prim = 1;
	}
	/* the current connection is not parimary, check if there is any primary */
	else {
		/* check other connections to see if there is any primary */
		for (i = 0; i < MAX_NVPARSE; i ++) {
			/* skip the current connection */
			if (i == unit)
				continue;
			/* primary connection exists, honor it */
			wan_prefix(i, prefix);
			if (nvram_match(strcat_r(prefix, "primary", tmp), "1")) {
				wan_prim = 1;
				break;
			}
		}
	}
	/* no one is primary, pick the first enabled one as primary */
	if (!wan_prim)
		wan_primary(wp);
}
#endif /* ROUTER */
#endif	/* __CONFIG_NAT__ */

static void
time_zone(webs_t wp, char *value, struct variable *v)
{
	nvram_set(v->name, value);
}

/* This is the monster V-block
 * 
 * It controls the following functions
 *
 * Configuration variables are validated and saved in NVRAM
 * Variables saved by NVRAM save/restore routine
 * Method in which the variables are to be validates
 * SES handling
 *
 * The control flags:
 * EZC_FLAGS_READ,EZC_FLAGS_WRITE :Ses read/write flags
 * NVRAM_ENCRYPT: Encrypt variable prior to downloading NVRAM variable to file
 * NVRAM_MI: Multi instance NVRAM variable eg wlXX,wanXX
 * NVRAM_VLAN_MULTI: Special flag to handle oddball vlanXXname and its cousins
 * NVRAM_MP: Variable can be single & multi instance eg lan_ifname, lanX_ifname
 * NVRAM_IGNORE: Dont save/restore this var.
 * WEB_IGNORE: Don't validate or process this var during web validation.
 * 
 * Below is the definition of the structure
 *struct variable {
 *	char *name; <- name of variable
 * 	char *longname; <- display name
 * 	char *prefix; <- prefix for processing the multi-instance versions
 *	void (*validate)(); <- Validation routine
 * 	char **argv; <- Optional argument vector for validation routine
 *	int nullok; <- value can be NULL
 *	int ezc_flags; <- control flags
 * };
 *
 * IMPORTANT:
 * =========
 *
 * The variables in tne table below determine if they will be saved/restored 
 * by the UI NVRAM save/restore feature
 *
 * If an NVRAM variable is not present in this list it will
 * not be processed and thus will not be saved or restored.
 *
*/

/* 
 * Variables are set in order (put dependent variables later). Set
 * nullok to TRUE to ignore zero-length values of the variable itself.
 * For more complicated validation that cannot be done in one pass or
 * depends on additional form components or can throw an error in a
 * unique painful way, write your own validation routine and assign it
 * to a hidden variable (e.g. filter_ip).
 *
 * EZC_FLAGS_READ : implies the variable will be returned for ezconfig read request
 * EZC_FLAGS_WRITE : allows the variable to be modified by the ezconfig tool
 *
 * The variables marked with EZConfig have to maintain backward compatibility.
 * If they cannot then the ezc_version has to be bumped up
 *
 */
static char wl_prefix[]="wl";
static char wan_prefix[]="wan";
static char lan_prefix[]="lan";
static char dhcp_prefix[]="dhcp";
static char vlan_prefix[]="vlan";

static struct hsearch_data vtab;

struct variable variables[] = {
	/* basic settings */
	{ "http_username", "Username", NULL, validate_name, ARGV("0", "63"), TRUE, EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "http_passwd", "Password", NULL, validate_name, ARGV("0", "63"), TRUE, NVRAM_ENCRYPT | EZC_FLAGS_WRITE },
	{ "http_wanport", "WAN Port", NULL, validate_range, ARGV("0", "65535"), TRUE, 0 },
	{ "http_lanport", "HTTP daemon lanport", NULL, NULL, NULL, TRUE, 0 },
	{ "router_disable", "Mode", NULL, validate_choice, ARGV("0", "1"), FALSE, 0 },
	{ "fw_disable", "Firewall", NULL, validate_choice, ARGV("0", "1"), FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "time_zone", "Time Zone", NULL, time_zone, NULL, TRUE, 0 },
	{ "upnp_enable", "UPnP", NULL, validate_choice, ARGV("0", "1"), FALSE, 0 },
	{ "ezc_enable", "EZConfig", NULL, validate_choice, ARGV("0", "1"), FALSE, EZC_FLAGS_READ | EZC_FLAGS_WRITE},
	{ "ntp_server", "NTP Servers", NULL, validate_ipaddrs, NULL, TRUE, 0 },
	{ "log_level", "Connection Logging", NULL, validate_range, ARGV("0", "3"), FALSE, 0 },
	{ "log_ipaddr", "Log LAN IP Address", NULL, validate_ipaddr, ARGV("lan_ipaddr", "lan_netmask"), TRUE, 0 },
	{ "log_ram_enable", "Syslog in RAM", NULL, validate_choice, ARGV("0", "1"), FALSE, 0 },
	/* LAN settings */
	{ "lan_ifname" "LAN Interface Name", lan_prefix, NULL, NULL, FALSE, NVRAM_MI | NVRAM_IGNORE },
	{ "lan_dhcp", "DHCP Client", lan_prefix, validate_choice, ARGV("1", "0"), FALSE, NVRAM_MI | NVRAM_MP },
	{ "lan_ipaddr", "IP Address", lan_prefix, validate_lan_ipaddr, NULL, FALSE, NVRAM_MI | NVRAM_MP | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "lan_netmask", "Subnet Mask", lan_prefix, validate_ipaddr, NULL, TRUE, NVRAM_MI | NVRAM_MP },
	{ "lan_gateway", "Gateway Address", lan_prefix, validate_ipaddr, NULL, TRUE, NVRAM_MI | NVRAM_MP },
	{ "lan_proto", "DHCP Server", lan_prefix, validate_choice, ARGV("dhcp", "static"), FALSE, NVRAM_MI | NVRAM_MP | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "dhcp_start", "DHCP Server LAN IP Address Range", dhcp_prefix, validate_dhcp, NULL, FALSE, NVRAM_MI | NVRAM_MP | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "dhcp_end", "DHCP Server LAN IP End Address", dhcp_prefix, validate_ipaddr, NULL, TRUE, NVRAM_MI | NVRAM_MP },
	{ "dhcp_wins", "DHCP WINS domain", dhcp_prefix, NULL, NULL, TRUE, NVRAM_MI | NVRAM_MP},
	{ "dhcp_domain", "DHCP domain", dhcp_prefix, NULL, NULL, TRUE, NVRAM_MI | NVRAM_MP },
	{ "lan_lease", "DHCP Server Lease Time", lan_prefix, validate_range, ARGV("1", "604800"), FALSE,  NVRAM_MI | NVRAM_MP | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "lan_stp", "Spanning Tree Protocol", lan_prefix, validate_choice, ARGV("0", "1"), FALSE, NVRAM_MI | NVRAM_MP | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "lan_route", "Static Routes", lan_prefix, validate_lan_route, NULL, FALSE, NVRAM_MI | NVRAM_MP },
	{ "lan_wins", "Lan WINS", lan_prefix, NULL, NULL, TRUE, NVRAM_MI | NVRAM_MP},
	{ "lan_domain", "Lan Domain", lan_prefix,NULL, NULL, TRUE,  NVRAM_MI | NVRAM_MP },
	//{ "lan_route","LAN route", lan_prefix, NULL, NULL, TRUE,  NVRAM_MI| NVRAM_MP },
	//{ "lan_guest_ifname","Guest LAN 1", lan_prefix, validate_guest_lan_ifname, ARGV("1"), FALSE,  NVRAM_MI| NVRAM_IGNORE},
	/*VLAN config vars. These are multi-instance Used by NVRAM save/restore for the moment */
	{ "vlanhwname", "VLAN HW name", vlan_prefix, NULL, NULL, TRUE, NVRAM_VLAN_MULTI},
	{ "vlanports", "VLAN Ports", vlan_prefix, NULL, NULL, TRUE, NVRAM_VLAN_MULTI},
#ifdef __CONFIG_NAT__
#ifdef ROUTER	
	/* ALL wan_XXXX variables below till wan_unit variable are per-interface */
	{ "wan_desc", "Description", wan_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_proto", "Protocol", wan_prefix, validate_choice, ARGV("dhcp", "static", "pppoe", "disabled"), FALSE,NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_hostname", "Host Name", wan_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_domain", "Domain Name", wan_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_ifname", "Interface Name", wan_prefix, validate_wan_ifname, NULL, TRUE, NVRAM_MI },
	{ "wan_hwaddr", "MAC Address", wan_prefix, validate_hwaddr, NULL, TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_ipaddr", "IP Address", wan_prefix, validate_ipaddr, NULL, FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_netmask", "Subnet Mask", wan_prefix, validate_ipaddr, NULL, FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_gateway", "Default Gateway", wan_prefix, validate_ipaddr, NULL, TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_dns", "DNS Servers", wan_prefix, validate_ipaddrs, NULL, TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_wins", "WINS Servers", wan_prefix, validate_ipaddrs, NULL, TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_ifname", "PPPoE Interface Name", wan_prefix, NULL, NULL, TRUE, NVRAM_MI },
	{ "wan_pppoe_username", "PPPoE Username", wan_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_passwd", "PPPoE Password", wan_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | EZC_FLAGS_WRITE },
	{ "wan_pppoe_service", "PPPoE Service Name", wan_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_ac", "PPPoE Access Concentrator", wan_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_keepalive", "PPPoE Keep Alive", wan_prefix, validate_choice, ARGV("0", "1"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_demand", "PPPoE Connect on Demand", wan_prefix, validate_choice, ARGV("0", "1"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_idletime", "PPPoE Max Idle Time", wan_prefix, validate_range, ARGV("1", "3600"), TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_mru", "PPPoE MRU", wan_prefix, validate_range, ARGV("128", "16384"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_pppoe_mtu", "PPPoE MTU", wan_prefix, validate_range, ARGV("128", "16384"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wan_primary", "Primary Interface", wan_prefix, validate_choice, ARGV("0", "1"), FALSE, NVRAM_MI },
	{ "wan_route", "Static Routes", wan_prefix, validate_wan_route, NULL, FALSE, NVRAM_MI },
	{ "wan_pptp_ifname", "PPTP Interface Name", NULL, NULL, NULL, TRUE, 0 },
	{ "wan_pptp_username", "PPTP Username", NULL, validate_name, ARGV("0", "255"), TRUE, 0 },
	{ "wan_pptp_passwd", "PPTP Password", NULL, validate_name, ARGV("0", "255"), TRUE, 0 },
	{ "wan_pptp_server", "PPTP Server Name or IP", NULL, NULL, NULL, TRUE, 0 },
	{ "wan_pptp_activated", "PPTP Disabled/Enabled", NULL, validate_pptp_activated, NULL, TRUE, 0 },
	{ "wan_pptp_mru", "PPTP MRU", NULL, validate_range, ARGV("128", "16384"), FALSE, 0 },
	{ "wan_pptp_mtu", "PPTP MTU", NULL, validate_range, ARGV("128", "16384"), FALSE, 0 },
	//{ "wan_pptp_mppe", "PPTP Encryption", NULL, validate_mppe, NULL, TRUE, 0 },
	//{ "wan_pptp_mppe_mode", "PPPoE Max Idle Time", NULL, validate_mppe_mode, NULL, TRUE, 0},
	/* MUST leave this entry here after all wl_XXXX per-interface variables */
	{ "wan_unit", "WAN Instance", wan_prefix, wan_unit, NULL, TRUE, NVRAM_MI | NVRAM_MP },
	/* USR - accomodate new web pages dessign */
	{ "connection", "PPPoE Connection", NULL, validate_connection, ARGV("timeout", "demand", "keepalive"), FALSE, 0},
	/* filter settings */
	{ "filter_macmode", "MAC Filter Mode", NULL, validate_choice, ARGV("disabled", "allow", "deny"), FALSE, 0 },
	{ "filter_maclist", "MAC Filter", NULL, validate_hwaddrs, NULL, TRUE, 0 },
	{ "filter_client", "LAN Client Filter", NULL, validate_filter_client, ARGV("0", XSTR(MAX_NVPARSE - 1)), FALSE, NVRAM_GENERIC_MULTI },
	/* routing settings */
	{ "forward_port", "Port Forward", NULL, validate_forward_port, ARGV("0", XSTR(MAX_NVPARSE - 1)), FALSE, NVRAM_GENERIC_MULTI },
	{ "autofw_port", "Application Specific Port Forward", NULL, validate_autofw_port, ARGV("0", XSTR(MAX_NVPARSE - 1)), FALSE, NVRAM_GENERIC_MULTI },
	{ "dmz_ipaddr", "DMZ LAN IP Address", NULL, validate_ipaddr, ARGV("lan_ipaddr", "lan_netmask"), TRUE, 0 },
#endif   /* ROUTER */
#endif	/* __CONFIG_NAT__ */
	/* USR - accomodate new web pages dessign - don't change position*/
	{ "cbAcceleration", "Acceleration", NULL, validate_wl_acceleration, ARGV("none", "frameburst", "afterburner"), FALSE, 0 },
	{ "cbAuthEncrypt", "Authentication", NULL, validate_cbAuthEncrypt, ARGV("none", "wpa12", "wpa2", "wpa", "open", "shared", "radius-wpa12", "radius-wpa2", "radius-wpa", "radius-wep"), FALSE, 0 },
	{ "cbEncryption", "Encryption", NULL, validate_cbEncryption, ARGV("none", "tkip", "aes", "tkip+aes", "wl_wep_open", "wl_wep_shared"), FALSE, 0 },
	/* ALL wl_XXXX variables are per-interface */
	{ "wl_ssid", "Network Name (ESSID)", wl_prefix, validate_name, ARGV("0", "32"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	//{ "wl_guest_ssid_1", "Guest Network Name 1(SSID)", wl_prefix, validate_guest_ssid, ARGV("1"), FALSE, NVRAM_MI|NVRAM_IGNORE },
	//{ "wl_guest_ssid_2", "Guest Network Name 2(SSID)", wl_prefix, validate_guest_ssid, ARGV("2"), FALSE, NVRAM_MI|NVRAM_IGNORE },
	//{ "wl_guest_ssid_3", "Guest Network Name 3(SSID)", wl_prefix, validate_guest_ssid, ARGV("3"), FALSE, NVRAM_MI|NVRAM_IGNORE },
	//{ "wl_guest_ssid_4", "Guest Network Name 4(SSID)", wl_prefix, validate_guest_ssid, ARGV("4"), FALSE, NVRAM_MI|NVRAM_IGNORE },
	//{ "wl_ure", "URE Mode",NULL, validate_ure, ARGV("0"), FALSE, NVRAM_MI|NVRAM_IGNORE },
	{ "wl_closed", "Network Type", wl_prefix, validate_choice, ARGV("0", "1"), FALSE,NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
        { "wl_ap_isolate", "AP Isolate", wl_prefix, validate_choice, ARGV("0", "1"), FALSE, NVRAM_MI | 0 },
	//{ "wl_country_code", "Country Code", wl_prefix, validate_country, NULL, FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_mode", "Mode", wl_prefix, validate_mode, ARGV("ap", "wds", "sta", "wet", "adhoc", "apsta"), FALSE, NVRAM_MI },
	{ "wl_infra", "Network", wl_prefix, validate_choice, ARGV("0", "1"), FALSE, NVRAM_MI },
	{ "wl_lazywds", "Bridge Restrict", wl_prefix, validate_wl_lazywds, ARGV("0", "1"), FALSE, NVRAM_MI },
	{ "wl_wds", "Bridges", wl_prefix, validate_wl_wds_hwaddrs, NULL, TRUE, NVRAM_MI },
	{ "wl_wds_timeout", "Link Timeout Interval", wl_prefix, NULL, NULL, TRUE, NVRAM_MI },
	{ "wl_macmode", "MAC Restrict Mode", wl_prefix, validate_choice, ARGV("disabled", "allow", "deny"), FALSE, NVRAM_MI },
	{ "wl_maclist", "Allowed MAC Address", wl_prefix, validate_hwaddrs, NULL, TRUE, NVRAM_MI },
	{ "wl_radio", "Radio Enable", wl_prefix, validate_choice, ARGV("0", "1"), FALSE, NVRAM_MI },
	{ "wl_phytype", "Radio Band", wl_prefix, validate_choice, ARGV("a", "b", "g"), TRUE, NVRAM_MI },
	{ "wl_antdiv", "Antenna Diversity", wl_prefix, validate_choice, ARGV("-1", "0", "1", "3"), FALSE, NVRAM_MI },
	/* Channel and rate are fixed in wlconf() if incorrect */
	{ "wl_channel", "Channel", wl_prefix, validate_range, ARGV("0", "216"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_reg_mode", "Regulatory Mode", wl_prefix, validate_choice, ARGV("off", "h", "d"), FALSE, NVRAM_MI },
	{ "wl_dfs_preism", "Pre-Network Radar Check", wl_prefix, validate_range, ARGV("0", "99"), FALSE, NVRAM_MI },
	{ "wl_dfs_postism", "In Network Radar Check", wl_prefix, validate_range, ARGV("10", "99"), FALSE, NVRAM_MI },
	{ "wl_tpc_db", "TPC Mitigation (db)",  wl_prefix, validate_range, ARGV("0", "99"), FALSE, NVRAM_MI },
	{ "wl_rate", "Rate", wl_prefix, validate_range, ARGV("0", "54000000"), FALSE, NVRAM_MI },
	{ "wl_rateset", "Supported Rates", wl_prefix, validate_choice, ARGV("all", "default", "12"), FALSE, NVRAM_MI },
	{ "wl_mrate", "Multicast Rate", wl_prefix, validate_range, ARGV("0", "54000000"), FALSE, NVRAM_MI },
	{ "wl_frag", "Fragmentation Threshold", wl_prefix, validate_range, ARGV("256", "2346"), FALSE, NVRAM_MI },
	{ "wl_rts", "RTS Threshold", wl_prefix, validate_range, ARGV("0", "2347"), FALSE, NVRAM_MI },
	{ "wl_dtim", "DTIM Period", wl_prefix, validate_range, ARGV("1", "255"), FALSE, NVRAM_MI },
	{ "wl_bcn", "Beacon Interval", wl_prefix, validate_range, ARGV("1", "65535"), FALSE, NVRAM_MI },
	{ "wl_plcphdr", "Preamble Type", wl_prefix, validate_choice, ARGV("long", "short"), FALSE, NVRAM_MI },
	{ "wl_maxassoc", "Max Assocation Limit", wl_prefix, validate_range, ARGV("1", "256"), FALSE, NVRAM_MI },
	{ "wl_gmode", "54g Mode", wl_prefix, validate_choice, ARGV(XSTR(GMODE_AUTO), XSTR(GMODE_ONLY), XSTR(GMODE_PERFORMANCE), XSTR(GMODE_LRS), XSTR(GMODE_LEGACY_B)), FALSE, NVRAM_MI | 0 },
	{ "wl_gmode_protection", "54g Protection", wl_prefix, validate_choice, ARGV("off", "auto"), FALSE, NVRAM_MI },
	{ "wl_frameburst", "XPress Technology", wl_prefix, validate_choice, ARGV("off", "on"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_afterburner", "AfterBurner Technology", wl_prefix, validate_wl_acceleration, ARGV("off", "auto"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_wme", "WMM Support", wl_prefix, validate_wl_wme, ARGV("off", "on"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_wme_no_ack", "No-Acknowledgement", wl_prefix, validate_noack, ARGV("off", "on"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	//{ "wl_wme_ap_bk", "WME AP BK", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	//{ "wl_wme_ap_be", "WME AP BE", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	//{ "wl_wme_ap_vi", "WME AP VI", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	//{ "wl_wme_ap_vo", "WME AP VO", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	//{ "wl_wme_sta_bk", "WME STA BK", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	//{ "wl_wme_sta_be", "WME STA BE", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	//{ "wl_wme_sta_vi", "WME STA VI", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	//{ "wl_wme_sta_vo", "WME STA VO", wl_prefix, validate_wl_wme_params, NULL, TRUE, NVRAM_MI },
	/* security parameters */
	{ "wl_key", "Network Key Index", wl_prefix, validate_key, ARGV("1", "4"), FALSE, NVRAM_MI | EZC_FLAGS_WRITE },
#ifdef CLIENT
	{ "wl_key1", "Network Key 1", wl_prefix, validate_wl_key, NULL, TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key2", "Network Key 2", wl_prefix, validate_wl_key, NULL, TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key3", "Network Key 3", wl_prefix, validate_wl_key, NULL, TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key4", "Network Key 4", wl_prefix, validate_wl_key, NULL, TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#endif
#ifdef ROUTER
	{ "wl_key1", "Network Key 1", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key2", "Network Key 2", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key3", "Network Key 3", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key4", "Network Key 4", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#endif
#ifdef AP
	{ "wl_key1", "Network Key 1", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key2", "Network Key 2", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key3", "Network Key 3", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key4", "Network Key 4", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#endif
#ifdef APB
	{ "wl_key1", "Network Key 1", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key2", "Network Key 2", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key3", "Network Key 3", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
	{ "wl_key4", "Network Key 4", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#endif
#ifdef REPEATER
	{ "wl_key1", "Key", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#endif
#ifdef GAMING
	{ "wl_key1", "Key", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#endif
#ifdef BRIDGE
	{ "wl_key1", "Key", wl_prefix, validate_wl_key, ARGV("5", "26"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#endif	
	{ "wl_auth", "802.11 Authentication", wl_prefix, validate_wl_auth, ARGV("0", "1"), FALSE, NVRAM_MI | 0},
	{ "wl_auth_mode", "Network Authentication", wl_prefix, validate_wl_auth_mode, ARGV("radius", "none"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_akm", "Authenticated Key Management", wl_prefix, validate_wl_akm, NULL, FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_wep", "WEP Encryption", wl_prefix, validate_wl_wep, ARGV("disabled", "enabled"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_crypto", "WPA Encryption", wl_prefix, validate_wl_crypto, ARGV("tkip", "aes", "tkip+aes"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
//#ifdef BCMWPA2
	{ "wl_net_reauth", "Network Re-auth Interval", wl_prefix, NULL, NULL, TRUE, NVRAM_MI  },
	{ "wl_preauth", "Network Preauthentication Support", wl_prefix, validate_wl_preauth, ARGV("disabled", "enabled"), FALSE, NVRAM_MI | 0 },
//#endif
	{ "wl_radius_ipaddr", "RADIUS Server", wl_prefix, validate_ipaddr, NULL, TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_radius_port", "RADIUS Port", wl_prefix, validate_range, ARGV("0", "65535"), FALSE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	{ "wl_radius_key", "RADIUS Shared Secret", wl_prefix, validate_name, ARGV("0", "255"), TRUE, NVRAM_MI | NVRAM_ENCRYPT | EZC_FLAGS_WRITE},
#ifdef CLIENT	
	{ "wl_wpa_psk", "WPA Pre-Shared Key", wl_prefix, validate_wl_wpa_psk, ARGV("8", "64"), TRUE, NVRAM_ENCRYPT | NVRAM_MI | EZC_FLAGS_WRITE},
#else
	{ "wl_wpa_psk", "Pass phrase", wl_prefix, validate_wl_wpa_psk, ARGV("8", "64"), TRUE, NVRAM_ENCRYPT | NVRAM_MI | EZC_FLAGS_WRITE},
#endif	
	{ "wl_wpa_gtk_rekey", "Network Key Rotation Interval", wl_prefix, NULL, NULL, TRUE, NVRAM_MI | EZC_FLAGS_READ | EZC_FLAGS_WRITE },
	/* Multi SSID Guest interface flag */
	//{ "wl_guest", "Guest SSID Interface", wl_prefix, NULL, NULL, TRUE, WEB_IGNORE | NVRAM_MI  },
	{ "wl_power", "Transmit Power", NULL, validate_power, ARGV("76", "64", "52"), FALSE, 0 },
	/* MUST leave this entry here after all wl_XXXX variables */
	{ "wl_unit", "802.11 Instance", wl_prefix, wl_unit, NULL, TRUE, NVRAM_IGNORE | NVRAM_MI | NVRAM_MP },
	/* Internal variables */
	{ "os_server", "OS Server", NULL, NULL, NULL, TRUE, 0 },
	{ "stats_server", "Stats Server", NULL, NULL, NULL, TRUE, 0 },
	{ "timer_interval", "Timer Interval", NULL, NULL, NULL, TRUE, 0 },
	{ "lan_ifname", "LAN Interface Name", lan_prefix, NULL, NULL, TRUE, NVRAM_MI | NVRAM_MP },
	{ "lan_ifnames", "LAN Interface Names", NULL, NULL, NULL, TRUE, 0 },
	{ "wan_ifnames", "WAN Interface Names", NULL, NULL, NULL, TRUE, 0 },
};
/* build hashtable of the monster v-block

   Inputs:
   -tab: 	hash table  structure
   -vblock:	the monster V-block
   -num_items:	estimated size of hash table
   
   Returns: 0 on success -1 on error
*/

int 
hash_vtab(struct hsearch_data *tab,struct variable *vblock,int num_items)
{
	ENTRY e, *ep=NULL;
	int count;
	
	assert(tab);
	assert(vblock);

	if (!num_items) return -1;
	 
	if (!hcreate_r(num_items,tab))  
	{
			return -1;
	}
	
	
	for (count=0; count < num_items; count++)
	{
		e.key = vblock[count].name;
		e.data = &vblock[count];
		if (!hsearch_r(e, ENTER, &ep, tab))
			return -1;		
	}
	
	return 0;
	
}

/* The routine gets the pointer into the giant "V" block above give the cgi var name..
   Inputs:
   -varname: Pointer to cgi var name
   
   Returns: entry within "V" block or NULL if not found.
   
   It will try the following forms
   varXX_type
   var_typeXX
   vlanXXtype
   
   where XX is the instance number

*/

static struct variable*
get_var_handle(char *varname)
{
	ENTRY e, *ep=NULL;
	struct variable *variable=NULL;
	char *ptr=NULL,*ptr2=NULL;
	int offset;
	char prefix[8],tmp[64]; 
	
	if (!varname ) return NULL;
	if (!*varname) return NULL;
	
	ep=NULL;
	e.key = varname;
	hsearch_r(e, FIND, &ep, &vtab);
	
	
	/* found something */
	if (ep)
	{
		variable = (struct variable *)ep->data;
		if ( variable->ezc_flags & NVRAM_MP) return variable;
		
		/* Drop the variable is multi instance but not
		   Multi personality
		*/
		
		if ( variable->ezc_flags & NVRAM_MI) return NULL;
		if ( variable->ezc_flags & NVRAM_VLAN_MULTI) return NULL;
		if ( variable->ezc_flags & NVRAM_GENERIC_MULTI) return NULL;

		return variable;
		
	}
	
	/* variable not found could be the form of vlanXXtype, varXX_type or var_typeXX*/
	
	ptr=strchr(varname,'_');
	
	/* Is it  vlanXXtype ? */
	if (!ptr)
	{
		if (!strstr(varname,"vlan")) return NULL;
		strcpy(tmp,"vlan");
		offset=4;
		ptr=&varname[offset];
		while (*ptr){
				if (!isdigit((int)*ptr)) tmp[offset++] = *ptr;
				ptr++;
		}

		tmp[offset] = '\0';
		ep = NULL;
		e.key = tmp;
		hsearch_r(e, FIND, &ep, &vtab);
		
		/* check to see if this has the 
		   NVRAM_VLAN MULTI flag set in the v-block
		*/
		if (ep){
			variable = (struct variable *)ep->data;
			if (variable->ezc_flags & NVRAM_VLAN_MULTI) 
					return variable;
		}

		return  NULL;
	}
	
	/* Is it varXX_type ? */
	ptr2=varname;
	offset=0;
	memset(prefix,0,sizeof(prefix));
	while (ptr2 < ptr){
		if (isdigit((int)*ptr2)) break;
		prefix[offset++] =*ptr2;
		ptr2++;
	}
	snprintf(tmp,sizeof(tmp),"%s%s",prefix,ptr);

	ep = NULL;
	e.key = tmp;
	hsearch_r(e, FIND, &ep, &vtab);
	
	if (ep){
		variable = (struct variable *)ep->data;
		if (variable->ezc_flags & NVRAM_MI)
	 			return  variable;
	}
	
	/* Is is a var_typeXX */
	strncpy(tmp,varname,sizeof(tmp));
	offset = strlen(varname) - 1 ;
	while(isdigit((int)tmp[offset]))
			tmp[offset--] = '\0';
	ep = NULL;
	e.key = tmp;
	hsearch_r(e, FIND, &ep, &vtab);
		
	if (ep){
		variable = (struct variable *)ep->data;
		if (variable->ezc_flags & NVRAM_GENERIC_MULTI)
	 			return  variable;
	}

	return  NULL;
}

int 
variables_arraysize(void)
{
	return ARRAYSIZE(variables);
}

static void
validate_cgi(webs_t wp)
{
	struct variable *v=NULL;
	char *value=NULL;

	websBufferInit(wp);
	if (!webs_buf) {
		websWrite(wp, "out of memory<br>");
		websDone(wp, 0);
		return;
	}

	/* Validate and set variables in table order */
	for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {
		if (!(value = websGetVar(wp, v->name, NULL)))
			continue;
		if ((!*value && v->nullok) || !v->validate)
			nvram_set(v->name, value);
		else
			v->validate(wp, value, v, NULL);
	}

	websBufferFlush(wp);
}

static int
apply_cgi(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg,
	  char_t *url, char_t *path, char_t *query)
{
	int action = NOTHING;
	char *value, *page;
	int finish=0;

	value = websGetVar(wp, "action", "");
	page = websGetVar(wp, "page", "");
	cprintf("apply_cgi: action=%s \n", value);
	cprintf("apply_cgi: page=%s \n", page);

#ifdef ROUTER	
	int change = 0;
	int change_ip = 0;
	if (!strcmp(page,"lan.asp")) {
		char *lanip  = websGetVar(wp, "lan_ipaddr", "");
		char *lanmask  = websGetVar(wp, "lan_netmask", "");
		char *dhcpstart  = websGetVar(wp, "dhcp_start", "");
		char *dhcpstop  = websGetVar(wp, "dhcp_stop", "");
		
		if (strcmp(lanip, nvram_safe_get("lan_ipaddr")) ||
			strcmp(lanmask, nvram_safe_get("lan_netmask")) ||
			strcmp(dhcpstart, nvram_safe_get("dhcp_start")) ||
			strcmp(dhcpstop, nvram_safe_get("dhcp_stop")))
			change = 1;
		if (strcmp(lanip, nvram_safe_get("lan_ipaddr")))
			change_ip = 1;
	}		
#endif
		
	/* Apply values */
	if (!strcmp(value, "Apply")) {
		validate_cgi(wp);
		nvram_set("is_modified", "1");
		nvram_set("is_default", "0");
		nvram_commit();
		if (!error) {
			websHeader(wp);
			websWrite(wp, (char_t *) apply_header);
		#ifdef CLIENT
			websWrite(wp, "<h3>Wireless Client Configured</h3>");	
			websWrite(wp, "<p>Your U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge has been configured.</p>");	
			websWrite(wp, "<p>Please press the <b>Continue</b> button if you wish to continue device configuration or press the <b>Finish</b> button if you finished device configuration.</p>");
			finish = 1;
			action = RESTART;
		#endif
		}
		else {
			finish = 3;
			action = RESTART;
		}
	}

	/* Save values */
	else if (!strcmp(value, "Save") || !strcmp(value, "Synchronize")) {
		validate_cgi(wp);
		nvram_set("is_modified", "1");
		nvram_set("is_default", "0");
		nvram_commit();
		if (!error) {
			websHeader(wp);
			websWrite(wp, (char_t *) apply_header);
		#ifdef CLIENT
			websWrite(wp, "<h3>Wireless Client Configured</h3>");	
			websWrite(wp, "<p>Your U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge has been configured.</p>");
			websWrite(wp, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
			//websWrite(wp, "<p>To verify the status of the U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge device, <b>wait about 15 seconds</b> in order for the device to reboot and click <b>Continue</b>.</p>");
		        //websWrite(wp, "<p><b>You may now connect your U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge to your gaming console or other Ethernet-ready device.<br>Power cycle your U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge.<br>Be sure to reconnect your wireless router or access point to your computer, if necessary.</b></p>");
			nvram_set("router_disable","1");
			nvram_commit();
			finish = 9;
			action = REBOOT;
		#endif
		#ifdef REPEATER
			websWrite(wp, "<h3>Range Extender Configured</h3>");	
			websWrite(wp, "<p>The Range Extender has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to return to the previous page.</p>");
			finish = 3;
			action = RESTART;
		#endif
		#ifdef AP
			websWrite(wp, "<h3>Access Point Configured</h3>");
			websWrite(wp, "<p>The access point has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to return to the previous page.</p>");
			finish = 3;
			action = RESTART;
		#endif
		#ifdef APB
			websWrite(wp, "<h3>Access Point Configured</h3>");
			websWrite(wp, "<p>The access point has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to return to the previous page.</p>");
			finish = 3;
			action = RESTART;
		#endif
		#ifdef GAMING
			websWrite(wp, "<h3>Gaming Adapter Configured</h3>");
			websWrite(wp, "<p>The gaming adapter has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to return to the previous page.</p>");
			finish = 3;
			action = RESTART;
		#endif
		#ifdef BRIDGE
			websWrite(wp, "<h3>Bridge Configured</h3>");
			if (!strcmp(page,"wireless.asp")) {
				websWrite(wp, "<p>The bridge has been configured and will reboot.</p>");
				websWrite(wp, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device is restarting.</p>");
				finish = 8;
				action = REBOOT;
			}
			else {
				websWrite(wp, "<p>The bridge has been configured.</p>");
				websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to return to the previous page.</p>");
				finish = 3;
				action = RESTART;
			}
		#endif
		#ifdef ROUTER
			websWrite(wp, "<h3>Router Configured</h3>");	
			if (change == 1) {
				websWrite(wp, "<p>The router has been configured and will reboot.</p>");
				if (change_ip == 1) {
					websWrite(wp, "<p>You have modified the router&rsquo;s IP address, but your browser will continue to use the old IP address.");
					websWrite(wp, "This means that you will need to enter the router&rsquo;s new IP address in your browser in order to access the router again.");
					websWrite(wp, " (First you may have to release and renew the IP addresses of all devices connected to the router so they can acquire a new IP address and re-connect.)</p>");
				}
				websWrite(wp, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
				finish = 8;		//was 10
				action = REBOOT;
			}
			else {
				websWrite(wp, "<p>The router has been configured.</p>");
				if (!strcmp(page,"wan.asp")) {
					websWrite(wp, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device is restarting.</p>");
					finish = 10;
				}
				else {
					websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to return to the previous page.</p>");
					finish = 3;
				}
				action = RESTART;
			}
		#endif
		//printf("Save finish\n");
		}
		else {
			action = RESTART;
			finish = 3;
		}
	}
	
	else if (!strncmp(value, "Next", 4)) {
		/*** KEO  check if we need to verify for pppoe login ***/
		validate_cgi(wp);
		nvram_set("is_modified", "1");
		nvram_set("is_default", "0");
		if (!error) {
			action = NOTHING;		
			nvram_commit();
			
			char *str;
			char *usr,*pwd;
		
			str = websGetVar(wp, "wan_proto", "");
			usr = websGetVar(wp, "wan_pppoe_username", "");
			pwd = websGetVar(wp, "wan_pppoe_passwd", "");
		
			/* check if we need to test a pppoe login */
			//cprintf("apply_cgi: wan_proto = %s \n", str);
			//cprintf("apply_cgi: wan_pppoe_username = %s \n", usr);
			//cprintf("apply_cgi: wan_pppoe_passwd = %s \n", pwd);
			if (!strcmp(str, "pppoe")) 
			{
#ifdef ROUTER		    		
				websHeader(wp);
		    		websWrite(wp, (char_t *) apply_header_detect);
		    		websWrite(wp, "<h3>Please Wait&hellip;</h3>");
		    		websWrite(wp, "<p>The router is attempting to establish an Internet connection. If it&rsquo;s successful, you&rsquo;ll automatically be redirected to Step 2 of the setup procedure. If it's not successful an error message will be displayed below.</p>");
		    		usleep(2000000);
				if (check_pppoelogin(wp, usr, pwd) > 0)
		    		{
					finish = 5;
		   		}
		     		else {
		      			websWrite(wp, "<div class=\"clsNotice\">The router was <strong>unable to establish a connection</strong> to the Internet.</p>");
					if (!strcmp(page,"setup_wan.asp")) {
		    				websWrite(wp, "<p>This may be because the router isn&rsquo;t connected to the DSL modem correctly or the modem may not be powered on. ");
						websWrite(wp, "It&rsquo;s also possible that your modem isn&rsquo;t connected to the Internet itself or your DSL service has not been activated yet. ");
						websWrite(wp, "You may need to contact your Internet Service Provider for assistance.</p>");
						websWrite(wp, "<p>If everything is connected properly, the user name or password may be invalid.");
		    				finish = 6;
		    			}
					else {
		    				websWrite(wp, "<p>The username or password is invalid. "); 
						finish = 6;
		    			}
					websWrite(wp, "You can re-enter the user name and password, or you can choose to continue and use the user name and password you have already entered. ");
					websWrite(wp, "If you choose to continue, you can always configure your Internet connection later from the router&rsquo;s <strong>Internet</strong> page.</p></div>");
		     		}
#else
			;
#endif
			}
			else {
				return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
			}
		}
		else {
			finish = 3;
			action = RESTART;
		}
	} /* Next */

	/* Refresh log.asp */ 
	else if (!strcmp(value, "Refresh") || !strcmp(value, "Scan")) {
			action = NOTHING;		
			return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
	}
	/* Reboot the device */
	else if (!strcmp(value, "Reboot")) {
			websHeader(wp);
			websWrite(wp, (char_t *) apply_header);
	#ifdef ROUTER
			websWrite(wp, "<h3>Router Rebooting</h3>");
			websWrite(wp, "<p>The router is rebooting.</p>");
	#endif
	#ifdef AP
			websWrite(wp, "<h3>Access Point Rebooting</h3>");
			websWrite(wp, "<p>The access point is rebooting.</p>");
	#endif
	#ifdef APB
			websWrite(wp, "<h3>Access Point Rebooting</h3>");
			websWrite(wp, "<p>The access point is rebooting.</p>");
	#endif	
	#ifdef REPEATER
			websWrite(wp, "<h3>Range Extender Rebooting</h3>");
			websWrite(wp, "<p>The range extender is rebooting.</p>");
	#endif
	#ifdef GAMING
			websWrite(wp, "<h3>Gaming Adapter Rebooting</h3>");
			websWrite(wp, "<p>The gaming adapter is rebooting.</p>");
	#endif
	#ifdef BRIDGE
			websWrite(wp, "<h3>Bridge Rebooting</h3>");
			websWrite(wp, "<p>The bridge is rebooting.</p>");
	#endif			
			//websWrite(wp, "<p class=\"clsSave\">Please wait at least <b>15 seconds</b> and press the <b>Continue</b> button to proceed to the <strong>Status</strong> page.</p>");
			websWrite(wp, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
			action = REBOOT;
			finish = 8;
	}

	/* Restore defaults */
	else if (!strcmp(value, "Reload") || !strcmp(value, "Restore Defaults")) {
		finish = 2;
		websHeader(wp);
		websWrite(wp, (char_t *) apply_header);
	#ifdef CLIENT
		websWrite(wp, "<h3>Wireless Client Configured</h3>");
	#endif
	#ifdef ROUTER
		websWrite(wp, "<h3>Router Configured</h3>");
	#endif
	#ifdef AP
		websWrite(wp, "<h3>Access Point Configured</h3>");
	#endif
	#ifdef APB
		websWrite(wp, "<h3>Access Point Configured</h3>");
	#endif
	#ifdef REPEATER
		websWrite(wp, "<h3>Range Extender Configured</h3>");
	#endif
	#ifdef GAMING
		websWrite(wp, "<h3>Gaming Adapter Configured</h3>");
	#endif
	#ifdef BRIDGE
		websWrite(wp, "<h3>Bridge Configured</h3>");
	#endif
		nvram_set("sdram_ncdl", "0");
		nvram_set("restore_defaults", "1");
		nvram_commit();
		websWrite(wp, "<p>Default configuration has been restored.</p>");
		websWrite(wp, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
		action = REBOOT;
	}

	/* Release lease */
	else if (!strcmp(value, "Release")) {
		websHeader(wp);
		websWrite(wp, (char_t *) apply_header);
		websWrite(wp, "<p>Releasing lease...");
		if (sys_release()) 
			websWrite(wp, "error.</p>");
		else 
			websWrite(wp, "done.</p>");
		action = NOTHING;
		finish = 3;
	}

	/* Renew lease */
	else if (!strcmp(value, "Renew")) {
		websHeader(wp);
		websWrite(wp, (char_t *) apply_header);
		websWrite(wp, "<h3>WAN IP address</h3>");
		websWrite(wp, "<p>Releasing lease...");
		if (sys_release()) 
			websWrite(wp, "error.</p>");
		else {
			nvram_set("wan_ipaddr", "0.0.0.0");
			nvram_set("wan0_ipaddr", "0.0.0.0");
			nvram_set("wan_netmask", "0.0.0.0");
			nvram_set("wan0_netmask", "0.0.0.0");
			nvram_set("wan_gateway", "0.0.0.0");
			nvram_set("wan0_gateway", "0.0.0.0");
			nvram_set("wan_dns", "");
			nvram_set("wan0_dns", "");
			nvram_set("wan_wins", "");
			nvram_set("wan0_wins", "");
			nvram_commit();
			websWrite(wp, "done</p>");
		}
		websWrite(wp, "<p>Renewing lease...");
		if (sys_renew()) 
			websWrite(wp, "error</p>");
		else {
			usleep(3000000);
			char *ip = nvram_get("wan0_ipaddr");
			//printf("ip: %s\n", ip);
			if (strcmp(ip, "0.0.0.0"))
				websWrite(wp, "done.</p>");
			else
				websWrite(wp, "<p><b>The DHCP server is unavailable. Please check your Internet connection.</b></p>");
		}
		websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the router&rsquo;s <strong>Status</strong> page.</p>");
		action = NOTHING;
		finish = 3;
	}

	/* Reboot router */
	else if (!strcmp(value, "Finish")) {
		validate_cgi(wp);
		nvram_set("is_modified", "1");
		nvram_set("is_default", "0");
		if (!error) {
			websHeader(wp);
			websWrite(wp, (char_t *) apply_header);
		#ifdef CLIENT	
			websWrite(wp, "<h3>Wireless Client Configured</h3>");
			websWrite(wp, "<p>Your U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge has been configured.</p>");
			websWrite(wp, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
			nvram_set("router_disable","1");
			action = REBOOT;
			finish = 8;
		#endif
		#ifdef REPEATER	
			websWrite(wp, "<h3>Range Extender Configured</h3>");
			websWrite(wp, "<p>The range extender has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the range extender&rsquo;s <strong>Status</strong> page.</p>");
			action = RESTART;
			finish = 4;
		#endif
		#ifdef AP
			websWrite(wp, "<h3>Access Point Configured</h3>");
			websWrite(wp, "<p>The access point has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the access point&rsquo;s <strong>Status</strong> page.</p>");
			action = RESTART;
			finish = 4;
		#endif
		#ifdef APB
			websWrite(wp, "<h3>Access Point Configured</h3>");
			websWrite(wp, "<p>The access point has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the access point&rsquo;s <strong>Status</strong> page.</p>");
			action = RESTART;
			finish = 4;
		#endif
		#ifdef ROUTER
			websWrite(wp, "<h3>Router Configured</h3>");
			websWrite(wp, "<p>The router has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the router&rsquo;s <strong>Status</strong> page.</p>");
			action = RESTART;
			finish = 4;
		#endif
		#ifdef GAMING
			websWrite(wp, "<h3>Gaming Adapter Configured</h3>");
			websWrite(wp, "<p>The gaming adapter has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the gaming adapter&rsquo;s <strong>Status</strong> page.</p>");
			action = RESTART;
			finish = 4;
		#endif
		#ifdef BRIDGE
			websWrite(wp, "<h3>Bridge Configured</h3>");
			websWrite(wp, "<p>The bridge has been configured.</p>");
			websWrite(wp, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the bridge&rsquo;s <strong>Status</strong> page.</p>");
			action = RESTART;
			finish = 4;
		#endif
			nvram_commit();
		}
		else {
			action = RESTART;
			finish = 3;
		}
	}
	/* Remote Upgrade image 
	else if (!strncmp(value, "Upgrade",7)) {
	    
	    cprintf("process Upgrade now\n");
	    
	    websHeader(wp);
	    websWrite(wp, (char_t *) apply_header);
	    websWrite(wp, "<p>Retrieving the new firmware , please wait...</p>");
#ifdef ROUTER
	    if (sys_upgrade(DEFAULT_5461_URL, NULL, NULL)) {
		websWrite(wp, "<p>The router was unable to retrieve the new firmware, ");
		websWrite(wp, "<p>please check your <b>Internet connection</b> and try again.</p>");
		finish = 3;
		}
#endif
#ifdef AP
	    if (sys_upgrade(DEFAULT_5451_URL, NULL, NULL)) {
		websWrite(wp, "<p>The access point was unable to retrieve the new firmware, ");
		websWrite(wp, "<p>please check your <b>Internet connection</b> and try again.</p>");
		finish = 3;
		}
#endif
#ifdef APB
	    if (sys_upgrade(DEFAULT_5451_URL, NULL, NULL)) {
		websWrite(wp, "<p>The access point was unable to retrieve the new firmware, ");
		websWrite(wp, "<p>please check your <b>Internet connection</b> and try again.</p>");
		finish = 3;
		}
#endif
#ifdef REPEATER
	    if (sys_upgrade(DEFAULT_5441_URL, NULL, NULL)) {
		websWrite(wp, "<p>The repeater was unable to retrieve the new firmware, ");
		websWrite(wp, "<p>please check your <b>Internet connection</b> and try again.</p>");
		finish = 3;
		}
#endif
#ifdef CLIENT
	    if (sys_upgrade(DEFAULT_5431_URL, NULL, NULL)) {
		websWrite(wp, "<p>The wireless client was unable to retrieve the new firmware, ");
		websWrite(wp, "<p>please check your <b>Internet connection</b> and try again.</p>");
		finish = 3;
		}
#endif
	     else {
		new_fw_available = 0;
		sz_new_fw_version[0] = 0;
		//sz_new_fw_date[0] = 0;
		finish = 4;
	#ifdef CLIENT	
		nvram_set("router_disable","1");
	#endif	
		nvram_unset("os_date");
		nvram_commit();
		websWrite(wp, "<br><h3>Router Rebooting</h3>");
		websWrite(wp, "<p>The router is rebooting.</p>");
		websWrite(wp, "<p class=\"clsSave\">Please wait at least <b>15 seconds</b> and press the <b>Continue</b> button to proceed to the <strong>Status</strong> page.</p>");
		action = REBOOT;
	    }
	}
	
	// Upgrade image 
	else if (!strcmp(value, "Upgrade")) {
		char *os_name = nvram_safe_get("os_name");
		char *os_server = websGetVar(wp, "os_server", nvram_safe_get("os_server"));
		char *os_version = websGetVar(wp, "os_version", "current");
		char url[PATH_MAX];
		if (!*os_version)
			os_version = "current";
		snprintf(url, sizeof(url), "%s/%s/%s/%s.trx",
			 os_server, os_name, os_version, os_name);
		websHeader(wp);
		websWrite(wp, (char_t *) apply_header);
		websWrite(wp, "Retrieving %s...", url);
		if (sys_upgrade(url, NULL, NULL)) {
			websWrite(wp, "error<br>");
			finish = 3;
			goto footer;
		} else {
			finish = 4;
		#ifdef CLIENT	
			nvram_set("router_disable","1");
		#endif	
			nvram_unset("os_date");
			nvram_commit();
			websWrite(wp, "done<br>");
			action = REBOOT;
		}
	}

	// Report stats 
	else if (!strcmp(value, "Stats")) {
		char *server = websGetVar(wp, "stats_server", nvram_safe_get("stats_server"));
//		websWrite(wp, "Contacting %s...", server);
		if (sys_stats(server)) {
			websHeader(wp);
			websWrite(wp, (char_t *) apply_header);
			websWrite(wp, "<p>Contacting %s...", server);
			websWrite(wp, "error.</p>");
			finish = 3;
			goto footer;
		} else {
//			websWrite(wp, "done<br>");
			nvram_set("stats_server", server);
		}
	}
	// Radio On Off  
	else if (!strcmp(value, "RadioOff")) {
		websWrite(wp, "Turing Off Radio...");
		wl_radio_onoff(wp, 1);
		action = NOTHING;
	}
	else if (!strcmp(value, "RadioOn")) {
		websWrite(wp, "Radio on...");
		wl_radio_onoff(wp, 0);
		action = NOTHING;
	} 
#ifdef __CONFIG_NAT__
	// Delete connection 
	else if (!strcmp(value, "Delete")) {
		int unit;
		if (!(value = websGetVar(wp, "wan_unit", NULL)) ||
		    (unit = atoi(value)) < 0 || unit >= MAX_NVPARSE) {
		    	websHeader(wp);
			websWrite(wp, (char_t *) apply_header);
			websWrite(wp, "<p>Unable to delete connection, index error.</p>");
			finish = 3;
			action = NOTHING;
		}
		else {
			struct nvram_tuple *t;
			char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";
			int unit2, units = 0;
			//
			// We cann't delete the last connection since we can't differentiate
			// the cases where user does not want any connection (user deletes
			// the last one) vs. the router is booted the first time when there
			// is no connection at all (where a default one is created anyway).
			for (unit2 = 0; unit2 < MAX_NVPARSE; unit2 ++) {
				wan_prefix(unit2, prefix);
				if (nvram_get(strcat_r(prefix, "unit", tmp)) && units++ > 0)
					break;
			}
			if (units < 2) {
				websHeader(wp);
				websWrite(wp, (char_t *) apply_header);
				websWrite(wp, "<p>Can not delete the last connection.</p>");
				finish = 3;
				action = NOTHING;
			}
			else {
				// set prefix 
				wan_prefix(unit, prefix);
				// remove selected wan%d_ set 
//				websWrite(wp, "Deleting connection...");
				for (t = router_defaults; t->name; t ++) {
					if (!strncmp(t->name, "wan_", 4))
						nvram_unset(strcat_r(prefix, &t->name[4], tmp));
				}
				// fix unit number 
				unit2 = unit;
				for (; unit < MAX_NVPARSE; unit ++) {
					wan_prefix(unit, prefix);
					if (nvram_get(strcat_r(prefix, "unit", tmp)))
						break;
				}
				if (unit >= MAX_NVPARSE) {
					unit = unit2 - 1;
					for (; unit >= 0; unit --) {
						wan_prefix(unit, prefix);
						if (nvram_get(strcat_r(prefix, "unit", tmp)))
							break;
					}
				}
				snprintf(tmp, sizeof(tmp), "%d", unit);
				nvram_set("wan_unit", tmp);
				// check if there is any primary connection - see comment in wan_unit() 
				for (unit = 0; unit < MAX_NVPARSE; unit ++) {
					wan_prefix(unit, prefix);
					if (!nvram_get(strcat_r(prefix, "unit", tmp)))
						continue;
					if (nvram_invmatch(strcat_r(prefix, "proto", tmp), "disabled") &&
					    nvram_match(strcat_r(prefix, "primary", tmp), "1"))
						break;
				}
				// no one is primary, pick the first enabled one as primary 
				if (unit >= MAX_NVPARSE)
					wan_primary(wp);
				// save the change 
				nvram_set("is_modified", "1");
				nvram_set("is_default", "0");
				nvram_commit();
//				websWrite(wp, "done<br>");
				action = RESTART;
			}
		}
	}
#endif	// __CONFIG_NAT__ */
#ifdef __CONFIG_NAT__	
	else if (!strcmp(value, "Detect")) {
		int detection = wan_f_detection(1, wp, arg, query);
		//printf ("detection: %d\n", detection);
		//websWrite(wp, "detection=%d", detection);
		return 1;
	}
#endif	// __CONFIG_NAT__ */	
	/* Invalid action */
	else {
		websHeader(wp);
		websWrite(wp, (char_t *) apply_header);
		websWrite(wp, "<p>Invalid action %s.</p>", value);
		finish = 3;
	}

  footer:
	if (finish == 1) {
		websWrite(wp, (char_t *) apply_footer_finish, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
	else if (finish == 2) {
		websWrite(wp, (char_t *) apply_footer_restore, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
	else if (finish == 3) {
		usleep(2000000);
		websWrite(wp, (char_t *) apply_footer_error, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
	else if (finish == 4) {		 
//		return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
		sleep(2);
		websWrite(wp, (char_t *) apply_footer, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}	
#ifdef ROUTER	
	else if (finish == 5) {		 
		websWrite(wp, (char_t *) apply_footer_detect, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
	else if (finish == 6) {		 
		websWrite(wp, (char_t *) apply_footer_nodetect, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
#endif
/*	else if (finish == 7) {
		char *lanip;
		lanip = websGetVar(wp, "lan_ipaddr", "");
		//cprintf("ip2: %s\n", lanip);
		websWrite(wp, "<p><input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" ");
		websWrite(wp, "OnClick=\"document.location.href='http://");
		websWrite(wp, lanip);
		websWrite(wp, "/lan.asp';\"></p>");
		websWrite(wp, "</form>");
		websWrite(wp, "<p class=\"clsCopyright\">&copy; 2004-2005 U.S. Robotics Corporation.</p>");
		websWrite(wp, "</body>");
		websFooter(wp);
		websDone(wp, 200);
	} */
	else if (finish == 8) {		 
		websWrite(wp, (char_t *) apply_footer_reboot, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
#ifdef CLIENT
	else if (finish == 9) {		 
		websWrite(wp, (char_t *) apply_footer_save, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
#endif
#ifdef ROUTER
	else if (finish == 10) {		 
		websWrite(wp, (char_t *) apply_footer_restart, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
#endif
/*
#ifdef ROUTER
	else if (finish == 10) {
		char *lanip;
		lanip = websGetVar(wp, "lan_ipaddr", "");
		//cprintf("ip2: %s\n", lanip);
		websWrite(wp,"<div id=\"idContinue\" style=\"display: none;\">");
		websWrite(wp,"<p class=\"clsSave\">Please press the <b>Continue</b> button to proceed to the <b>Status</b> page.</p>");
		websWrite(wp,"<p><input type=\"button\" class=\"clsBtnSave\" name=\"action\" value=\"Continue -->\" onclick=\"document.location.href='status.asp';\"></p>");
		websWrite(wp, "OnClick=\"document.location.href='http://");
		websWrite(wp, lanip);
		websWrite(wp, "/lan.asp';\"></p>");		 
		websWrite(wp, (char_t *) apply_footer_lan, websGetVar(wp, "page", ""));
		websFooter(wp);
		websDone(wp, 200);
	}
#endif		*/
		
	flag = 0;
		
	if (action == RESTART)
		sys_restart();
	else if (action == REBOOT)
		sys_reboot();
	
	return 1;
}

#ifdef CLIENT
static int
index_asp(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg,
	     char_t *url, char_t *path, char_t *query)
{
	struct variable *v;
	char *value;
	char name[IFNAMSIZ], *next;
	int unit = 0;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";

	/* Can enter this function through GET or POST */
	if ((value = websGetVar(wp, "action", NULL))) {
		if (strcmp(value, "Select")){
			return apply_cgi(wp, urlPrefix, webDir, arg, url, path, query);
		}
	}
	/* copy wl%d_XXXXXXXX to wl_XXXXXXXX */
	if ((value = websGetVar(wp, "wl_unit", NULL))) {
		unit = atoi(value);
	} else {
		char ifnames[256];
		snprintf(ifnames, sizeof(ifnames), "%s %s",
			nvram_safe_get("lan_ifnames"),
			nvram_safe_get("wan_ifnames"));
		/* Probe for first wl interface */
		foreach(name, ifnames, next) {
			if (wl_probe(name) == 0 &&
			    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0)
				break;
		}
		if (!*name)
			unit = -1;
	}
	if (unit >= 0) {
		snprintf(prefix, sizeof(prefix), "wl%d_", unit);
		for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {
			if (!strncmp(v->name, "wl_", 3))
				nvram_set(v->name, nvram_safe_get(strcat_r(prefix, &v->name[3], tmp)));
			if (!strncmp(v->name, "wl_unit", 7))
				break;
		}
	}

	/* Set currently selected unit */
	snprintf(tmp, sizeof(tmp), "%d", unit);
	nvram_set("wl_unit", tmp);

	/* Display the page */
	return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
}

static int
advanced_asp(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg,
	     char_t *url, char_t *path, char_t *query)
{
	struct variable *v;
	char *value;
	char name[IFNAMSIZ], *next;
	int unit = 0;
	char tmp[100], prefix[] = "wlXXXXXXXXXX_";


	/* Can enter this function through GET or POST */
	if ((value = websGetVar(wp, "action", NULL))) {
		if (strcmp(value, "Select")){
			return apply_cgi(wp, urlPrefix, webDir, arg, url, path, query);
		}
	}

	/* copy wl%d_XXXXXXXX to wl_XXXXXXXX */
	if ((value = websGetVar(wp, "wl_unit", NULL))) {
		unit = atoi(value);
	} else {
		char ifnames[256];
		snprintf(ifnames, sizeof(ifnames), "%s %s", 
			nvram_safe_get("lan_ifnames"),
			nvram_safe_get("wan_ifnames"));
		/* Probe for first wl interface */
		foreach(name, ifnames, next) {
			if (wl_probe(name) == 0 &&
			    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0)
				break;
		}
		if (!*name)
			unit = -1;
	}
	if (unit >= 0) {
		snprintf(prefix, sizeof(prefix), "wl%d_", unit);
		for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {
			if (!strncmp(v->name, "wl_", 3))
				nvram_set(v->name, nvram_safe_get(strcat_r(prefix, &v->name[3], tmp)));
			if (!strncmp(v->name, "wl_unit", 7))
				break;
		}
	}

	/* Set currently selected unit */
	snprintf(tmp, sizeof(tmp), "%d", unit);
	nvram_set("wl_unit", tmp);

	/* Display the page */
	return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
}
#endif

static int
wireless_asp(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg,
	     char_t *url, char_t *path, char_t *query)
{
	struct variable *v;
	char *value;
	char name[IFNAMSIZ], *next;
	int unit = 0;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";

	/* Can enter this function through GET or POST */
	if ((value = websGetVar(wp, "action", NULL))) {
		if (strcmp(value, "Select")){
			return apply_cgi(wp, urlPrefix, webDir, arg, url, path, query);
		}
	}

	/* copy wl%d_XXXXXXXX to wl_XXXXXXXX */
	if ((value = websGetVar(wp, "wl_unit", NULL))) {
		unit = atoi(value);
	} else {
		char ifnames[256];
		snprintf(ifnames, sizeof(ifnames), "%s %s",
			nvram_safe_get("lan_ifnames"),
			nvram_safe_get("wan_ifnames"));
		/* Probe for first wl interface */
		foreach(name, ifnames, next) {
			if (wl_probe(name) == 0 &&
			    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0)
				break;
		}
		if (!*name)
			unit = -1;
	}
	if (unit >= 0) {
		snprintf(prefix, sizeof(prefix), "wl%d_", unit);
		for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {
			if (!strncmp(v->name, "wl_", 3))
				nvram_set(v->name, nvram_safe_get(strcat_r(prefix, &v->name[3], tmp)));
			if (!strncmp(v->name, "wl_unit", 7))
				break;
		}
	}

	/* Set currently selected unit */
	snprintf(tmp, sizeof(tmp), "%d", unit);
	nvram_set("wl_unit", tmp);

	/* Display the page */
	return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
}

static int
security_asp(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg, 
	     char_t *url, char_t *path, char_t *query)
{
	struct variable *v;
	char *value;
	char name[IFNAMSIZ], *next;
	int unit = 0;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wlXXXXXXXXXX_";

	/* Can enter this function through GET or POST */
	if ((value = websGetVar(wp, "action", NULL))) {
		if (strcmp(value, "Select"))
			return apply_cgi(wp, urlPrefix, webDir, arg, url, path, query);
	}

	/* copy wl%d_XXXXXXXX to wl_XXXXXXXX */
	if ((value = websGetVar(wp, "wl_unit", NULL))) {
		unit = atoi(value);
	} else {
		char ifnames[256];
		snprintf(ifnames, sizeof(ifnames), "%s %s", 
			nvram_safe_get("lan_ifnames"),
			nvram_safe_get("wan_ifnames"));
		/* Probe for first wl interface */
		foreach(name, ifnames, next) {
			if (wl_probe(name) == 0 &&
			    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0)
				break;
		}
		if (!*name)
			unit = -1;
	}
	if (unit >= 0) {
		snprintf(prefix, sizeof(prefix), "wl%d_", unit);
		for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {
			if (!strncmp(v->name, "wl_", 3))
				nvram_set(v->name, nvram_safe_get(strcat_r(prefix, &v->name[3], tmp)));
			if (!strncmp(v->name, "wl_unit", 7))
				break;
		}
	}

	/* Set currently selected unit */
	snprintf(tmp, sizeof(tmp), "%d", unit);
	nvram_set("wl_unit", tmp);

	/* Display the page */
	return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
}

#ifdef __CONFIG_NAT__
#ifdef ROUTER
static int
wan_asp(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg, 
	     char_t *url, char_t *path, char_t *query)
{
	//struct variable *v;
	char *value;
	int unit = 0;
	char tmp[NVRAM_BUFSIZE], prefix[] = "wanXXXXXXXXXX_";
	char ustr[16];
	struct nvram_tuple *t=NULL;

	/* Can enter this function through GET or POST */
	if ((value = websGetVar(wp, "action", NULL))) {
		if (!strcmp(value, "New")) {
			/* pick one that 'does not' exist */
			for (unit = 0; unit < MAX_NVPARSE; unit ++) {
				wan_prefix(unit, prefix);
				if (!nvram_get(strcat_r(prefix, "unit", tmp)))
					break;
			}
			if (unit >= MAX_NVPARSE) {
				websHeader(wp);
				websWrite(wp, (char_t *) apply_header);
				websWrite(wp, "Unable to create new connection. Maximum %d.", MAX_NVPARSE);
				websWrite(wp, (char_t *) apply_footer, websGetVar(wp, "page", ""));
				websFooter(wp);
				websDone(wp, 200);
				return 1;
			}
			/* copy default to newly created wan%d_ set */
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wan_", 4))
					nvram_set(strcat_r(prefix, &t->name[4], tmp), t->value);
			}
			/* the following variables must be overwritten */
			snprintf(ustr, sizeof(ustr), "%d", unit);
			nvram_set(strcat_r(prefix, "unit", tmp), ustr);
			nvram_set(strcat_r(prefix, "proto", tmp), "disabled");
			nvram_set(strcat_r(prefix, "ifname", tmp), nvram_safe_get("wan_ifname"));
			nvram_set(strcat_r(prefix, "hwaddr", tmp), nvram_safe_get("wan_hwaddr"));
			nvram_set(strcat_r(prefix, "ifnames", tmp), nvram_safe_get("wan_ifnames"));
			/* commit change */
			nvram_set("is_modified", "1");
			nvram_set("is_default", "0");
			nvram_commit();
		}
		else if (!strcmp(value, "Select")) {
			if ((value = websGetVar(wp, "wan_unit", NULL)))
				unit = atoi(value);
			else
				unit = -1;
		}
		else
			return apply_cgi(wp, urlPrefix, webDir, arg, url, path, query);
	}
	else if ((value = websGetVar(wp, "wan_unit", NULL)))
		unit = atoi(value);
	else
		unit = atoi(nvram_safe_get("wan_unit"));
	if (unit < 0 || unit >= MAX_NVPARSE)
		unit = 0;

	/* Set prefix */
	wan_prefix(unit, prefix);
	/* USR -you don't need this */
	/* copy wan%d_ set to wan_ set  
	for (v = variables; v < &variables[ARRAYSIZE(variables)]; v++) {
		if (strncmp(v->name, "wan_", 4))
			continue;
		value = nvram_get(strcat_r(prefix, &v->name[4], tmp));
		if (value) {
			printf("wan_asp: %s %d\n", v->name, &value);
			nvram_set(v->name, value);
		}
		if (!strncmp(v->name, "wan_unit", 8))
			break;
	}
	*/
	/* Set currently selected unit */
	snprintf(tmp, sizeof(tmp), "%d", unit);
	nvram_set("wan_unit", tmp);

	/* Display the page */
	return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
}
#endif  /* ROUTER */
#endif /* __CONFIG_NAT__ */

static int
setup_sec_asp(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg,
	     char_t *url, char_t *path, char_t *query)
{
	char *value;
	
	/* Can enter this function through GET or POST */
	if ((value = websGetVar(wp, "action", NULL))) {
		if (strcmp(value, "Select")){
			return apply_cgi(wp, urlPrefix, webDir, arg, url, path, query);
		}
	}
	return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
}
	

#ifdef CLIENT
static int
sitesurvey_asp(webs_t wp, char_t *urlPrefix, char_t *webDir, int arg,
	     char_t *url, char_t *path, char_t *query)
{
	char *value;
	char name[IFNAMSIZ],ssid[40],eaddr[40];
	char *next;
	int unit = 0, flag_client = 0;
	int unit2;
        char ifnames[256];
	wlc_ssid_t new_ssid;
	int rssi, channel;
	char chan[3];

	if ((unit2 = atoi(nvram_safe_get("wl_unit"))) < 0)
		return -1;

	snprintf(ifnames, sizeof(ifnames), "%s %s",
		nvram_safe_get("lan_ifnames"),
		nvram_safe_get("wan_ifnames"));
	
	foreach(name, ifnames, next) {
		if (wl_probe(name) == 0 &&
		    wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit)) == 0 &&
		    unit == unit2) {

		/* Can enter this function through GET or POST */
			if ((value = websGetVar(wp, "action", NULL))) {
				if (!strcmp(value, "Select")) {
					if(value = websGetVar(wp, "ap_sel", NULL)) {
						//printf("value is: %s \n", value);
						sscanf(value,"%s %s %s %d %d\n",name,ssid,eaddr,&rssi,&channel); //check == 3
						strcpy(new_ssid.SSID, ssid);
						new_ssid.SSID_len = strlen(ssid);
						sprintf(chan, "%d", channel);
						cprintf("channel: %s\n", chan);
						if (!wl_ioctl(name, WLC_SET_SSID, &new_ssid, sizeof(new_ssid))) {
							usleep(1000000);
							unsigned char buf[ETHER_ADDR_LEN];
			        			char addr[32];
							flag_client = 1;
							if (!wl_ioctl(name, WLC_GET_BSSID, buf, sizeof(buf))) {
								if (strcmp(ether_etoa(buf,addr),"00:00:00:00:00:00")) {
								//save new SSID in nvram
									nvram_set("wl_ssid",new_ssid.SSID);
									nvram_set("wl0_ssid",new_ssid.SSID);
									nvram_set("wl_channel", chan);
									nvram_set("wl0_channel", chan);
									nvram_commit();
									websHeader(wp);
									websWrite(wp, (char_t *) apply_header);
									websWrite(wp, "<h3>Wireless Client Configured</h3>");
									websWrite(wp, "<p>Your wireless device is associating with the selected network.</p>");
									websWrite(wp, "<p>Please press the <b>Continue</b> button if you wish to continue device configuration or press the <b>Finish</b> button if you finished device configuration.</p>");
									websWrite(wp, (char_t *) apply_footer_finish, websGetVar(wp, "page", "status.asp"));
									websFooter(wp);
									websDone(wp, 200);
								}
								else {
									printf("bssid is null!");
									flag_client = 0;
								}
							}
							else {
								printf("GET BSSID failed!\n");
								flag_client = 0;
							}
						}
						else {
							printf("SET SSID failed!\n");
						}
						if (flag_client == 0) {
							websHeader(wp);
							websWrite(wp, (char_t *) apply_header);
							websWrite(wp, "<h3>Site Survey result</h3>");
							websWrite(wp, "Your wireless device failed associating with the selected network.<br>");
							websWrite(wp, (char_t *) apply_footer, websGetVar(wp, "page", "tools.asp"));
							websFooter(wp);
							websDone(wp, 200);
						}
					}
					return 1;
				}
			}
		}
	}
	return websDefaultHandler(wp, urlPrefix, webDir, arg, url, path, query);
}
#endif

#ifdef WEBS

void
initHandlers(void)
{
	websAspDefine("nvram_get", ej_nvram_get);
	websAspDefine("nvram_match", ej_nvram_match);
	websAspDefine("nvram_invmatch", ej_nvram_invmatch);
	websAspDefine("nvram_list", ej_nvram_list);
	websAspDefine("filter_client", ej_filter_client);
	websAspDefine("forward_port", ej_forward_port);
	websAspDefine("static_route", ej_static_route);
	websAspDefine("localtime", ej_localtime);
	websAspDefine("dumplog", ej_dumplog);
	websAspDefine("syslog", ej_syslog);
	websAspDefine("dumpleases", ej_dumpleases);
	websAspDefine("link", ej_link);
	//websAspDefine("wme_match_op", ej_wme_match_op);

	websUrlHandlerDefine("/apply.cgi", NULL, 0, apply_cgi, 0);
	websUrlHandlerDefine("/wireless.asp", NULL, 0, wireless_asp, 0);
#ifdef CLIENT
	websUrlHandlerDefine("/index.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/ssid.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/peer.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/advanced.asp", NULL, 0, advanced_asp, 0);
        websUrlHandlerDefine("/security.asp", NULL, 0, security_asp, 0);
	websUrlHandlerDefine("/sitesurvey.asp", NULL, 0, sitesurvey_asp, 0);
#endif
#ifdef GAMING
	websUrlHandlerDefine("/setup_mode.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/setup_infra.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/setup_adhoc.asp", NULL, 0, index_asp, 0);
#endif
#ifdef BRIDGE
	websUrlHandlerDefine("/setup_mode.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/setup_infra.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/setup_adhoc.asp", NULL, 0, index_asp, 0);
	websUrlHandlerDefine("/setup_wds.asp", NULL, 0, index_asp, 0);
#endif

	websSetPassword(nvram_safe_get("http_passwd"));

#ifdef CLIENT
	websSetRealm("U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge");
#endif
#ifdef REPEATER
	websSetRealm("U.S. Robotics Wireless MAXg Range Extender");
#endif
#ifdef AP
	websSetRealm("U.S. Robotics Wireless MAXg Access Point");
#endif
#ifdef APB
	websSetRealm("U.S. Robotics Wireless MAXg Professional Access Point");
#endif
#ifdef ROUTER
	websSetRealm("U.S. Robotics Wireless MAXg Router");
#endif
#ifdef GAMING
	websSetRealm("U.S. Robotics Wireless MAXg Gaming Adapter");
#endif
#ifdef BRIDGE
	websSetRealm("U.S. Robotics Wireless MAXg Bridge");
#endif
	/* Initialize hash table */
	hash_vtab(&vtab,v,variables_arraysize());
}

#else /* !WEBS */

int internal_init(void)
{
	/* Initialize hash table */
	if (hash_vtab(&vtab,variables,variables_arraysize()))
		return -1;
	return 0;
}

static void
do_auth(char *userid, char *passwd, char *realm)
{
	assert(userid);
	assert(passwd);
	assert(realm);
	
	strncpy(userid, nvram_safe_get("http_username"), AUTH_MAX);
	strncpy(passwd, nvram_safe_get("http_passwd"), AUTH_MAX);
#ifdef CLIENT
	strncpy(realm, "U.S. Robotics 5430 Wireless Gaming Adapter and Ethernet Bridge", AUTH_MAX);
#endif
#ifdef REPEATER
	strncpy(realm, "U.S. Robotics Wireless MAXg Range Extender", AUTH_MAX);
#endif
#ifdef AP
	strncpy(realm, "U.S. Robotics Wireless MAXg Access Point", AUTH_MAX);
#endif
#ifdef APB
	strncpy(realm, "U.S. Robotics MAXg Professional Access Point", AUTH_MAX);
#endif
#ifdef ROUTER
	strncpy(realm, "U.S. Robotics Wireless MAXg Router", AUTH_MAX);
#endif
#ifdef GAMING
	strncpy(realm, "U.S. Robotics Wireless MAXg Gaming Adapter", AUTH_MAX);
#endif
#ifdef BRIDGE
	strncpy(realm, "U.S. Robotics Wireless MAXg Bridge", AUTH_MAX);
#endif
}

char post_buf[POST_BUF_SIZE];
char ezc_version[128];
char no_cache[] =
"Cache-Control: no-cache\r\n"
"Pragma: no-cache\r\n"
"Expires: 0"
;


static void
do_apply_post(char *url, FILE *stream, int len, char *boundary)
{
	assert(url);
	assert(stream);
			
	/* Get query */
	if (!fgets(post_buf, MIN(len + 1, sizeof(post_buf)), stream))
		return;
	len -= strlen(post_buf);
	printf("post_buf: %s \n", post_buf);
	/* Initialize CGI */
	init_cgi(post_buf);

	/* Slurp anything remaining in the request */
	while (len--)
		(void) fgetc(stream);
}

static void
do_apply_cgi(char *url, FILE *stream)
{
	char *path=NULL, *query=NULL;

	assert(url);
	assert(stream);

	cprintf("do_apply_cgi\n");
	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;

	apply_cgi(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}

static int upgrade_ret;

static void
do_upgrade_post(char *url, FILE *stream, int len, char *boundary)
{
	char buf[1024];
	upgrade_ret = EINVAL;
	int ret = 0;
	
	ret = nvram_commit();	
	/* Look for our part */
	while (len > 0) {
		if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
		len -= strlen(buf);
		if (!strncasecmp(buf, "Content-Disposition:", 20) &&
		    strstr(buf, "name=\"file\""))
			break;
	}

	/* Skip boundary and headers */
	while (len > 0) {
		if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
		len -= strlen(buf);
		if (!strcmp(buf, "\n") || !strcmp(buf, "\r\n"))
			break;
	}
	
	if (ret < 0) {
		cprintf("error commit!\n");
		upgrade_ret = 10;
	}
	else
		upgrade_ret = sys_upgrade(NULL, stream, &len);

	// Slurp anything remaining in the request 
	while (len--)
		(void) fgetc(stream);
}

static void
do_upgrade_cgi(char *url, FILE *stream)
{
	cprintf("do_upgrade_cgi\n");
	websHeader(stream);
	websWrite(stream, (char_t *) apply_header);
#ifdef ROUTER
	websWrite(stream, "<h3>Upgrade Router</h3>");
#endif
#ifdef AP
	websWrite(stream, "<h3>Upgrade Access Point</h3>");
#endif
#ifdef APB
	websWrite(stream, "<h3>Upgrade Access Point</h3>");
#endif
#ifdef REPEATER
	websWrite(stream, "<h3>Upgrade Range Extender</h3>");
#endif
#ifdef CLIENT
	websWrite(stream, "<h3>Upgrade Wireless Client</h3>");
#endif
#ifdef GAMING
	websWrite(stream, "<h3>Upgrade Gaming Adapter</h3>");
#endif
#ifdef BRIDGE
	websWrite(stream, "<h3>Upgrade Bridge</h3>");
#endif
	//websWrite(stream, "<p>The upgrade process has been started. Please wait...</p>");
	if (upgrade_ret) {
		websWrite(stream, "<p>Error during upgrade.");
		//if (COMPAT == 1)
		//	websWrite(stream, "</p><p>The new firmware is not compatible with your device. Please try again with the correct firmware.</p>");
		//else
		if (upgrade_ret == 10)
			websWrite(stream, " The device does not have enough memory to handle the upgrade process. Please <strong>reboot</strong> the device and try again. </p>");	
		else
			websWrite(stream, " Please try again.</p>");
		websWrite(stream, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the <strong>Device</strong> page.</p>");
#ifdef CLIENT		
		websWrite(stream, (char_t *) apply_footer_error, "tools.asp");
#else
		websWrite(stream, (char_t *) apply_footer_error, "device.asp");
#endif	
	}
	else {
#ifdef ROUTER
		websWrite(stream, "<p>The router has been upgraded.</p>");
#endif
#ifdef AP
		websWrite(stream, "<p>The access point has been upgraded.</p>");
#endif
#ifdef APB
		websWrite(stream, "<p>The access point has been upgraded.</p>");
#endif
#ifdef REPEATER
		websWrite(stream, "<p>The range extender has been upgraded.</p>");
#endif
#ifdef CLIENT		
		websWrite(stream, "<p>The wireless client has been upgraded.</p>");
		nvram_set("router_disable","1");
#endif
#ifdef GAMING
		websWrite(stream, "<p>The gaming adapter has been upgraded.</p>");
#endif
#ifdef BRIDGE
		websWrite(stream, "<p>The bridge has been upgraded.</p>");
#endif
		nvram_unset("os_date");
		nvram_commit();
		websWrite(stream, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
		websWrite(stream, (char_t *) apply_footer_reboot, "status.asp");
	}
	websFooter(stream);
	websDone(stream, 200);

	/* Reboot if successful */
	if (upgrade_ret == 0)
		sys_reboot();

}

extern int http_get(const char *server, char *buf, size_t count, off_t offset);


static void splitversionstr(char *v, int *M, int *m, int *rc, int *inc, int *b)
{

    char s[16],*p1,*p2,*p3,*p4,*p5;
    int n;
    
    *M = *m = *rc = *inc = *b = 0;
    p1 = p2 = p3 = p4 = p5 = 0;
    
    if (v)
    {
	p1 = strstr(v, ".");
	if (p1) p2 = strstr(p1+1, ".");
	if (p2) p3 = strstr(p2+1, ".");
	if (p3) p4 = strstr(p3+1, ".");
	if (p4) p5 = strstr(p4+1, ".");
	if ((!p1) || (!p2) || (!p3) || (!p4) || (!p5)) return;
	
	// get Major value 
	n = p1 - v;
	strncpy(s,v,n);
	s[n] = 0;
	*M = atoi(s);
	
	// get minor value 
	n = (p2 - p1) - 1;
	strncpy(s, p1+1,n);
	s[n] = 0;
	*m = atoi(s);
	
	// get RC value 
	n = (p3 - p2) - 1;
	strncpy(s, p2+1,n);
	s[n] = 0;
	*rc = atoi(s);
	
	// get incremental value 
	n = (p4 - p3) - 1;
	strncpy(s, p3+1,n);
	s[n] = 0;
	*inc = atoi(s);
	
	// get Build value 
	n = (p5 - p4) - 1;
	strncpy(s, p4+1,n);
	s[n] = 0;
	*b = atoi(s);
	
    }
}

/*
static int local_check_n_fw()
{

    char tmp[30];
    char *s, currentversion[64];
    int ret = 0;
    struct usr_header hdr;
    long count;
    struct tm *t = 0;
    int cM, cm, cB, cb, nM, nm, nB, nb;
    
    if (p3) p4 = strstr(p3+1, ".");
    force_check_fw = 0;

    sprintf(tmp, nvram_safe_get("wan0_ipaddr"));
    cprintf("check_n_fw: wan IP: %s\n",tmp);
    // We check for new firmware only if we get a wan connection
    if (strcmp(tmp,"0.0.0.0"))
    {
	// KEO Sep 27 2004
	//   We need to check that resolv.conf file should not content 0.0.0.0 
	//   as DNS ip address, because gethostbyname in http_get routine
	//   will hang if we try to call http_get routine 
	
	FILE *f;
	
        if ((f = fopen("/etc/resolv.conf", "r")))
	{
	    char buf[128];
	    int l;
	    
	    while (l = fgets(buf, sizeof(buf), f))
	    {
		if (strstr(buf,"0.0.0.0")) {
    		    ret = -1;	
		    goto nonew_fw;
		}
	    }
	    fclose(f);
	}
	// KEO Sep 17 2004
	 
	    NOTICE:
	    
		Even if we got wan_ipaddr, this does not mean that we can go 
		to the outside, so we need to check if we can resolve the 
		URL onto IP address, if it does then we got the internet DNS
		server somewhere.

	   	Assuming that on our upgrade site, we have at least the lastest 
		firmware version, so when we try to fetch for the firmware's 
		header the http_get routine should return n bytes of the header,
		if http_get return 0 bytes length then, which means 
		that there's no Internet connection. 
		      
		      
	gettimeofday(&last_check, NULL);

	strcpy(currentversion,nvram_safe_get("os_version"));
	splitversionstr(currentversion, &cM, &cm, &cB, &cb);
	cprintf("check_n_fw: current os_version: %d.%d.%d.%d.1\n",
		cM, cm, cB, cb);

	count = http_get(DEFAULT_5461_URL, (char *) &hdr, sizeof(struct usr_header), 0);
	if (count == 0) { ret = -1; goto nonew_fw; }
	if (count < sizeof(struct usr_header)) 
	{
	    cprintf("check_n_fw: File is too small (%ld bytes)\n", count);
	    goto nonew_fw;
	}
	if (hdr.magic != USR_MAGIC ||
	    hdr.len < sizeof(struct usr_header)) 
	{
	    cprintf("check_n_fw: Bad usr header\n");
	    goto nonew_fw;
	}
	
	nM = hdr.version & 0xff;
	nm = (hdr.version >> 8) & 0xff;
	nB = (hdr.version >> 16) & 0xff;
	nb = (hdr.version >> 24) & 0xff;
	sprintf(sz_new_fw_version,"%d.%d.%d.%d.1",nM,nm,nB,nb);
	
	if (strcmp(sz_new_fw_version,currentversion) > 0)
	{
	    cprintf("check_n_fw: new firmware = %s\n",sz_new_fw_version);
	    t = gmtime(&hdr.time);
	    if (t) {
	        sprintf(sz_new_fw_date,"%s %#2d %d",
			sz_month[t->tm_mon], t->tm_mday, t->tm_year+1900);
	    }
	    cprintf("check_n_fw: new date = %s\n",sz_new_fw_date);
	    ret = 1;
	}
    } else ret = -1;
nonew_fw:
    new_fw_available = ret;
    
    return ret;
}
*/

/* 
    Automatic check for a new firmware every 30mn 
    This routine is called by httpd 
*
int check_for_new_fw()
{

    struct timeval now;
    int ret = new_fw_available;

    if (new_fw_available != 1)
    {
	gettimeofday(&now, NULL);
	if ((force_check_fw) || 
	    (!last_check.tv_sec) || 
	    ((now.tv_sec - last_check.tv_sec) > 1800))
	{
	    cprintf("check_for_new_fw: call local_check_n_fw()\n");
	    ret = local_check_n_fw();
	}
    }

    return ret;
}
*/

/* 
    This routine is called by the asp pages to check for  
    a new firmware available. 
    This routine just returns the value of new_fw_available variable
    to the asp page.
*
int get_fw_status(int eid, webs_t wp, int argc, char_t **argv)
{
    int ret = 0;

    cprintf("check_new_fw: new_fw_available=%d\n",new_fw_available);
    if (new_fw_available == 1) 
	ret += websWrite(wp, "1");
    else
	ret += websWrite(wp, "0");
	
    return ret;
}
*/

static void
do_check_fw_post(char *url, FILE *stream, int len, char *boundary)
{
	char buf[1024];

	cprintf("do_check_fw_post: len=%d\n",len);
	/* Look for our part */
	while (len > 0) {
		if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
		len -= strlen(buf);
		if (!strncasecmp(buf, "Content-Disposition:", 20) &&
		    strstr(buf, "name=\"file\""))
			break;
	}

	cprintf("do_check_fw_post: after look for our part\n");
	/* Skip boundary and headers */
	while (len > 0) {
		if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
		len -= strlen(buf);
		if (!strcmp(buf, "\n") || !strcmp(buf, "\r\n"))
			break;
	}
	/*if (new_fw_available <= 0)
	{
	    cprintf("do_check_fw_post: call local_check_n_fw()\n");
	    do_check_fw_cgi(url, stream);
	}*/
	/* Slurp anything remaining in the request */
	while (len--)
		(void) fgetc(stream);
}

static void
do_check_fw_cgi(char *url, FILE *stream)
{
    char tmp[30];
    char currentversion[64], curr_version[64], sz_new_fw_version1[64];
    int ret = 0;
    struct usr_header hdr;
    long count;
    //struct tm *t = 0;
    int cM, cm, crc, cinc, cb;
    int nM, nm, nrc, ninc, nb;
       
    //force_check_fw = 0;

    cprintf("do_check_fw_cgi\n");
#ifdef ROUTER	
    sprintf(tmp, nvram_safe_get("wan0_ipaddr"));
#else
    sprintf(tmp, nvram_safe_get("lan_dns"));
#endif
    cprintf("check_n_fw: wan IP: %s\n",tmp);
    // We check for new firmware only if we get a wan connection
    /* load a waiting page, in case that http_get take too long to return */	
    websHeader(stream);
    websWrite(stream, (char_t *) apply_header);
    websWrite(stream, "<h3>Check for Update</h3>");
    websWrite(stream, "<p>It may take a minute or so to check the U.S. Robotics Web site for an update. Please wait...</p>");		
    if (strcmp(tmp,"0.0.0.0"))
    {
	/* KEO Sep 27 2004
	   We need to check that resolv.conf file should not content 0.0.0.0 
	   as DNS ip address, because gethostbyname in http_get routine
	   will hang if we try to call http_get routine 
	*/
	FILE *f;
	
        if ((f = fopen("/etc/resolv.conf", "r")))
	{
	    char buf[128];
	    int l;
	    
	    while (l = fgets(buf, sizeof(buf), f))
	    {
		if (strstr(buf,"0.0.0.0")) {
    		    ret = -1;	
		    goto nonew_fw;
		}
	    }
	    fclose(f);
	}
	/* KEO Sep 17 2004
	 
	    NOTICE:
	    
		Even if we got wan_ipaddr, this does not mean that we can go 
		to the outside, so we need to check if we can resolve the 
		URL onto IP address, if it does then we got the internet DNS
		server somewhere.

	   	Assuming that on our upgrade site, we have at least the lastest 
		firmware version, so when we try to fetch for the firmware's 
		header the http_get routine should return n bytes of the header,
		if http_get return 0 bytes length then, which means 
		that there's no Internet connection. */
		      
		      
	//gettimeofday(&last_check, NULL);

	strcpy(currentversion,nvram_safe_get("os_version"));
	strcat(currentversion, ".");
	splitversionstr(currentversion, &cM, &cm, &crc, &cinc, &cb);
	cprintf("check_n_fw: current os_version: %03d.%03d.%03d.%03d.%03d\n",cM, cm, crc, cinc, cb);
	sprintf(curr_version,"%03d.%03d.%03d.%03d.%03d",cM,cm,crc,cinc,cb);
#ifdef ROUTER	
	count = http_get(DEFAULT_5461_URL, (char *) &hdr, sizeof(struct usr_header), 0);
#endif
#ifdef AP
	count = http_get(DEFAULT_5451_URL, (char *) &hdr, sizeof(struct usr_header), 0);
#endif
#ifdef APB
	count = http_get(DEFAULT_5453_URL, (char *) &hdr, sizeof(struct usr_header), 0);
#endif
#ifdef CLIENT
	count = http_get(DEFAULT_5430_URL, (char *) &hdr, sizeof(struct usr_header), 0);
#endif
#ifdef REPEATER
	count = http_get(DEFAULT_5441_URL, (char *) &hdr, sizeof(struct usr_header), 0);
#endif
#ifdef GAMING
	count = http_get(DEFAULT_5431_URL, (char *) &hdr, sizeof(struct usr_header), 0);
#endif
#ifdef BRIDGE
	count = http_get(DEFAULT_5432_URL, (char *) &hdr, sizeof(struct usr_header), 0);
#endif
	if (count == 0) { ret = -1; goto nonew_fw; }
	if (count < sizeof(struct usr_header)) 
	{
	    cprintf("check_n_fw: File is too small (%ld bytes)\n", count);
	    goto nonew_fw;
	}
	cprintf("compat_id: %d\n", hdr.compatibility_id);
	cprintf("hardware_rev: %d\n", hdr.hardware_revision);
	if ((hdr.magic != USR_MAGIC) || (hdr.compatibility_id != COMPAT_ID)
	 || (hdr.len < sizeof(struct usr_header)) || (hdr.hardware_revision < HARDWARE_REV)) 
	{
	    cprintf("check_n_fw: Bad usr header\n");
	    goto nonew_fw;
	}
	
	nM = hdr.version & 0xff;
	nm = (hdr.version >> 8) & 0xff;
	nrc = (hdr.version >> 16) & 0xff;
	ninc = (hdr.version >> 24) & 0xf0;
	nb = (hdr.version >> 24) & 0x0f;
	sprintf(sz_new_fw_version,"%03d.%03d.%03d.%03d.%03d",nM,nm,nrc,ninc,nb);
	sprintf(sz_new_fw_version1,"%d.%d.%d.%d.%d",nM,nm,nrc,ninc,nb);
	cprintf("check_n_fw: new version: %s\n", sz_new_fw_version);
	if (strcmp(sz_new_fw_version,curr_version) > 0)
	{
	    cprintf("check_n_fw: new firmware = %s\n",sz_new_fw_version);
	    //t = gmtime(&hdr.time);
	    //if (t) {
	    //    sprintf(sz_new_fw_date,"%s %#2d %d",
	//		sz_month[t->tm_mon], t->tm_mday, t->tm_year+1900);
	  //  }
	  //  cprintf("check_n_fw: new date = %s\n",sz_new_fw_date);
	    ret = 1;
	}
	cprintf("ret is: %d\n",ret);
    } else ret = -1;
    
nonew_fw:
    new_fw_available = ret;
    
    if (new_fw_available < 0) {
#ifdef ROUTER
        websWrite(stream, "<div class=\"clsNotice\">The router was unable to check if there is a new firmware available.");
#endif
#ifdef AP
        websWrite(stream, "<div class=\"clsNotice\">The access point was unable to check if there is a new firmware available.");
#endif
#ifdef APB
        websWrite(stream, "<div class=\"clsNotice\">The access point was unable to check if there is a new firmware available.");
#endif	
#ifdef REPEATER
        websWrite(stream, "<div class=\"clsNotice\">The range extender was unable to check if there is a new firmware available.");
#endif
#ifdef CLIENT
        websWrite(stream, "<div class=\"clsNotice\">The wireless client was unable to check if there is a new firmware available.");
#endif
#ifdef GAMING
        websWrite(stream, "<div class=\"clsNotice\">The gaming adapter was unable to check if there is a new firmware available.");
#endif
#ifdef BRIDGE
        websWrite(stream, "<div class=\"clsNotice\">The bridge was unable to check if there is a new firmware available.");
#endif				
	websWrite(stream, "<p>Please check your <b>Internet connection</b> and try again.</p></div>");
        websWrite(stream, (char_t *) apply_footer_check_fw_nonew,"device.asp" );
    } else if (new_fw_available == 0) {
        websWrite(stream, "<div class=\"clsNotice\"><p>No new version of firmware is available at this time.</p></div>");
        websWrite(stream, (char_t *) apply_footer_check_fw_nonew,"device.asp" );
    } else {
        //websWrite(stream, (char_t *) apply_header);
        websWrite(stream, (char_t *) apply_footer_check_fw_new, sz_new_fw_version1);
    }

    websFooter(stream);
    websDone(stream, 200);
}


#ifdef CLIENT
static void
do_index_asp(char *url, FILE *stream)
{
	char *path, *query;

	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;

	index_asp(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}

static void
do_advanced_asp(char *url, FILE *stream)
{
	char *path, *query;

	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;

	advanced_asp(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}
#endif

static void
do_wireless_asp(char *url, FILE *stream)
{
	char *path, *query;

	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;
	
	//printf("do_wireless_asp: %s %s\n", path, query);
	wireless_asp(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}

static void
do_security_asp(char *url, FILE *stream)
{
	char *path, *query;

	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;

	security_asp(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}

static void
do_setup_sec_asp(char *url, FILE *stream)
{
	char *path, *query;

	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;

	setup_sec_asp(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}

#ifdef CLIENT
static void
do_sitesurvey_asp(char *url, FILE *stream)
{
	char *path, *query;

	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;

	sitesurvey_asp(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}
#endif

#ifdef __CONFIG_NAT__
#ifdef ROUTER
static void
do_wan_asp(char *url, FILE *stream)
{
	char *path, *query;

	/* Parse path */
	query = url;
	path = strsep(&query, "?") ? : url;

	wan_asp(stream, NULL, NULL, 0, url, path, query);

	/* Reset CGI */
	init_cgi(NULL);
}
#endif  /* ROUTER */
#endif	/* __CONFIG_NAT__ */

/* For Backup/Restore settings */
static char backup_headers[] = 
"Cache-Control: no-cache\r\n"
"Pragma: no-cache\r\n"
"Expires: 0\r\n"
"Content-Type: application/download\r\n"
#ifdef ROUTER
"Content-Disposition: attachment ; filename=USR5461_backup.dat"
#endif
#ifdef CLIENT
"Content-Disposition: attachment ; filename=USR5430_backup.dat"
#endif
#ifdef AP
"Content-Disposition: attachment ; filename=USR5451_backup.dat"
#endif
#ifdef APB
"Content-Disposition: attachment ; filename=USR5453_backup.dat"
#endif
#ifdef REPEATER
"Content-Disposition: attachment ; filename=USR5441_backup.dat"
#endif
#ifdef GAMING
"Content-Disposition: attachment ; filename=USR5431_backup.dat"
#endif
#ifdef BRIDGE
"Content-Disposition: attachment ; filename=USR5432_backup.dat"
#endif
;
/* Old backup start */
/*
static void 
do_backup_cgi(char *path, FILE *stream)
{
	char *name, buf[NVRAM_SPACE];
	
	// TODO: some variables like os_version shouldn't be backup/restored.....
	printf("\nBackup NVRAM variables.....");
	nvram_getall(buf, sizeof(buf));
	for (name = buf; *name; name += strlen(name) + 1)
	{
		fputs(name, stream);
		fputs("\r\n", stream);
	}
	fputs("\r\n", stream); // terminator
	printf("DONE\n\n");
}

static void 
do_restore_post(char *url, FILE *stream, int len, char *boundary)
{
	FILE *fp;
	char buf[NVRAM_SPACE], buf2[1024];
	char *name, *next, *name2 = NULL, *value2;
	char tmp[] = "/tmp/restore.XXXXXX";
	int toset = 1, err = 1, dirty = 0;
	
	restore_ret = EINVAL;

	printf("\n\nReading restored data from HTTP.....\n");

	// Look for our part
	while (len > 0) {
		if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
		len -= strlen(buf);
		if (!strncasecmp(buf, "Content-Disposition:", 20) &&
		    strstr(buf, "name=\"file\""))
			break;
	}

	// Skip boundary and headers
	while (len > 0) {
		if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
		len -= strlen(buf);
		if (!strcmp(buf, "\n") || !strcmp(buf, "\r\n"))
			break;
	}
	
	if (mktemp(tmp) && (fp = fopen(tmp, "w")))
	{
		// Save the file first
		while (len > 0)
		{
			if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
				return;
			len -= strlen(buf);
			fputs(buf, fp);
			if (!strcmp(buf, "\n") || !strcmp(buf, "\r\n"))
				break;
		}
		fclose(fp);
		
		// Slurp anything remaining in the request
		while (len--)
			(void) fgetc(stream);

		// TODO: format checking should be done completely before writing anything 
		printf("Restoring settings.....\n");
		if ((fp = fopen(tmp, "r")))
		{
			// Put all current NVRAM variables into buf
			nvram_getall(buf, sizeof(buf));
			
			// Get NVRAM variable from file one by one
			while(fgets(buf2, sizeof(buf2), fp))
			{
				if (!strcmp(buf2, "\n") || !strcmp(buf2, "\r\n"))
					break;
				while (buf2[strlen(buf2)-1] == '\r' || buf2[strlen(buf2)-1] == '\n')
					buf2[strlen(buf2)-1] = 0;

				toset = 1;	
				name2 = buf2;	
				next = strchr(name2, '=');
				if (next == NULL) // format error
				{	
					err = 1;
					break;
				}
				err = 0; // at least we got one valid entry
				
				// Check if the variable existing in current settings
				for (name = buf; *name; name = name + strlen(name) + 1)
				{
					// Skip if it's already updated/skipped
					if (*name == '\1')
						continue;
						
					if (!strncmp(name2, name, (int)(next - name2)) && name[(int)(next - name2)] == '=')
					{
						// If found; check the value by comparing the whole string (name=value);
						// If it's the same, then no updating is required 
						if (!strcmp(name2, name))
							toset = 0;
							
						*name = '\1'; // Marked as it's been checked and/or updated
						break;
					}
				}

				if (toset) // Updating required? 
				{
					printf("  + update %s\n", name2);
					value2 = buf2;
					name2 = strsep(&value2, "=");
					//USR check compatibility ID
					if (!strcmp(name2, "COMPAT_ID") && (atoi(value2) != COMPAT_ID)) {
						printf("COMPAT_ID is different\n");
						err = 1;
						restore_ret = -2;
						break;
					}
					else if (!strcmp(name2, "HARDWARE_REV")) {
						printf("hardware_rev: %s\n", name2);
						if (atoi(value2) < HARDWARE_REV) {
							printf("Hardware_Rev is smaller than: %d\n", HARDWARE_REV);
							err = 1;
							restore_ret = -2;
							break;
						}
						else
							printf("Hardware_Rev is greater than: %d\n", HARDWARE_REV);
					}
					else {
						nvram_set(name2, value2);
						dirty++;
					}
				}
				else
				{
					printf("  . skip %s\n", name2);
				}
			}
			fclose(fp);
			
			if (!err)
			{
				// Remove all variables in buf, which are not marked, i.e.,
				//   remove all current variables which don't exist in backuped file. 
				for (name = buf; *name; name = name + strlen(name) + 1)
				{
					if (*name == '\1')
						continue; // marked, i.e., existing in new settings
						
					strcpy(buf2, name);
					value2 = name2;
					name2 = strsep(&value2, "=");
					printf("  - unset %s\n", name2);
					nvram_unset(name2);
					dirty++;
				}

				if (dirty) // anyone got updated? 
				{
					printf("Flushing settings.....\n");
					nvram_commit();
					restore_ret = 0;
					printf("Restore settings is done. Now rebooting.....\n\n");
				}
				else 
				{
					restore_ret = -1;
					printf("No settings has been updated. There is no need to reboot.\n\n");
				}
			}
			else
				printf("***** ERROR Restoring settings: Invalid file format!\n\n");
		}
		else
			printf("***** ERROR Restoring settings: Failed to open %s, errno=%d\n\n", tmp, errno);
			
		unlink(tmp);
	}
	else
	{
		printf("***** ERROR Restoring settings: Failed to create temp file, errno=%d\n\n", errno);

		// Slurp anything remaining in the request
		while (len--)
			(void) fgetc(stream);
	}
}

static void do_restore_cgi(char *url, FILE *stream)
{
	websHeader(stream);
	websWrite(stream, (char_t *) apply_header);
	websWrite(stream, "<h3>Restore Settings</h3>");
	if (restore_ret == 0) {
		websWrite(stream, "<p>New settings has been loaded.</p>");
		websWrite(stream, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
		websWrite(stream, (char_t *) apply_footer_reboot, "status.asp");
	}
	else if (restore_ret == -1) {
		websWrite(stream, "<p>The settings are exactly the same. No restore is required.</p>");
		websWrite(stream, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the <strong>Status</strong> page.</p>");
		websWrite(stream, (char_t *) apply_footer, "status.asp");
	}
	else if (restore_ret == -2) {
		websWrite(stream, "<p>Error occurred during restore. The settings are not compatible with your product. Please verify your back up settings.</p>");
		websWrite(stream, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
		websWrite(stream, (char_t *) apply_footer_reboot, "status.asp");
	}
	else {
		websWrite(stream, "<p>Error occurred during restore.</p>");
		websWrite(stream, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the <strong>Device</strong> page.</p>");
		websWrite(stream, (char_t *) apply_footer_error, "device.asp");
	}
	websFooter(stream);
	websDone(stream, 200);

	// Reboot if successful or is not compatible
	if ( (restore_ret == 0) || (restore_ret == -2))
		sys_reboot();
}
*/
/*Old backup end */

/* New backup start */
/* Format of the NVRAM text file is as follows
 * the following major sections
 * Header Block
 * Constraint Block
 * NVRAM variables
 *  
 * Header consists of :
 *	LineCount
 *	Checksum
 *
 * Constraint block consists of :
 * Defined by NVRAM_CONSTRAINT_VARS 
 * 
 *
*/  				
/* Get major os version. The nvram variable gives the whole thing.
 * this function avoids the messy parsing.
 * Used by NVRAM constraint validation routines.
 * The input argument is not used. This is used to maintain 
 * argument list compatability with nvram_get().
 *
*/ 
static  char*
osversion_get(const char *name)
{

	static char ret_string[32];
	
	assert(name);
  
 	snprintf(ret_string,sizeof(ret_string),"%d.%d",
  			EPI_MAJOR_VERSION,EPI_MINOR_VERSION);
		
		
	return ret_string;
	
}
/* This routine strips the CRLF from the HTTP stream */
/* Returns new length or -1 if there is an error*/
static int 
remove_crlf(char *buf, int len)
{

	int slen;
	if (!buf) return -1;
	if (!len) return 0;

	slen = strlen(buf);
	len = len-slen;
	if ((buf[slen-1] == '\n' ) && (buf[slen-2] == '\r' )){
		buf[slen-1]='\0';
		buf[slen-2]='\0';

	} 
	else if ( buf[slen-1] == '\n' ) buf[slen-1]='\0';
	
	return len;
}
/* Utility routine to add a string from a string buffer.
 * 
 * Routine expects that the buffer is composed of text strings with 
 * a null as the delimiter.
 *
 * Input is expected to be a null terminated string
 *
 * Inputs -Pointer to char buffer
 *        -NULL input terminated string
 *	  -Offset to end of used area. Including the NULL terminator
 *	  -Size of the buffer (to avoid running of the end of the buffer)
 *
 * 
 * Returns - Number of bytes occupied including NULL terminator. -1 if error
 *
 *
*/
static int
add_string(char *buf, char *string, int start, int bufsize)
{
	int len;
	
	assert(buf);
	
	if (!string) return -1;
	if (!*string) return -1;
	if (!bufsize) return -1;
	if (start >=bufsize) return -1;
	
	len = strlen(string);
	
	if ( (start + len + 1) >= bufsize ) return -1;
	
	strncpy(buf,string,len);
		
	return start + len + 1;
}
static void
do_restore_post(char *url, FILE *stream, int len, char *boundary)
{
	char buf[1024];
	int  index,cur_entry;
	char *ptr=NULL;
	char *name=NULL;
	char tmp[NVRAM_MAX_STRINGSIZE];
	char file_checksum[NVRAM_SHA1BUFSIZE],checksum[NVRAM_SHA1BUFSIZE];
	char key[NVRAM_FILEKEYSIZE];
	int  checksum_linenum = NVRAM_CHECKSUM_LINENUM;
	char salt[NVRAM_SALTSIZE];
	int  numlines=0;
	int  entries,offset,slen;
	char passphrase[]=NVRAM_PASSPHRASE;
	char nvram_file_header[][NVRAM_MAX_STRINGSIZE/2] = NVRAM_FILEHEADER;
	upload_constraints constraint_vars [] = NVRAM_CONSTRAINT_VARS;
	struct pb { 
			char header[NVRAM_HEADER_LINECOUNT(nvram_file_header)][NVRAM_MAX_STRINGSIZE];
			char buf[NVRAM_SPACE];
		 }*tmpbuf=NULL;

	assert(stream);
	
	entries = NVRAM_HEADER_LINECOUNT(nvram_file_header);
	restore_ret = EINVAL;
	
	/* Look for our part */
	while (len > 0) {
		if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;

		/* Remove LF that fgets()drags in */
		len=remove_crlf(buf,len);
		/* look for start of attached file header */

		if (*buf){
		  if (strstr(buf,NVRAM_BEGIN_WEBFILE)) break;
		}
		
	}
	
	if (!len) return;
	
	/* loop thru the header lines until we get the blank line
	   that signifies the start of file contents */
	while (len > 0){
	 	if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
		/* Remove LF that fgets()drags in */
		len=remove_crlf(buf,len);
		
		/* look for the blank line */
		if (*buf == NVRAM_END_WEBFILE)  break;
	}
	
	if (!len) return;
	
	/* Found start of upload file. Proceed with the upload */
	/* Look for start of upload data */
	if (!fgets(buf, MIN(len + 1, sizeof(buf)), stream))
			return;
					
	len = len - strlen(buf);
	
	/* Get the number of NVRAM variables */ 
	sprintf(tmp,"%s=%s",nvram_file_header[NVRAM_LINECOUNT_LINENUM],"%d");
	
	if ( (sscanf(buf,tmp,&numlines) != 1) || !len ){
		strncpy(posterr_msg,"<p>Invalid NVRAM header.</p>",ERR_MSG_SIZE);
	 	return ;
	}
	
	if ((numlines > NVRAM_MAX_LINES ) || !numlines ){
		strncpy(posterr_msg,"<p>Invalid NVRAM header.</p>",ERR_MSG_SIZE);
	 	return;
	}
	
	tmpbuf = (struct pb *) malloc( sizeof(struct pb));
	if (!tmpbuf){
		strncpy(posterr_msg,"<p>Memory allocation error.</p>",ERR_MSG_SIZE);
	 	return;
	} 
	
	memset(tmpbuf,0,sizeof(struct pb));
	
	/* Copy over the first line of the file. Needed for 
	   proper checksum calculations */
	   
	strcpy(tmpbuf->header[0],buf);
	
	/* read in checksum */
	if (!fgets(tmpbuf->header[1], MIN(len + 1,NVRAM_MAX_STRINGSIZE), stream))
			goto do_restore_post_cleanup0;
		
	len = len - strlen(tmpbuf->header[1]);
	
	/* start reading in the rest of the NVRAM contents into memory buffer*/
	offset = 0;
	for (index = 2 ; (len > 0) &&  (index < numlines); index++){
		char filebuf[NVRAM_MAX_STRINGSIZE];
		
	 	if (!fgets(filebuf, MIN(len + 1,NVRAM_MAX_STRINGSIZE), stream))
			goto do_restore_post_cleanup0;
		
		offset = add_string(&tmpbuf->buf[offset],filebuf,offset,
						sizeof(((struct pb *)0)->buf));
		if (offset < 0){
			snprintf(posterr_msg,ERR_MSG_SIZE,
				"<p>Error processing NVRAM variable:%s.</p>",filebuf);
	 		goto do_restore_post_cleanup0;
		}
		len = len - strlen(filebuf);

		/* don't remove the LFs since the we are using  multipart
		   MIME encoding that does not touch the contents of the
		   uploaded file. LFs are actually part of the NVRAM var lines.
		*/
	}
	
	/*  Bail if the number of actual lines is less than that
	    in the header 
	*/
	
	if ( !len && (index < numlines) ){
		strncpy(posterr_msg,"<p>NVRAM file incomplete.</p>",ERR_MSG_SIZE);
	 	goto do_restore_post_cleanup0;
	}
	
	/* Save and decode checksum from file */
	ptr = tmpbuf->header[checksum_linenum];
	
	sprintf(tmp,"%s=",nvram_file_header[checksum_linenum]);
	
	if ( !strstr(ptr,tmp) ) {
		snprintf(posterr_msg,ERR_MSG_SIZE,
			"<p>No checksum present at line %d.</p>",checksum_linenum+1);
		goto do_restore_post_cleanup0 ;
	}
	ptr = &ptr[strlen(tmp)];
	
	if ( b64_decode(ptr,tmp,NVRAM_MAX_STRINGSIZE) != NVRAM_FILECHKSUM_SIZE){
		strncpy(posterr_msg,"<p>Invalid checksum.</p>",ERR_MSG_SIZE);
		goto do_restore_post_cleanup0 ;
	};
	
	memset(file_checksum,0,sizeof(file_checksum));
	memcpy(file_checksum,tmp,NVRAM_HASHSIZE);
	
	/* Extract salt value from stored checksum*/
	memcpy(salt,&tmp[NVRAM_HASHSIZE],NVRAM_SALTSIZE);

	/* Regenerate encryption key */
	memset(key,0,NVRAM_FILEKEYSIZE);
	fPRF(passphrase,strlen(passphrase),NULL,0,
			(unsigned char*)salt,sizeof(salt),key,sizeof(key));
	
	/* Plug in filler for checksum into read buffer*/
	memset(tmpbuf->header[checksum_linenum],0,NVRAM_MAX_STRINGSIZE);
	snprintf(tmpbuf->header[checksum_linenum],
			NVRAM_MAX_STRINGSIZE,"%s\n",NVRAM_CHECKSUM_FILLER);
	
	/* Calculate checksum and compare with stored value*/ 
	
	memset(checksum,0,sizeof(checksum));
	hmac_sha1((unsigned char*)tmpbuf,sizeof(struct pb) ,
							key,sizeof(key),checksum);
	
	if (memcmp(checksum,file_checksum,NVRAM_HASHSIZE)){
		memcpy(posterr_msg,"<p>File checksum error.</p>",ERR_MSG_SIZE);
		goto do_restore_post_cleanup0;	
	}
	
	/* Check constraints on the data */
	
	cur_entry = NVRAM_HEADER_LINECOUNT(nvram_file_header);
	slen=0;
	name = tmpbuf->buf;
	for (index = 0  ; *constraint_vars[index].name && *name; index++,cur_entry++,name += slen + 1) {
		char *var=NULL,*ptr=NULL;
		int comparesize ;
		
		slen = strlen(name);
		
		if (cur_entry > numlines){		
			memcpy(posterr_msg,
				"<p>Constraints mismatch between file and running image.</p>",
										ERR_MSG_SIZE);
			goto do_restore_post_cleanup0;	
		}
	
	
		var = constraint_vars[index].get(constraint_vars[index].name);

		if (!var){
			snprintf(posterr_msg,ERR_MSG_SIZE,
				"<p>NVRAM variable:%s not found in running image.</p>",
								constraint_vars[index].name);
			goto do_restore_post_cleanup0;	
		}
		
		if (!strstr(name,constraint_vars[index].name)){
			snprintf(posterr_msg,ERR_MSG_SIZE,
				"<p>NVRAM variable:%s not found in uploaded file.</p>",
								constraint_vars[index].name);
			goto do_restore_post_cleanup0;	
		}
		
		/*Move past separator*/
		ptr = &name[strlen(constraint_vars[index].name) + 1] ;
		
		/* Ignore last character which is a \n */
		comparesize = strlen(ptr) -1;
		
		/* If the primary match fails try for the secomdary matches */
		if (strncmp(ptr,var,comparesize)){
		int sec_fail=0;
		
			/* If partial march is defined match altval pattern against 
			   file and image values*/
			if (constraint_vars[index].flags & NVRAM_CONS_PARTIAL_MATCH){
				sec_fail=( !strstr(ptr,constraint_vars[index].altval) && 
							!strstr(var,constraint_vars[index].altval));		
			}
			else if (constraint_vars[index].flags & NVRAM_CONS_ALT_MATCH){
				sec_fail=(strncmp(ptr,constraint_vars[index].altval,comparesize));
			}
			else sec_fail =1;
			
		if (sec_fail){
			snprintf(posterr_msg,ERR_MSG_SIZE,
				"<p>NVRAM constraint mismatch FILE:<b>%s=%s</b>  IMAGE:<b>%s=%s</b>.</p>",
					constraint_vars[index].name,ptr,constraint_vars[index].name,var);
			goto do_restore_post_cleanup0;	
			}
		}
	
	}

	/* Process the NVRAM payload 
	 *
	 * Remove the carriage returns at the end of the NVRAM variables 
	 *
	*/
	
	cur_entry = NVRAM_HEADER_LINECOUNT(nvram_file_header) + NVRAM_HEADER_LINECOUNT(constraint_vars);
	
	slen=0;
	for (index = cur_entry ;(index < numlines) && *name; index++,name += slen + 1){
		int offset;
		char *ptr=NULL;		
		char *varname=NULL;
		char *myptr=NULL;
		struct variable *v=NULL;
	
		offset = index * NVRAM_MAX_STRINGSIZE;
		
		/* The length of the actual var must be saved here as the subsequent
		 * manipulation using strsep() changes the string buffer.
		 * This to ensure that the buffer (tmpbuf->buf) is correctly procesed
		 * This buffer separates the individual AVPs using a null character
		 */
		  
		slen =strlen(name);
		ptr = name;
		
		/* Remove the CR at the end of the string */
		ptr[slen-1] = '\0';
		
		/* Look for tag that indicates that the value is encrypted */
		if  (*ptr==NVRAM_ENCTAG){
			char buf[NVRAM_MAX_STRINGSIZE];
			int bufsize = sizeof(buf);
			char *varname=NULL;
		
			varname=strsep(&ptr, "=");
		
			/* Increment pointer to move past tag */
			varname++;

			if (!decrypt_var(varname,ptr,strlen(ptr),buf,&bufsize,key,sizeof(key))){
				snprintf(posterr_msg,ERR_MSG_SIZE,
					"<p>Error decrypting value %s at line %d.</p>",ptr,index);
				goto do_restore_post_cleanup0;	
			}
			
		snprintf(name,NVRAM_MAX_STRINGSIZE,"%s=%s",varname,buf);

		}
				
		/* 
		   Write out NVRAM variables.
		*/
		
		myptr = name;
		varname=strsep(&myptr, "=");
		
		v=get_var_handle(varname);
		
		/* Restore only those NVRAM vars in the validation table 
		   Ignore those with the obvious flag set
		*/
		
		if (!v) 
			continue;
		
		if (v->ezc_flags & NVRAM_IGNORE)
			continue;
			
		if (nvram_set(varname,myptr)) 
			goto do_restore_post_cleanup0;
			
	}
	
	nvram_commit();

	/* We are done */
	restore_ret = 0;
	
do_restore_post_cleanup0:
	/* Clear up any outstanding stuff */
	/* Slurp anything remaining in the request */
	while (len--)
		(void) fgetc(stream);
	
	if (tmpbuf) free(tmpbuf);

	return;
}


static void
do_restore_cgi(char *url, FILE *stream)
{
	assert(stream);
	assert(url);
	
	websHeader(stream);
	websWrite(stream, (char_t *) apply_header);
	websWrite(stream, "<h3>Restore Settings</h3>");
	
	if (restore_ret){
		websWrite(stream, "<p>Error occurred during restore.</p>");
		if (*posterr_msg){
			websWrite(stream, posterr_msg);
			memset(posterr_msg,0,ERR_MSG_SIZE);
		}
		websWrite(stream, "<p class=\"clsSave\">Please press the <strong>Continue</strong> button to proceed to the <strong>Device</strong> page.</p>");
		websWrite(stream, (char_t *) apply_footer_error, "device.asp");
		
		
	} else {
		websWrite(stream, "<p>New settings has been loaded.</p>");
		websWrite(stream, "<p id=\"idWait\">Please wait <b><span id=\"idSeconds\"></span></b> while the device reboots.</p>");
		websWrite(stream, (char_t *) apply_footer_reboot, "status.asp");
	}
	
	websFooter(stream);
	websDone(stream, 200);
	
	/* Reboot if successful */
	if (restore_ret == 0)
		sys_reboot();

}

/* 
   This routine validates and formats the NVRAM variable into the file output format.
   If the variable is not in the monster V-block or its allowed multi instance variants
   it is dropped 
   
   Returns number of characters printed or -1 on error
*/
static int
save_nvram_var(char *name, char *var_val,char *buf, int buflen,char *key, int keylen)
{
	char tmp[NVRAM_MAX_STRINGSIZE];
	int len=sizeof(tmp);
	int retval=0;
	struct variable *v = NULL; 
	
	assert(name);
	assert(buf);
	assert(key);
	assert(buflen);
	assert(keylen);
	

	/* If var_val is null, this forces the variable to be unset when the file is uploaded 
	  If the variable is supposed to be encrypted but is null, skip it and do not 
	  mark the string as encrypted.
				  
	  If var_val is null the variable does not exist. Skip and do not save in that case.
	*/
	
	v = get_var_handle(name);
	
	if (!v) 
		return 0;
	
	if (v->ezc_flags & NVRAM_IGNORE)
		return 0;
	
	if (var_val) {	
			if (strlen(var_val) > NVRAM_MAX_STRINGSIZE){
				cprintf("get_nvram_var():String too long Len=%d String=%s\n",	strlen(var_val) ,var_val);	
				return -1;
			}
			
			if ( (v->ezc_flags & NVRAM_ENCRYPT) && (*var_val) ){
				var_val = encrypt_var(name,var_val,strlen(var_val),tmp,&len,key,keylen);
					
				if (!var_val){
					cprintf("get_nvram_var():Error encrypting %s\n",name);
					return -1;
				}
					
				retval=snprintf(buf,buflen,"%c%s=%s\n",
						NVRAM_ENCTAG,
						(char_t *)name,
						(char_t *)var_val);
		    } else retval=snprintf(buf,buflen,"%s=%s\n",
						(char_t *)name,
						(char_t *)var_val);			
	}
	return retval;
}

/* This is the cgi handler for the NVRAM download function
 * Inputs: -url of the calling file (not used but HTTPD expects this form)
 *         -Pointer to the post buffer
 *	
 *	   
 *  Returns: None	  
*/
static void
//do_nvramdl_cgi(char *url, FILE *stream)
do_backup_cgi(char *url, FILE *stream)
{
	char checksum[NVRAM_SHA1BUFSIZE];
	char passphrase[]=NVRAM_PASSPHRASE;
	char salt[NVRAM_SALTSIZE];
	char key[NVRAM_FILEKEYSIZE];
	int entries;
	char tmp[NVRAM_MAX_STRINGSIZE],tmp1[NVRAM_MAX_STRINGSIZE];
	char tmp_buf[NVRAM_MAX_STRINGSIZE];
	char *buf=NULL;
	char *var_val=NULL;
	char *var_name=NULL;
	char *name=NULL;
	char *ptr=NULL;
	int index;
	int retval;
	upload_constraints constraint_vars [] = NVRAM_CONSTRAINT_VARS;
	char nvram_file_header[][NVRAM_MAX_STRINGSIZE/2] = NVRAM_FILEHEADER;
	struct pb { 
			char header[NVRAM_HEADER_LINECOUNT(nvram_file_header)][NVRAM_MAX_STRINGSIZE];
			char buf[NVRAM_SPACE];
		 } *post_buf=NULL;

	int offset = 0;
	
	assert(stream);
	
	post_buf = (struct pb *)malloc(sizeof (struct pb));
	
	if (!post_buf) {
		cprintf("do_nvramdl_cgi():Error allocating %d bytes for post_buf\n",
						sizeof (struct pb));
		goto do_nvramdl_cgi_error;
	}
	
	buf = (char *)malloc(NVRAM_SPACE);
	
	if (!buf) {
		cprintf("do_nvramdl_cgi():Error allocating %d bytes for buf\n",
						NVRAM_SPACE);
		goto do_nvramdl_cgi_error;
	}
	
	memset (post_buf,0,sizeof(struct pb));
	memset (buf,0,NVRAM_SPACE);

	assert(stream);
	
	entries = NVRAM_HEADER_LINECOUNT(nvram_file_header);
	memset(tmp_buf,0,sizeof(tmp_buf));
	
	memset(salt,0,sizeof(salt));

	srand(time((time_t *)NULL));
	for (index = 0 ; index < 30 ; index ++) rand();
	index =rand();
	memcpy(&salt[sizeof(index)],&index,sizeof(index));
	for (index = 0 ; index < 30 ; index ++) rand();
	index =rand();
	memcpy(&salt,&index,sizeof(index));

	/*
	   The first entry of the file is the number of variables
	   The second entry is the offset to the start of the NVRAM variables 
	   The third entry is the SHA1 checksum
	*/

	/* PopulateHeader info */
	offset = 0;
	for  (index =0 ;*constraint_vars[index].name;index++){
		entries++;
		var_val=constraint_vars[index].get(constraint_vars[index].name);
		snprintf(tmp_buf,NVRAM_MAX_STRINGSIZE,"%s=%s\n",
					constraint_vars[index].name,
					(var_val) ? var_val : "unknown");
		if (!offset) cprintf("Post_buf address = %p\n",&post_buf->buf[offset]);
		offset = add_string(&post_buf->buf[offset],tmp_buf,offset,sizeof(struct pb));
		if (offset < 0){
			cprintf("httpd: Error Adding NVRAM header info.\n");
			goto do_nvramdl_cgi_error;
		}
	};
		
	memset(tmp,0,sizeof(tmp));

	fPRF(passphrase,strlen(passphrase),NULL,0,
			(unsigned char*)salt,sizeof(salt),key,sizeof(key));	
	
	/* Plug in filler for checksum */
	snprintf(post_buf->header[NVRAM_CHECKSUM_LINENUM],NVRAM_MAX_STRINGSIZE,"%s\n",NVRAM_CHECKSUM_FILLER);
	/* Grab the NVRAM buffer */

	nvram_getall(buf, NVRAM_SPACE);
	
	for(name = buf; *name; name += strlen(name) + 1){
		var_val =tmp;
		strncpy(tmp,name,sizeof(tmp));
		var_name=strsep(&var_val,"=");
		retval = save_nvram_var(var_name,var_val,tmp_buf,
						sizeof(tmp_buf),key,sizeof(key));
						
		if (retval <0){
			cprintf("httpd: Error Adding NVRAM variable info.\n");
			goto do_nvramdl_cgi_error;
		};
		
		if  (retval > 0 ) {
			entries++;
			offset = add_string(&post_buf->buf[offset],tmp_buf,offset,sizeof(struct pb));
			if (offset < 0){
				cprintf("httpd: Error Adding NVRAM variable info.\n");
				goto do_nvramdl_cgi_error;
			}
		};
		
		if (offset > NVRAM_SPACE) {
			cprintf("httpd: NVRAM_SPACE of %d (%d) exceeded\n",NVRAM_SPACE,offset);
			goto do_nvramdl_cgi_error;
		};
	}
	
	/* Add the header info */
	snprintf(post_buf->header[NVRAM_LINECOUNT_LINENUM],NVRAM_MAX_STRINGSIZE,"%s=%d\n",
					nvram_file_header[NVRAM_LINECOUNT_LINENUM],entries);
	
	/*Generate the hash */

	memset(checksum,0,sizeof(checksum));
	hmac_sha1((unsigned char*)post_buf,sizeof(struct pb),key,sizeof(key),checksum);
		
	memcpy(tmp,checksum,NVRAM_HASHSIZE);
	memcpy(&tmp[NVRAM_HASHSIZE],salt,NVRAM_SALTSIZE);

	ptr = b64_encode(tmp,NVRAM_FILECHKSUM_SIZE, buf, NVRAM_SPACE );
	
	if (!ptr){
		cprintf("do_nvramdl_cgi():Error performing base-64 encode of NVRAM checksum.\n");
		goto do_nvramdl_cgi_error;
	}
	
	strncpy(tmp1,ptr,NVRAM_MAX_STRINGSIZE);
	snprintf(post_buf->header[NVRAM_CHECKSUM_LINENUM],NVRAM_MAX_STRINGSIZE,"%s=%s\n",
					nvram_file_header[NVRAM_CHECKSUM_LINENUM],tmp1);
	
	/* Write out header */
	for (index =0; index < NVRAM_HEADER_LINECOUNT(nvram_file_header) ; index ++) {
		websWrite(stream, "%s", post_buf->header[index]);
	}
		
	/* Write out rest of file */

	for(name = post_buf->buf; *name; name += strlen(name) + 1){
		/*cprintf("Val->%s\n",name);*/
		websWrite(stream, "%s", name);
		};
	
do_nvramdl_cgi_error:

	if (post_buf) 
		free(post_buf);
	if (buf)
		free(buf);
		
	websDone(stream, 200);
	
	return;
}


/* 
 * This routine takes the cipher text ctext and produces the plaintext ptext
 * using the provided key.
 *
 * This routine accepts the cipher text  in the MIME Base64 format.
 *
 * The plain text is 8 bytes less than the ciphertext
 */
static char*
decrypt_var(char *varname,char *ctext, int ctext_len, char *ptext, int *ptext_len,char *key, int keylen)
{
	unsigned char tmp[NVRAM_MAX_STRINGSIZE];
	int len;
	char *end=NULL;
	
	assert(ptext);
	assert(ctext);
	assert(ptext_len);
	assert(key);
	if (keylen < 1 ) return NULL;

	if (ctext_len > NVRAM_MAX_STRINGSIZE){
		cprintf("decrypt_var():Encrypted string is too long MAXSTRINGSIZE=%d Strlen=%d\n",
							NVRAM_MAX_STRINGSIZE,ctext_len);
	}

	/* Cipher text must be more than 8 chars for this aes_unwrap() routine to work*/
	if (ctext_len < 8) return NULL;
	
	len=b64_decode(ctext,tmp,NVRAM_MAX_STRINGSIZE);
	
	if (!len) return NULL;
	
	if (aes_unwrap(keylen,key,len,tmp,ptext)) return NULL;
	
	*ptext_len = len - 8;
	
	end = strstr(ptext,varname);
	
	if (end) 
		(*end)= '\0';
	else
		return NULL;
	
	return ptext;
}
/* 
 * This routine takes the plaintext text ptext and produces the ciphertext ctext
 * using the provided key.
 *
 * It accepts the plaintext in binary form and produces ciphertext in MIME Base64 format.
 *
 * 
 */
static char* 
encrypt_var(char *varname,char *ptext, int ptext_len, char *ctext, int *ctext_len,char *key, int keylen)
{
	unsigned char tmp[NVRAM_MAX_STRINGSIZE];
	char *buf=NULL;
	int newlen;
	int varname_len;
	
	assert(ptext);
	assert(ctext);
	assert(ctext_len);
	assert(key);
	if (keylen < 1 ) return NULL;
	if (ptext_len < 1) return NULL;
	
	varname_len = strlen (varname);
	 
	 /* Include the NULL at the end */
	newlen = ptext_len + varname_len + 1;
	/* Align the incoming buffer to AES block length boundaries */
	if (newlen % AES_BLOCK_LEN)
		newlen = (1 + newlen /AES_BLOCK_LEN ) * AES_BLOCK_LEN;

	/* Do a string length check. When the binary string is base-64 encoded
	   it becomes 30% larger as every 3 bytes are represented by 4 ascii
	   characters */
	  
	if ( (4*(1+(newlen+8)/3)) > NVRAM_MAX_STRINGSIZE ) 
	{
		cprintf("encrypt_var():The encrypted string is too long. MAXSTRINGSIZE=%d Strlen=%d\n",
				NVRAM_MAX_STRINGSIZE,(4*(1+(newlen+8)/3)));
		return NULL;
	}
		
	buf  = malloc (newlen);
	if (!buf) return NULL;
	memset(buf,0,newlen);
	memcpy(buf,ptext,ptext_len);
	memcpy(buf + ptext_len,varname,varname_len);
			
	if (aes_wrap(keylen,key,newlen,buf,ctext)){
		if (buf) free(buf);
	 	return NULL;
	};
	
	if (buf) free(buf);
	
	buf = b64_encode(ctext,newlen+8,tmp,NVRAM_MAX_STRINGSIZE-(ptext_len + 1));

	if (buf){ 
		strncpy(ctext,buf,NVRAM_MAX_STRINGSIZE);
		*ctext_len = strlen(ctext) ;
		return ctext;
	}else{
		cprintf("encrypt_var():base-64 encode error\n");
		*ctext_len = 0;
		return NULL;
	}
}
/* New backup end */

struct mime_handler mime_handlers[] = {
#ifdef __CONFIG_NAT__
#ifdef ROUTER
	{ "wan.asp", "text/html", no_cache, do_apply_post, do_wan_asp, do_auth },
#endif
#endif	 /* __CONFIG_NAT__ */
	{ "wireless.asp", "text/html", no_cache, do_apply_post, do_wireless_asp, do_auth },
	{ "security.asp", "text/html", no_cache, do_apply_post, do_security_asp, do_auth },
	{ "setup_sec.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "setup_login.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "log.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "status.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
//#ifdef __CONFIG_EZC__
//	{ "ezconfig.asp", "text/html", ezc_version, do_apply_ezconfig_post, do_ezconfig_asp, do_auth },
//#endif /* __CONFIG_EZC__ */
#ifdef CLIENT
	{ "index.asp", "text/html", no_cache, do_apply_post, do_index_asp, do_auth },
	{ "ssid.asp", "text/html", no_cache, do_apply_post, do_index_asp, do_auth },
	{ "peer.asp", "text/html", no_cache, do_apply_post, do_index_asp, do_auth },
	{ "advanced.asp", "text/html", no_cache, do_apply_post, do_advanced_asp, do_auth },
	{ "sitesurvey.asp", "text/html", no_cache, do_apply_post, do_sitesurvey_asp, do_auth },
#endif
#ifdef GAMING
	{ "setup_mode.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "setup_infra.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "setup_adhoc.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
#endif
#ifdef BRIDGE
	{ "setup_mode.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "setup_infra.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "setup_adhoc.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
	{ "setup_wds.asp", "text/html", no_cache, do_apply_post, do_setup_sec_asp, do_auth },
#endif
	{ "**.asp", "text/html", no_cache, NULL, do_ej, do_auth },
	{ "**.css", "text/css", NULL, NULL, do_file, do_auth },
	{ "**.gif", "image/gif", NULL, NULL, do_file, do_auth },
	{ "**.jpg", "image/jpeg", NULL, NULL, do_file, do_auth },
	{ "**.png", "image/png", NULL, NULL, do_file, do_auth },
	{ "**.js", "text/javascript", NULL, NULL, do_file, do_auth },
	{ "apply.cgi*", "text/html", no_cache, do_apply_post, do_apply_cgi, do_auth },
	{ "upgrade.cgi*", "text/html", no_cache, do_upgrade_post, do_upgrade_cgi, do_auth },
	{ "check_fw.cgi*", "text/html", no_cache, do_check_fw_post, do_check_fw_cgi, do_auth },
	//{ "backup.cgi*", "application/octet-stream", backup_headers, NULL, do_backup_cgi, do_auth },
 	//{ "restore.cgi*", "text/html", no_cache, do_restore_post, do_restore_cgi, do_auth },
	/* set MIME type to NULL to override the one built into the webserver. download_hdr
	   defines its content type
	*/
	{ "backup.cgi*", NULL, backup_headers, NULL, do_backup_cgi, do_auth },
	{ "restore.cgi*", NULL, "text/html", do_restore_post, do_restore_cgi , do_auth },
	{ NULL, NULL, NULL, NULL, NULL, NULL }
};

struct ej_handler ej_handlers[] = {
	{ "nvram_get", ej_nvram_get },
	{ "nvram_match", ej_nvram_match },
	{ "nvram_invmatch", ej_nvram_invmatch },
	{ "nvram_list", ej_nvram_list },
	{ "nvram_inlist", ej_nvram_inlist },
	{ "nvram_invinlist", ej_nvram_invinlist },
#ifdef __CONFIG_NAT__
#ifdef ROUTER
	//{ "wan_list", ej_wan_list },
	//{ "wan_iflist", ej_wan_iflist },
	{ "wan_route", ej_wan_route },
	//{ "wan_link", ej_wan_link },
	{ "wan_lease", ej_wan_lease },
	//{ "filter_client", ej_filter_client },
	//{ "forward_port", ej_forward_port },
	//{ "autofw_port", ej_autofw_port },
	{ "usr_filter_client", ej_usr_filter_client },
	{ "usr_forward_port", ej_usr_forward_port },
	{ "usr_autofw_port", ej_usr_autofw_port },
#endif /* ROUTER */
#endif	/*  __CONFIG_NAT__ */
	{ "localtime", ej_localtime },
	{ "sysuptime", ej_sysuptime },
	{ "dumplog", ej_dumplog },
	{ "syslog", ej_syslog },
	//{ "wl_list", ej_wl_list },
	//{ "wl_guest_ssid", ej_wl_guest_ssid},
	{ "wl_cur_channel", ej_wl_cur_channel },
	//{ "wl_cur_phytype", ej_wl_cur_phytype },
#ifdef CLIENT
	{ "wl_phytypes", ej_wl_phytypes },
	{ "wl_radioid", ej_wl_radioid },
	{ "wl_corerev", ej_wl_corerev },	
	{ "wl_cur_country", ej_wl_cur_country },
	{ "wl_country_list", ej_wl_country_list },
	{ "wl_channel_list", ej_wl_channel_list },
#endif	
	{ "wl_auth_list", ej_wl_auth_list },
	//{ "wl_mode_list", ej_wl_mode_list },
	//{ "wl_inlist", ej_wl_inlist },
#ifdef REPEATER
	{ "wl_wds_status", ej_wl_wds_status },
#endif	
	//{ "wl_radio_roam_option", ej_wl_radio_roam_option},
	//{ "wl_ure_list", ej_ure_list },
	//{ "wl_ure_enabled", ej_ure_enabled },
	//{ "ses_button_display", ej_ses_button_display},
	//{ "ses_cl_button_display", ej_ses_cl_button_display},
	{ "lan_list", ej_lan_list },
	{ "wireless_list", ej_wireless_list },
	{ "lan_route", ej_lan_route },
	//{ "lan_guest_iflist", ej_lan_guest_iflist },
	{ "lan_leases", ej_lan_leases },
	//{ "wme_match_op", ej_wme_match_op },
	{ "wl_cur_mode", ej_wl_cur_mode },
#ifdef CLIENT
	{ "wl_cur_SSID", ej_wl_cur_SSID },
	{ "wl_cur_BSSID", ej_wl_cur_BSSID },
	{ "wl_scan", ej_wl_scan },
	{ "wl_cur_rate", ej_wl_cur_rate },
	{ "wl_cur_signal", ej_wl_cur_signal },	
	{ "nvram_adhoc", ej_nvram_adhoc },
#endif	
	{ "retrieve_mac", ej_retrieve_mac },
	//{ "ses_button_display", ej_ses_button_display},
#ifdef __CONFIG_NAT__
#ifdef ROUTER
	// U.S. Robotics primary connection auto-sense
	{ "wan_detection", wan_detection },
	{ "wan_f_detection", wan_f_detection },
	{ "check_pppoelogin", check_pppoelogin },
	{ "get_printer", ej_get_printer },
	//{ "get_fw_status", get_fw_status },
#endif //ROUTER
#endif	//  __CONFIG_NAT__
	{ "lan_clients", ej_lan_clients },
	{ "wl_clients", ej_wl_clients },
#ifdef REPEATER	
	{ "wl_scan", ej_wl_scan },
#endif
#ifdef GAMING	
	{ "wl_scan", ej_wl_scan },
	{ "wl_cur_BSSID", ej_wl_cur_BSSID },
	{ "wl_cur_rate", ej_wl_cur_rate },
	{ "wl_cur_signal", ej_wl_cur_signal },
#endif
#ifdef BRIDGE	
	{ "wl_scan", ej_wl_scan },
	{ "wl_cur_BSSID", ej_wl_cur_BSSID },
	{ "wl_cur_rate", ej_wl_cur_rate },
	{ "wl_cur_signal", ej_wl_cur_signal },
	{ "wl_wds_status", ej_wl_wds_status },
#endif 
	{ NULL, NULL }
};

#endif /* !WEBS */

/*
 * Country names and abbreviations from ISO 3166
 */
country_name_t country_names[] = {

{"COUNTRY Z1",		 "Z1"},
{"COUNTRY Z2",		 "Z2"},
{"AFGHANISTAN",		 "AF"},
{"ALBANIA",		 "AL"},
{"ALGERIA",		 "DZ"},
{"AMERICAN SAMOA", 	 "AS"},
{"ANDORRA",		 "AD"},
{"ANGOLA",		 "AO"},
{"ANGUILLA",		 "AI"},
{"ANTARCTICA",		 "AQ"},
{"ANTIGUA AND BARBUDA",	 "AG"},
{"ARGENTINA",		 "AR"},
{"ARMENIA",		 "AM"},
{"ARUBA",		 "AW"},
{"ASCENSION ISLAND",		 "AC"},
{"ASHMORE AND BARBUDA",		 "AG"},
{"AUSTRALIA",		 "AU"},
{"AUSTRIA",		 "AT"},
{"AZERBAIJAN",		 "AZ"},
{"BAHAMAS",		 "BS"},
{"BAHRAIN",		 "BH"},
{"BAKER ISLAND",		 "Z2"},
{"BANGLADESH",		 "BD"},
{"BARBADOS",		 "BB"},
{"BELARUS",		 "BY"},
{"BELGIUM",		 "BE"},
{"BELIZE",		 "BZ"},
{"BENIN",		 "BJ"},
{"BERMUDA",		 "BM"},
{"BHUTAN",		 "BT"},
{"BOLIVIA",		 "BO"},
{"BOSNIA AND HERZEGOVINA","BA"},
{"BOTSWANA",		 "BW"},
{"BOUVET ISLAND",	 "BV"},
{"BRAZIL",		 "BR"},
{"BRITISH INDIAN OCEAN TERRITORY", 	"IO"},
{"BRUNEI DARUSSALAM",	 "BN"},
{"BULGARIA",		 "BG"},
{"BURKINA FASO",	 "BF"},
{"BURUNDI",		 "BI"},
{"CAMBODIA",		 "KH"},
{"CAMEROON",		 "CM"},
{"CANADA",		 "CA"},
{"CAPE VERDE",		 "CV"},
{"CAYMAN ISLANDS",	 "KY"},
{"CENTRAL AFRICAN REPUBLIC","CF"},
{"CHAD",		 "TD"},
{"CHANNEL ISLANDS",		 "Z1"},
{"CHILE",		 "CL"},
{"CHINA",		 "CN"},
{"CHRISTMAS ISLAND",	 "CX"},
{"CLIPPERTON ISLAND",	 "CP"},
{"COCOS (KEELING) ISLANDS","CC"},
{"COLOMBIA",		 "CO"},
{"COMOROS",		 "KM"},
{"CONGO",		 "CG"},
{"CONGO, THE DEMOCRATIC REPUBLIC OF THE", "CD"},
{"COOK ISLANDS",	 "CK"},
{"COSTA RICA",		 "CR"},
{"COTE D'IVOIRE",	 "CI"},
{"CROATIA",		 "HR"},
{"CUBA",		 "CU"},
{"CYPRUS",		 "CY"},
{"CZECH REPUBLIC",	 "CZ"},
{"DENMARK",		 "DK"},
{"DJIBOUTI",		 "DJ"},
{"DOMINICA",		 "DM"},
{"DOMINICAN REPUBLIC", 	 "DO"},
{"ECUADOR",		 "EC"},
{"EGYPT",		 "EG"},
{"EL SALVADOR",		 "SV"},
{"EQUATORIAL GUINEA",	 "GQ"},
{"ERITREA",		 "ER"},
{"ESTONIA",		 "EE"},
{"ETHIOPIA",		 "ET"},
{"FALKLAND ISLANDS (MALVINAS)",	"FK"},
{"FAROE ISLANDS",	 "FO"},
{"FIJI",		 "FJ"},
{"FINLAND",		 "FI"},
{"FRANCE",		 "FR"},
{"FRENCH GUIANA",	 "GF"},
{"FRENCH POLYNESIA",	 "PF"},
{"FRENCH SOUTHERN TERRITORIES",	 "TF"},
{"GABON",		 "GA"},
{"GAMBIA",		 "GM"},
{"GEORGIA",		 "GE"},
{"GERMANY",		 "DE"},
{"GHANA",		 "GH"},
{"GIBRALTAR",		 "GI"},
{"GREECE",		 "GR"},
{"GREENLAND",		 "GL"},
{"GRENADA",		 "GD"},
{"GUADELOUPE",		 "GP"},
{"GUAM",		 "GU"},
{"GUANTANAMO BAY",		 "Z1"},
{"GUATEMALA",		 "GT"},
{"GUERNSEY",		 "GG"},
{"GUINEA",		 "GN"},
{"GUINEA-BISSAU",	 "GW"},
{"GUYANA",		 "GY"},
{"HAITI",		 "HT"},
{"HEARD ISLAND AND MCDONALD ISLANDS",	"HM"},
{"HOLY SEE (VATICAN CITY STATE)", 	"VA"},
{"HONDURAS",		 "HN"},
{"HONG KONG",		 "HK"},
{"HOWLAND ISLAND",		 "Z2"},
{"HUNGARY",		 "HU"},
{"ICELAND",		 "IS"},
{"INDIA",		 "IN"},
{"INDONESIA",		 "ID"},
{"IRAN, ISLAMIC REPUBLIC OF",		"IR"},
{"IRAQ",		 "IQ"},
{"IRELAND",		 "IE"},
{"ISRAEL",		 "IL"},
{"ITALY",		 "IT"},
{"JAMAICA",		 "JM"},
{"JAN MAYEN AMAICA",		 "Z1"},
{"JAPAN",		 "JP"},
{"JAPAN_1",		 "J1"},
{"JAPAN_2",		 "J2"},
{"JAPAN_3",		 "J3"},
{"JAPAN_4",		 "J4"},
{"JAPAN_5",		 "J5"},
{"JAPAN_6",		 "J6"},
{"JAPAN_7",		 "J7"},
{"JAPAN_8",		 "J8"},
{"JARVIS ISLAND",		 "Z2"},
{"JERSEY",		 "JE"},
{"JOHNSTON ATOLL",		 "Z2"},
{"JORDON",		 "JO"},
{"KAZAKHSTAN",		 "KZ"},
{"KENYA",		 "KE"},
{"KINGMAN REEF",		 "Z2"},
{"KIRIBATI",		 "KI"},
{"KOREA, DEMOCRATIC PEOPLE'S REPUBLIC OF", "KP"},
{"KOREA, REPUBLIC OF",	 "KR"},
{"KUWAIT",		 "KW"},
{"KYRGYZSTAN",		 "KG"},
{"LAO PEOPLE'S DEMOCRATIC REPUBLIC", 	"LA"},
{"LATVIA",		 "LV"},
{"LEBANON",		 "LB"},
{"LESOTHO",		 "LS"},
{"LIBERIA",		 "LR"},
{"LIBYAN ARAB JAMAHIRIYA","LY"},
{"LIECHTENSTEIN",	 "LI"},
{"LITHUANIA",		 "LT"},
{"LUXEMBOURG",		 "LU"},
{"MACAO",		 "MO"},
{"MACEDONIA, THE FORMER YUGOSLAV REPUBLIC OF",	 "MK"},
{"MADAGASCAR",		 "MG"},
{"MALAWI",		 "MW"},
{"MALAYSIA",		 "MY"},
{"MALDIVES",		 "MV"},
{"MALI",		 "ML"},
{"MALTA",		 "MT"},
{"MAN, ISLE OF",		 "IM"},
{"MARSHALL ISLANDS",	 "MH"},
{"MARTINIQUE",		 "MQ"},
{"MAURITANIA",		 "MR"},
{"MAURITIUS",		 "MU"},
{"MAYOTTE",		 "YT"},
{"MEXICO",		 "MX"},
{"MICRONESIA, FEDERATED STATES OF", 	"FM"},
{"MIDWAY ISLANDS", 	"Z2"},
{"MOLDOVA, REPUBLIC OF", "MD"},
{"MONACO",		 "MC"},
{"MONGOLIA",		 "MN"},
{"MONTSERRAT",		 "MS"},
{"MOROCCO",		 "MA"},
{"MOZAMBIQUE",		 "MZ"},
{"MYANMAR",		 "MM"},
{"NAMIBIA",		 "NA"},
{"NAURU",		 "NR"},
{"NEPAL",		 "NP"},
{"NETHERLANDS",		 "NL"},
{"NETHERLANDS ANTILLES", "AN"},
{"NEW CALEDONIA",	 "NC"},
{"NEW ZEALAND",		 "NZ"},
{"NICARAGUA",		 "NI"},
{"NIGER",		 "NE"},
{"NIGERIA",		 "NG"},
{"NIUE",		 "NU"},
{"NORFOLK ISLAND",	 "NF"},
{"NORTHERN MARIANA ISLANDS","MP"},
{"NORWAY",		 "NO"},
{"OMAN",		 "OM"},
{"PAKISTAN",		 "PK"},
{"PALAU",		 "PW"},
{"PALESTINIAN TERRITORY, OCCUPIED", 	"PS"},
{"PALMYRA ATOLL", 	"Z2"},
{"PANAMA",		 "PA"},
{"PAPUA NEW GUINEA",	 "PG"},
{"PARAGUAY",		 "PY"},
{"PERU",		 "PE"},
{"PHILIPPINES",		 "PH"},
{"PITCAIRN",		 "PN"},
{"POLAND",		 "PL"},
{"PORTUGAL",		 "PT"},
{"PUERTO RICO",		 "PR"},
{"QATAR",		 "QA"},
{"REUNION",		 "RE"},
{"ROMANIA",		 "RO"},
{"ROTA ISLAND",		 "Z1"},
{"RUSSIAN FEDERATION",	 "RU"},
{"RWANDA",		 "RW"},
{"SAINT HELENA",	 "SH"},
{"SAINT KITTS AND NEVIS","KN"},
{"SAINT LUCIA",		 "LC"},
{"SAINT PIERRE AND MIQUELON",	 	"PM"},
{"SAINT VINCENT AND THE GRENADINES", 	"VC"},
{"SAIPAN", 	"Z1"},
{"SAMOA",		 "WS"},
{"SAN MARINO",		 "SM"},
{"SAO TOME AND PRINCIPE","ST"},
{"SAUDI ARABIA",	 "SA"},
{"SENEGAL",		 "SN"},
{"SEYCHELLES",		 "SC"},
{"SIERRA LEONE",	 "SL"},
{"SINGAPORE",		 "SG"},
{"SLOVAKIA",		 "SK"},
{"SLOVENIA",		 "SI"},
{"SOLOMON ISLANDS",	 "SB"},
{"SOMALIA",		 "SO"},
{"SOUTH AFRICA",	 "ZA"},
{"SOUTH GEORGIA AND THE SOUTH SANDWICH ISLANDS", "GS"},
{"SPAIN",		 "ES"},
{"SRI LANKA",		 "LK"},
{"SUDAN",		 "SD"},
{"SURINAME",		 "SR"},
{"SVALBARD AND JAN MAYEN","SJ"},
{"SWAZILAND",		 "SZ"},
{"SWEDEN",		 "SE"},
{"SWITZERLAND",		 "CH"},
{"SYRIAN ARAB REPUBLIC", "SY"},
{"TAIWAN, PROVINCE OF CHINA", 		"TW"},
{"TAJIKISTAN",		 "TJ"},
{"TANZANIA, UNITED REPUBLIC OF",	"TZ"},
{"THAILAND",		 "TH"},
{"TIMOR-LESTE",		 "TL"},
{"TOGO",		 "TG"},
{"TOKELAU",		 "TK"},
{"TONGA",		 "TO"},
{"TRINIDAD AND TOBAGO",	 "TT"},
{"TRISTAN DA CUNHA",	 "TA"},
{"TUNISIA",		 "TN"},
{"TURKEY",		 "TR"},
{"TURKMENISTAN",	 "TM"},
{"TURKS AND CAICOS ISLANDS",		"TC"},
{"TUVALU",		 "TV"},
{"UGANDA",		 "UG"},
{"UKRAINE",		 "UA"},
{"UNITED ARAB EMIRATES", "AE"},
{"UNITED KINGDOM",	 "GB"},
{"UNITED STATES",	 "US"},
{"UNITED STATES MINOR OUTLYING ISLANDS","UM"},
{"URUGUAY",		 "UY"},
{"UZBEKISTAN",		 "UZ"},
{"VANUATU",		 "VU"},
{"VENEZUELA",		 "VE"},
{"VIET NAM",		 "VN"},
{"VIRGIN ISLANDS, BRITISH", "VG"},
{"VIRGIN ISLANDS, U.S.", "VI"},
{"WALLIS AND FUTUNA",	 "WF"},
{"WESTERN SAHARA", 	 "EH"},
{"YEMEN",		 "YE"},
{"YUGOSLAVIA",		 "YU"},
{"ZAMBIA",		 "ZM"},
{"ZIMBABWE",		 "ZW"},
{"EUROPE",               "ETSI"},
{"ALL",		 	 "ALL"},
{"RADAR CHANNELS",	 "RDR"},
{NULL, 			 NULL}
};
